(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(require('substance-test'), require('substance')) :
	typeof define === 'function' && define.amd ? define(['substance-test', 'substance'], factory) :
	(factory(global.window.substanceTest,global.window.substance));
}(this, (function (substanceTest,substance) { 'use strict';

class TestEditor extends substance.AbstractEditor {

  constructor(...args) {
    super(...args);
    this.handleActions({
      domSelectionRendered: function() {}
    });
  }

  render($$) {
    let doc = this.editorSession.getDocument();
    let el = $$('div');
    let body = $$(substance.ContainerEditor, {
      node: doc.get('body')
    }).ref('surface');
    el.append(body);
    return el
  }
}

class TestNode extends substance.DocumentNode {}

TestNode.type = 'test-node';

TestNode.schema = {
  boolVal: { type: "boolean", default: false },
  stringVal: { type: "string", default: "" },
  arrayVal: { type: ["array","string"], default: [] },
  objectVal: { type: "object", default: {} },
};

TestNode.isBlock = true;

class TestBlockNode extends substance.BlockNode {}
TestBlockNode.type = 'test-block';

class TestContainerAnnotation extends substance.ContainerAnnotation {}

TestContainerAnnotation.type = 'test-container-anno';

class StructuredNode extends substance.DocumentNode {}

StructuredNode.schema = {
  type: "structured-node",
  title: "text",
  body: "text",
  caption: "text"
};

StructuredNode.isBlock = true;

class TestStructuredNodeComponent extends substance.Component {

  render($$) {
    var node = this.props.node;
    var el = $$('div').addClass('sc-structured-node');
    el.append(
      $$(substance.TextPropertyEditor, {
        disabled: this.props.disabled,
        path: [node.id, 'title']
      }).ref('titleEditor')
    );
    el.append(
      $$(substance.TextPropertyEditor, {
        disabled: this.props.disabled,
        path: [node.id, 'body']
      }).ref('bodyEditor')
    );
    el.append(
      $$(substance.TextPropertyEditor, {
        disabled: this.props.disabled,
        path: [node.id, 'caption']
      }).ref('captionEditor')
    );
    return el
  }
}

TestStructuredNodeComponent.fullWidth = true;

class TestInlineNode extends substance.InlineNode {}

TestInlineNode.type = 'test-inline-node';

TestInlineNode.schema = {
  content: 'text'
};

class TestInlineNodeComponent extends substance.InlineNodeComponent {
  didMount() {
    super.didMount();
    this.context.editorSession.onRender('document', this.rerender, this, {
      path: [this.props.node.id, 'content']
    });
  }

  dispose() {
    super.dispose();
    this.context.editorSession.off(this);
  }

  getClassNames() {
    // ATTENTION: ATM it is necessary to add .sc-inline-node
    return 'sc-test-inline sc-inline-node'
  }

  renderContent($$) {
    return $$('span').append(this.props.node.content)
  }
}

class TestContainerComponent extends substance.Component {

  render($$) {
    var el = $$('div').addClass('sc-container');
    el.append(
      $$(substance.ContainerEditor, {
        node: this.props.node
      }).ref('editor')
    );
    return el
  }
}

TestContainerComponent.fullWidth = true;

class TestMetaNode extends substance.DocumentNode {}

TestMetaNode.schema = {
  type: 'meta',
  title: 'text'
};

var TestMetaNodeXMLConverter = {

  tagName: 'meta',
  type: 'meta',

  import: function(el, node, converter) {
    node.id = 'meta';
    var titleEl = el.find('title');
    if (titleEl) {
      node.title = converter.annotatedText(titleEl, ['meta', 'title']);
    } else {
      node.title = '';
    }
  },

  export: function(node, el, converter) {
    el.append(el.createElement('title').append(
      converter.annotatedText(['meta', 'title'])
    ));
  }

};

class TestHTMLImporter extends substance.HTMLImporter {

  convertDocument(documentEl) {
    this.state.doc = this.createDocument();
    var bodyEl = documentEl.find('body');
    this.convertContainer(bodyEl.children, 'body');
  }

}

class TestXMLImporter extends substance.XMLImporter {

  convertDocument(documentEl) {
    this.state.doc = this.createDocument();
    var bodyEl = documentEl.find('body');
    this.convertContainer(bodyEl.children, 'body');
  }

}

class TestHTMLExporter extends substance.HTMLExporter {

  convertDocument(doc) {
    let el = this.createElement('div');
    el.append(
      this.convertContainer(doc.get('body'))
    );
    return el
  }

}

class TestXMLExporter extends substance.XMLExporter {

  convertDocument(doc) {
    var articleEl = this.createElement('article');
    var body = doc.get('body');
    articleEl.append(
      this.convertContainer(body)
    );
    return articleEl
  }

}

class TestArticle extends substance.Document {

  constructor(schema) {
    super(schema);
    this.create({
      type: "container",
      id: "body",
      nodes: []
    });
  }

}

const { InlineWrapper, InlineWrapperComponent } = substance.InlineWrapperPackage;

function getTestConfig() {
  let config = new substance.Configurator();
  config.addToolGroup('default');
  config.addToolGroup('annotations');
  config.addToolGroup('overlay');
  config.addToolGroup('insert');
  config.defineSchema({
    name: 'test-article',
    DocumentClass: TestArticle,
    defaultTextType: 'paragraph',
    version: 1.0
  });

  config.import(substance.ParagraphPackage);
  config.import(substance.HeadingPackage);
  config.import(substance.StrongPackage);
  config.import(substance.EmphasisPackage);
  config.import(substance.LinkPackage);
  config.import(substance.ListPackage);
  config.import(substance.TablePackage);
  config.import(substance.CodeblockPackage);
  config.import(substance.FilePackage);
  config.import(substance.ImagePackage);

  config.addComponent('container', TestContainerComponent);

  config.addNode(TestNode);

  config.addNode(TestBlockNode);
  config.addComponent(TestBlockNode.type, substance.Component);

  config.addNode(TestContainerAnnotation);

  config.addNode(TestInlineNode);
  config.addComponent(TestInlineNode.type, TestInlineNodeComponent);

  config.addNode(StructuredNode);
  config.addComponent(StructuredNode.type, TestStructuredNodeComponent);

  config.addNode(TestMetaNode);
  config.addConverter('xml', TestMetaNodeXMLConverter);

  config.addNode(InlineWrapper);
  config.addComponent(InlineWrapper.type, InlineWrapperComponent);


  config.addImporter('html', TestHTMLImporter);
  config.addImporter('xml', TestXMLImporter);
  config.addExporter('html', TestHTMLExporter);
  config.addExporter('xml', TestXMLExporter);

  return config
}

function createEditorSession(...seeds) {
  let config = getTestConfig();
  let doc = new substance.Document(config.getSchema());
  let body = doc.create({
    type: 'container',
    id: 'body'
  });
  seeds.forEach((seed)=>{
    seed(doc, body);
  });
  let editorSession = new substance.EditorSession(doc, { configurator: config });
  let first = body.getNodeAt(0);
  if (first) {
    if (first.isText()) {
      editorSession.setSelection({
        type: 'property',
        path: first.getTextPath(),
        startOffset: 0,
        containerId: 'body',
        surfaceId: 'body'
      });
    } else if (first.isList()) {
      editorSession.setSelection({
        type: 'property',
        path: first.getItemAt(0).getTextPath(),
        startOffset: 0,
        containerId: 'body',
        surfaceId: 'body'
      });
    } else {
      editorSession.setSelection({
        type: 'node',
        nodeId: first.id,
        mode: 'before',
        containerId: 'body',
        surfaceId: 'body'
      });
    }
  }
  return editorSession
}

function getMountPoint(t) {
  let mountPoint;
  if (substance.platform.inBrowser) {
    mountPoint = t.sandbox;
  } else {
    let htmlDoc = substance.DefaultDOMElement.createDocument('html');
    mountPoint = htmlDoc.createElement('div');
    htmlDoc.appendChild(mountPoint);
  }
  return mountPoint
}

function setupEditor(t, ...f) {
  let editor = TestEditor.mount({ editorSession: createEditorSession(...f) }, getMountPoint(t));
  let editorSession = editor.editorSession;
  let doc = editorSession.getDocument();
  let surface = editor.refs.surface;
  return { editor, editorSession, doc, surface }
}

var containerAnnoSample = function(tx) {
  let body = tx.get('body');
  tx.create({
    type: 'paragraph',
    id: 'p1',
    content: '0123456789'
  });
  tx.create({
    type: 'paragraph',
    id: 'p2',
    content: '0123456789'
  });
  tx.create({
    type: 'paragraph',
    id: 'p3',
    content: '0123456789'
  });
  tx.create({
    type: 'paragraph',
    id: 'p4',
    content: '0123456789'
  });
  tx.create({
    type: 'test-container-anno',
    id: 'a1',
    start: {
      path: ['p1', 'content'],
      offset: 5,
    },
    end: {
      path: ['p3', 'content'],
      offset: 4
    },
    containerId: 'body'
  });
  tx.create({
    type: 'strong',
    id: 'a2',
    start: {
      path: ['p1', 'content'],
      offset: 0,
    },
    end: {
      offset: 2
    }
  });
  body.show('p1');
  body.show('p2');
  body.show('p3');
  body.show('p4');
  return tx
};

const test = substanceTest.module('AnnotationCommand');

test("can 'create' property annotation", function(t) {
  let { doc } = fixture(t);
  let selectionState = new substance.SelectionState(doc);
  let cmd = new ToggleStrongCommand();
  let sel = doc.createSelection({
    type: 'property',
    path: ['p4', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  selectionState.setSelection(sel);
  let cmdState = cmd.getCommandState({
    selectionState: selectionState
  });
  t.equal(cmdState.mode, 'create', "Mode should be correct.");
  t.end();
});

test("execute 'create' property annotation", function(t) {
  let { editorSession, doc } = fixture(t);
  let cmd = new ToggleStrongCommand();
  editorSession.setSelection({
    type: 'property',
    path: ['p4', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  let res = cmd.execute({
    commandState: {
      mode: 'create'
    },
    editorSession: editorSession,
    selectionState: editorSession.getSelectionState()
  });
  let newAnno = res.anno;
  t.notNil(newAnno, 'A new anno should have been created');
  newAnno = doc.get(newAnno.id);
  t.equal(newAnno.type, 'strong', '.. of correct type');
  t.deepEqual(newAnno.start.path, ['p4', 'content'], ".. with correct path");
  t.equal(newAnno.start.offset, 1, '.. with correct startOffset');
  t.equal(newAnno.end.offset, 6, '.. with correct endOffset');
  t.end();
});

test("can 'delete' property annotation", function(t) {
  let { doc } = fixture(t);
  let selectionState = new substance.SelectionState(doc);
  let cmd = new ToggleStrongCommand();
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 5,
    endOffset: 7
  });
  selectionState.setSelection(sel);
  let cmdState = cmd.getCommandState({
    selectionState: selectionState
  });
  t.equal(cmdState.mode, 'delete', "Mode should be correct.");
  t.end();
});

test("execute 'delete' property annotation", function(t) {
  let { doc } = fixture(t);
  let selectionState = new substance.SelectionState(doc);
  let cmd = new ToggleStrongCommand();
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 5,
    endOffset: 7
  });
  selectionState.setSelection(sel);
  let cmdState = cmd.getCommandState({
    selectionState: selectionState
  });
  t.equal(cmdState.mode, 'delete', "Mode should be correct.");
  t.end();
});

test("can 'expand' property annotation", function(t) {
  let { editorSession, doc } = fixture(t);
  let cmd = new ToggleStrongCommand();
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 5,
    endOffset: 7
  });
  cmd.execute({
    commandState: {
      mode: 'delete'
    },
    editorSession: editorSession,
    selectionState: editorSession.getSelectionState()
  });
  t.isNil(doc.get('a3'), 'a3 should be gone.');
  t.end();
});

class ToggleStrongCommand extends substance.AnnotationCommand {
  constructor() {
    super({ name: 'strong', nodeType: 'strong' });
  }
}

function fixture(t) {
  return setupEditor(t, containerAnnoSample, (tx) => {
    tx.create({
      id: 'a3',
      type: 'strong',
      start: { path: ['p1', 'content'], offset: 4 },
      end: { offset: 8 }
    });
  })
}

function createTestArticle(seedFn) {
  let config = getTestConfig();
  var doc = new TestArticle(config.getSchema());
  if (seedFn) {
    seedFn(doc);
  }
  return doc
}

/*
  Creates a document with the following content

  ```
  body:
    p1: '0123456789'
    p2: '0123456789'
    p3: '0123456789'
    p4: '0123456789'
  ```
*/
function simple(doc) {
  let body = doc.get('body');
  doc.create({
    type: 'paragraph',
    id: 'p1',
    content: '0123456789'
  });
  body.show('p1');
  doc.create({
    type: 'paragraph',
    id: 'p2',
    content: '0123456789'
  });
  body.show('p2');
  doc.create({
    type: 'paragraph',
    id: 'p3',
    content: '0123456789'
  });
  body.show('p3');
  doc.create({
    type: 'paragraph',
    id: 'p4',
    content: '0123456789'
  });
  body.show('p4');
}

const test$1 = substanceTest.module('annotationHelpers');

let truncateAnnotation = substance.annotationHelpers.truncateAnnotation;
let expandAnnotation = substance.annotationHelpers.expandAnnotation;
let fuseAnnotation = substance.annotationHelpers.fuseAnnotation;

test$1("Truncate property annotation with a given property selection", function(t) {
  let doc = fixture$1(A1);
  // Put cursor inside an the existing annotation
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 2
  });
  let anno = doc.get('a1');
  truncateAnnotation(doc, anno, sel);
  t.equal(anno.start.offset, 0, 'startOffset should be 0');
  t.equal(anno.end.offset, 1, 'endOffset should have changed from 2 to 1');
  t.end();
});

test$1("Truncate container annotation with a given property selection", function(t) {
  let doc = fixture$1(CA1);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p3', 'content'],
    startOffset: 1,
    endOffset: 4
  });
  let anno = doc.get('ca1');
  truncateAnnotation(doc, anno, sel);
  t.equal(anno.end.offset, 1, 'endOffset should be 1');
  t.end();
});

test$1("Truncate container annotation with a given container selection", function(t) {
  let doc = fixture$1(CA1);
  let sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  let anno = doc.get('ca1');
  truncateAnnotation(doc, anno, sel);
  t.deepEqual(anno.end.path, ['p2', 'content'], 'endPath should be p2.content');
  t.equal(anno.end.offset, 1, 'endOffset should be 1');
  t.end();
});

test$1("Expand-right of property annotation for a given property selection", function(t) {
  let doc = fixture$1(A1);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  let anno = doc.get('a1');
  expandAnnotation(doc, anno, sel);
  t.equal(anno.start.offset, 0, 'startOffset should be 0');
  t.equal(anno.end.offset, 6, 'endOffset should have changed from 2 to 1');
  t.end();
});

test$1("Expand container annotation for a given property selection (right expansion)", function(t) {
  let doc = fixture$1(CA1);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p3', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  let anno = doc.get('ca1');
  expandAnnotation(doc, anno, sel);
  t.equal(anno.end.offset, 6, 'endOffset should be 6');
  t.end();
});

test$1("Expand container annotation for a given container selection (expand right)", function(t) {
  let doc = fixture$1(CA1);
  let sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 6,
  });
  let anno = doc.get('ca1');
  expandAnnotation(doc, anno, sel);
  t.deepEqual(anno.end.path, ['p3', 'content'], 'endPath should be p2.content');
  t.equal(anno.end.offset, 6, 'endOffset should be 6');
  t.end();
});

test$1("Fuse two property annotations for a given property selection", function(t) {
  let tx = new substance.EditingInterface(fixture$1(A1, A2));
  // Put selection so that it touches both strong annotations
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  let a1 = tx.get('a1');
  let a2 = tx.get('a2');
  fuseAnnotation(tx, [a1, a2]);
  t.isNil(tx.get('a2'), 'a2 should be gone.');
  t.equal(a1.start.offset, 0, 'startOffset should be 0');
  t.equal(a1.end.offset, 6, 'endOffset should be 6');
  t.end();
});

test$1("Fuse two conatiner annotations for a given property selection", function(t) {
  let tx = new substance.EditingInterface(fixture$1(CA1, CA2));
  tx.setSelection({
    type: 'property',
    path: ['p3', 'content'],
    startOffset: 3,
    endOffset: 8
  });
  let ca1 = tx.get('ca1');
  let ca2 = tx.get('ca2');
  fuseAnnotation(tx, [ca1, ca2]);
  t.isNil(tx.get('ca2'), 'ca2 should be gone.');
  t.deepEqual(ca1.start.path, ['p1', 'content'], 'start pPath should be p1.content');
  t.equal(ca1.start.offset, 5, 'start offset should be 5');
  t.deepEqual(ca1.end.path, ['p4', 'content'], 'end path should be p1.content');
  t.equal(ca1.end.offset, 9, 'end offset should be 9');
  t.end();
});

function fixture$1(...fns) {
  let doc = createTestArticle(simple);
  fns.forEach((fn)=>{
    fn(doc);
  });
  return doc
}

function A1(doc) {
  doc.create({
    type: 'strong',
    id: 'a1',
    start: {
      path: ['p1', 'content'],
      offset: 0,
    },
    end: {
      offset: 2
    }
  });
}

function A2(doc) {
  doc.create({
    id: 'a2',
    type: 'strong',
    start: {
      path: ['p1', 'content'],
      offset: 4
    },
    end: {
      offset: 6
    }
  });
}

function CA1(doc) {
  doc.create({
    type: 'test-container-anno',
    id: 'ca1',
    start: {
      path: ['p1', 'content'],
      offset: 5,
    },
    end: {
      path: ['p3', 'content'],
      offset: 4
    },
    containerId: 'body'
  });
}

function CA2(doc) {
  doc.create({
    type: 'test-container-anno',
    id: 'ca2',
    containerId: 'body',
    start: {
      path: ['p3', 'content'],
      offset: 7,
    },
    end: {
      path: ['p4', 'content'],
      offset: 9,
    }
  });
}

const test$2 = substanceTest.module('ArrayOperation');

function checkArrayOperationTransform(t, a, b, input, expected) {
  let ops = substance.ArrayOperation.transform(a, b);
  let output = ops[1].apply(a.apply(input.slice(0)));
  t.deepEqual(output, expected, `(b' o a)('${JSON.stringify(input)}') == '${JSON.stringify(expected)}' with a=${a.toString()}, b'=${ops[1].toString()}`);
  output = ops[0].apply(b.apply(input.slice(0)));
  t.deepEqual(output, expected, `(a' o b)('${JSON.stringify(input)}') == '${JSON.stringify(expected)}' with b=${b.toString()}, a'=${ops[0].toString()}`);
}

test$2("Insert element", (t) => {
  let arr = [1,2,4];
  let expected = [1,2,3,4];
  let op = substance.ArrayOperation.Insert(2, 3);
  op.apply(arr);
  t.deepEqual(arr, expected, 'Should insert element.');
  t.end();
});

test$2("Insert element after last position", (t) => {
  let arr = [1,2,3];
  let expected = [1,2,3,4];
  let op = substance.ArrayOperation.Insert(arr.length, 4);
  op.apply(arr);
  t.deepEqual(arr, expected, 'Should append element.');
  t.end();
});

test$2("Delete element", (t) => {
  let arr = [1,2,3];
  let expected = [1,3];
  let op = substance.ArrayOperation.Delete(1, 2);
  op.apply(arr);
  t.deepEqual(arr, expected, 'Should delete element.');
  t.end();
});

test$2("Create operation with invalid data", (t) => {
  t.throws(function() {
    new substance.ArrayOperation();
  }, "Should throw if no data given.");
  t.throws(function() {
    new substance.ArrayOperation({ type: 'foo' });
  }, "Should throw for invalid type.");
  t.throws(function() {
    new substance.ArrayOperation({ type: substance.ArrayOperation.INSERT, value: 1});
  }, "Should throw for missing position.");
  t.throws(function() {
    new substance.ArrayOperation({ type: substance.ArrayOperation.INSERT, pos: -1, value: 1});
  }, "Should throw for position < 0.");
  t.end();
});

test$2("Operation can be NOP", (t) => {
  let op = new substance.ArrayOperation({type: substance.ArrayOperation.NOP});
  t.ok(op.isNOP(), 'Operation should be NOP');
  t.end();
});

test$2("Apply operation on too short array.", (t) => {
  let arr = [1,2,3];
  let op = substance.ArrayOperation.Insert(4, 5);
  t.throws(function() {
    op.apply(arr);
  }, "Should throw if applying insert operation out-of-bounds");
  op = substance.ArrayOperation.Delete(4, 5);
  t.throws(function() {
    op.apply(arr);
  }, "Should throw if applying delete operation out-of-bounds");
  t.end();
});

// Note: it is better to fail in such cases, as this is an indicator for other greater problems.
test$2("Apply delete operation on wrong array.", (t) => {
  let arr = [1,2,3];
  let op = substance.ArrayOperation.Delete(2, 4);
  t.throws(function() {
    op.apply(arr);
  }, "Should throw if applying delete operation with wrong value");
  t.end();
});

test$2("JSON de-/serialisation", (t) => {
  let op = substance.ArrayOperation.Delete(1,2);
  let out = op.toJSON();
  t.equal(out.type, substance.ArrayOperation.DELETE);
  t.equal(out.pos, 1);
  t.equal(out.val, 2);
  op = substance.ArrayOperation.fromJSON(out);
  t.ok(op.isDelete());
  t.equal(op.getOffset(), 1);
  t.equal(op.getValue(), 2);
  op = new substance.ArrayOperation({type: substance.ArrayOperation.NOP});
  out = op.toJSON();
  t.deepEqual(out, {type: substance.ArrayOperation.NOP});
  t.end();
});

// Insert-Insert Transformations
// --------
// Cases:
//  1. `a < b`:   operations should not be affected
//  2. `b < a`:   dito
//  3. `a == b`:  result depends on preference (first applied)

test$2("Transformation: a=Insert, b=Insert, a < b and b < a", (t) => {
  let input = [1,3,5];
  let expected = [1,2,3,4,5];
  let a = substance.ArrayOperation.Insert(1, 2);
  let b = substance.ArrayOperation.Insert(2, 4);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});

// Example:
//     A = [1,4], a = [+, 1, 2], b = [+, 1, 3]
//     A  - a ->  [1, 2, 4]   - b' ->   [1,2,3,4]     => b'= [+, 2, 3], transform(a, b) = [a, b']
//     A  - b ->  [1, 3, 4]   - a' ->   [1,3,2,4]     => a'= [+, 2, 2], transform(b, a) = [a', b]
test$2("Transformation: a=Insert, b=Insert, a == b", (t) => {
  let input = [1,4];
  let expected = [1,2,3,4];
  let expected_2 = [1,3,2,4];
  let a = substance.ArrayOperation.Insert(1, 2);
  let b = substance.ArrayOperation.Insert(1, 3);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected_2);
  t.end();
});

// Delete-Delete Transformations
// --------
// Cases:
//  1. `a < b`:   operations should not be affected
//  2. `b < a`:   dito
//  3. `a == b`:  second operation should not have an effect
//                user should be noticed about conflict

test$2("Transformation: a=Delete, b=Delete (1,2), a < b and b < a", (t) => {
  let input = [1,2,3,4,5];
  let expected = [1,3,5];
  let a = substance.ArrayOperation.Delete(1, 2);
  let b = substance.ArrayOperation.Delete(3, 4);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});

test$2("Transformation: a=Delete, b=Delete (3), a == b", (t) => {
  let input = [1,2,3];
  let expected = [1,3];
  let a = substance.ArrayOperation.Delete(1, 2);
  let b = substance.ArrayOperation.Delete(1, 2);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});

// Insert-Delete Transformations
// --------
// Cases: (a = insertion, b = deletion)
//  1. `a < b`:   b must be shifted right
//  2. `b < a`:   a must be shifted left
//  3. `a == b`:  ???


// A = [1,3,4,5], a = [+, 1, 2], b = [-, 2, 4]
// A  - a ->  [1,2,3,4,5] - b' ->   [1,2,3,5]     => b'= [-, 3, 4]
// A  - b ->  [1,3,5]     - a' ->   [1,2,3,5]     => a'= [+, 1, 2] = a
test$2("Transformation: a=Insert, b=Delete (1), a < b", (t) => {
  let input = [1,3,4,5];
  let expected = [1,2,3,5];
  let a = substance.ArrayOperation.Insert(1, 2);
  let b = substance.ArrayOperation.Delete(2, 4);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});

// A = [1,2,3,5], a = [+,3,4], b = [-,1,2]
// A  - a ->  [1,2,3,4,5] - b' ->   [1,3,4,5]     => b'= [-,1,2] = b
// A  - b ->  [1,3,5]     - a' ->   [1,3,4,5]     => a'= [+,2,4]
test$2("Transformation: a=Insert, b=Delete (2), b < a", (t) => {
  let input = [1,2,3,5];
  let expected = [1,3,4,5];
  let a = substance.ArrayOperation.Insert(3, 4);
  let b = substance.ArrayOperation.Delete(1, 2);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});


// A = [1,2,3], a = [+,1,4], b = [-,1,2]
// A  - a ->  [1,4,2,3] - b' ->   [1,4,3]     => b'= [-,2,2]
// A  - b ->  [1,3]     - a' ->   [1,4,3]     => a'= [+,1,4] = a
test$2("Transformation: a=Insert, b=Delete (3), a == b", (t) => {
  let input = [1,2,3];
  let expected = [1,4,3];
  let a = substance.ArrayOperation.Insert(1, 4);
  let b = substance.ArrayOperation.Delete(1, 2);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});

test$2("Transformation: a=NOP || b=NOP", (t) => {
  let a = new substance.ArrayOperation.Insert(1, 4);
  let b = new substance.ArrayOperation({type: substance.ArrayOperation.NOP});
  let tr = substance.ArrayOperation.transform(a, b);
  t.deepEqual(tr[0].toJSON(), a.toJSON());
  t.deepEqual(tr[1].toJSON(), b.toJSON());
  tr = substance.ArrayOperation.transform(b, a);
  t.deepEqual(tr[0].toJSON(), b.toJSON());
  t.deepEqual(tr[1].toJSON(), a.toJSON());
  t.end();
});

test$2("Inverting operations", (t) => {
  let op = substance.ArrayOperation.Insert(1, 4);
  let inverse = op.invert();
  t.ok(inverse.isDelete(), 'Inverse of an insert op should be a delete op.');
  t.equal(inverse.getOffset(), op.getOffset(), 'Offset of inverted op should be the same.');
  t.equal(inverse.getValue(), op.getValue(), 'Value of inverted op should be the same.');
  op = substance.ArrayOperation.Delete(2, 3);
  inverse = op.invert();
  t.ok(inverse.isInsert(), 'Inverse of a delete op should be an insert op.');
  t.equal(inverse.getOffset(), op.getOffset(), 'Offset of inverted op should be the same.');
  t.equal(inverse.getValue(), op.getValue(), 'Value of inverted op should be the same.');
  op = new substance.ArrayOperation({type: substance.ArrayOperation.NOP});
  inverse = op.invert();
  t.ok(inverse.isNOP(), 'Inverse of a nop is a nop.');
  t.end();
});

test$2("Transformations can be done inplace (optimzation for internal use)", (t) => {
  let a = substance.ArrayOperation.Insert(2, 3);
  let b = substance.ArrayOperation.Insert(2, 3);
  let tr = substance.ArrayOperation.transform(a, b, {inplace: true});
  t.ok(a.getOffset() === tr[0].getOffset() && b.getOffset() === tr[1].getOffset(), "Transformation should be done inplace.");
  t.end();
});

test$2("With option 'no-conflict' conflicting operations can not be transformed.", (t) => {
  let a = substance.ArrayOperation.Insert(2, 2);
  let b = substance.ArrayOperation.Insert(2, 2);
  t.throws(function() {
    substance.ArrayOperation.transform(a, b, { "no-conflict": true });
  }, 'Transforming conflicting ops should throw when option "no-conflict" is enabled.');
  t.end();
});

test$2("Conflicts: inserting at the same position", (t) => {
  // this is considered a conflict as a decision is needed to determine which element comes first.
  let a = substance.ArrayOperation.Insert(2, 'a');
  let b = substance.ArrayOperation.Insert(2, 'b');
  t.ok(a.hasConflict(b) && b.hasConflict(a), "Inserts at the same position are considered a conflict.");
  t.end();
});

test$2("Conflicts: inserting at different positions", (t) => {
  let a = substance.ArrayOperation.Insert(2, 'a');
  let b = substance.ArrayOperation.Insert(4, 'b');
  t.ok(!a.hasConflict(b) && !b.hasConflict(a), "Inserts at different positions should be fine.");
  t.end();
});

test$2("Conflicts: deleting at the same position", (t) => {
  // this is *not* considered a conflict as it is clear how the result should look like.
  let a = substance.ArrayOperation.Delete(2, 'a');
  let b = substance.ArrayOperation.Delete(2, 'a');
  t.ok(!a.hasConflict(b) && !b.hasConflict(a), "Deletes at the same position are not a conflict.");
  t.end();
});

test$2("Conflicts: inserting and deleting at the same position", (t) => {
  // this is *not* considered a conflict as it is clear how the result should look like.
  let a = substance.ArrayOperation.Insert(2, 'a');
  let b = substance.ArrayOperation.Delete(2, 'b');
  t.ok(!a.hasConflict(b) && !b.hasConflict(a), "Inserting and deleting at the same position is not a conflict.");
  t.end();
});

test$2("Conflicts: when NOP involved", (t) => {
  let a = substance.ArrayOperation.Insert(2, 2);
  let b = new substance.ArrayOperation({type: substance.ArrayOperation.NOP});
  t.ok(!a.hasConflict(b) && !b.hasConflict(a), "NOPs should never conflict.");
  t.end();
});

var BrowserLinuxPLainTextFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 22.3999996185303px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">XXX</span>";

var BrowserLinuxAnnotatedTextFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 22.3999996185303px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">X</span><a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" title=\"United States Senate election in Illinois, 2004\" style=\"text-decoration: none; color: rgb(11, 0, 128); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 22.3999996185303px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; background: none rgb(255, 255, 255);\">X</a><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 22.3999996185303px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">X</span>";

var BrowserLinuxTwoParagraphsFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><p>AAA</p><p>BBB</p>\r\n";

var BrowserWindowsPlainTextFixture = "<html>\r\n<body>\r\n<!--StartFragment--><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 12.8px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">XXX</span><!--EndFragment-->\r\n</body>\r\n</html>\r\n";

var BrowserWindowsAnnotatedTextFixture = "<html>\r\n<body>\r\n<!--StartFragment--><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 12.8px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">X</span></span><a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" title=\"United States Senate election in Illinois, 2004\" style=\"text-decoration: none; color: rgb(11, 0, 128); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 12.8px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; background: none rgb(255, 255, 255);\">X</a><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 12.8px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">X</span></span><!--EndFragment-->\r\n</body>\r\n</html>";

var BrowserWindowsTwoParagraphsFixture = "<html>\r\n<body>\r\n<!--StartFragment--><p>AAA</p><p>BBB</p><!--EndFragment-->\r\n</body>\r\n</html>\r\n";

var BrowserLinuxFirefoxPlainTextFixture = "XXX";

var BrowserLinuxFirefoxAnnotatedTextFixture = "X<a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" title=\"United States Senate election in Illinois, 2004\">X</a>X";

var BrowserLinuxFirefoxTwoParagraphsFixture = "<p>AAA</p><p>BBB</p>";

var BrowserLinuxFirefoxWholePageFixture = "    <div id=\"mw-page-base\" class=\"noprint\"></div>\r\n    <div id=\"mw-head-base\" class=\"noprint\"></div>\r\n    <div id=\"content\" class=\"mw-body\" role=\"main\">\r\n      <a id=\"top\"></a>\r\n\r\n              <div id=\"siteNotice\"><div id=\"centralNotice\"></div></div>\r\n            <div class=\"mw-indicators\">\r\n</div>\r\n      <h1 id=\"firstHeading\" class=\"firstHeading\" lang=\"en\">Synonym</h1>\r\n                  <div id=\"bodyContent\" class=\"mw-body-content\">\r\n                  <div id=\"siteSub\">From Wikipedia, the free encyclopedia</div>\r\n                <div id=\"contentSub\"></div>\r\n                        </div></div>\r\n        <div id=\"mw-content-text\" dir=\"ltr\" class=\"mw-content-ltr\" lang=\"en\"><div class=\"hatnote\">This article is about the general meaning of \"synonym\".  For its use in biology, see <a href=\"https://en.wikipedia.org/wiki/Synonym_%28taxonomy%29\" title=\"Synonym (taxonomy)\">Synonym (taxonomy)</a>.</div>\r\n<table class=\"metadata plainlinks ambox ambox-content ambox-Refimprove\" role=\"presentation\">\r\n<tbody><tr>\r\n<td class=\"mbox-image\">\r\n<div style=\"width:52px\"><a href=\"https://en.wikipedia.org/wiki/File:Question_book-new.svg\" class=\"image\"><img alt=\"\" src=\"https://upload.wikimedia.org/wikipedia/en/thumb/9/99/Question_book-new.svg/50px-Question_book-new.svg.png\" srcset=\"//upload.wikimedia.org/wikipedia/en/thumb/9/99/Question_book-new.svg/75px-Question_book-new.svg.png 1.5x, //upload.wikimedia.org/wikipedia/en/thumb/9/99/Question_book-new.svg/100px-Question_book-new.svg.png 2x\" data-file-width=\"262\" data-file-height=\"204\" height=\"39\" width=\"50\"></a></div>\r\n</td>\r\n<td class=\"mbox-text\"><span class=\"mbox-text-span\">This article <b>needs additional citations for <a href=\"https://en.wikipedia.org/wiki/Wikipedia:Verifiability\" title=\"Wikipedia:Verifiability\">verification</a></b>. <span class=\"hide-when-compact\">Please help <a class=\"external text\" href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;action=edit\">improve this article</a> by <a href=\"https://en.wikipedia.org/wiki/Help:Introduction_to_referencing/1\" title=\"Help:Introduction to referencing/1\" class=\"mw-redirect\">adding citations to reliable sources</a>. Unsourced material may be challenged and removed.</span> <small><i>(May 2014)</i></small></span></td>\r\n</tr>\r\n</tbody></table>\r\n<div class=\"thumb tright\">\r\n<div class=\"thumbinner\" style=\"width:222px;\"><a href=\"https://en.wikipedia.org/wiki/File:Library_of_Ashurbanipal_synonym_list_tablet.jpg\" class=\"image\"><img alt=\"\" src=\"https://upload.wikimedia.org/wikipedia/commons/thumb/6/64/Library_of_Ashurb…onym_list_tablet.jpg/220px-Library_of_Ashurbanipal_synonym_list_tablet.jpg\" class=\"thumbimage\" srcset=\"//upload.wikimedia.org/wikipedia/commons/thumb/6/64/Library_of_Ashurbanipal_synonym_list_tablet.jpg/330px-Library_of_Ashurbanipal_synonym_list_tablet.jpg 1.5x, //upload.wikimedia.org/wikipedia/commons/thumb/6/64/Library_of_Ashurbanipal_synonym_list_tablet.jpg/440px-Library_of_Ashurbanipal_synonym_list_tablet.jpg 2x\" data-file-width=\"1829\" data-file-height=\"2676\" height=\"322\" width=\"220\"></a>\r\n<div class=\"thumbcaption\">\r\n<div class=\"magnify\"></div></div></div></div></div><div class=\"thumb tright\"><div class=\"thumbinner\" style=\"width:222px;\"><div class=\"thumbcaption\">\r\nSynonym list in <a href=\"https://en.wikipedia.org/wiki/Cuneiform\" title=\"Cuneiform\">cuneiform</a> on a clay tablet, <a href=\"https://en.wikipedia.org/wiki/Neo-Assyrian\" title=\"Neo-Assyrian\" class=\"mw-redirect\">Neo-Assyrian</a> period. Ref:<a rel=\"nofollow\" class=\"external text\" href=\"http://www.britishmuseum.org/research/search_the_collection_database/search_object_details.aspx?objectid=308401&amp;partid=1\">K.4375</a> .</div>\r\n</div>\r\n</div>\r\n<p>A <b>synonym</b> is a word or phrase that means exactly or nearly the\r\n same as another word or phrase in the same language. Words that are\r\nsynonyms are said to be <b>synonymous</b>, and the state of being a synonym is called <b>synonymy</b>. The word comes from <a href=\"https://en.wikipedia.org/wiki/Ancient_Greek_language\" title=\"Ancient Greek language\" class=\"mw-redirect\">Ancient Greek</a> <i>syn</i> (<span xml:lang=\"grc\" lang=\"grc\"><a href=\"https://en.wiktionary.org/wiki/%CF%83%CF%8D%CE%BD\" class=\"extiw\" title=\"wikt:σύν\">σύν</a></span>) (\"with\") and <i>onoma</i> (<span xml:lang=\"grc\" lang=\"grc\"><a href=\"https://en.wiktionary.org/wiki/%E1%BD%84%CE%BD%CE%BF%CE%BC%CE%B1\" class=\"extiw\" title=\"wikt:ὄνομα\">ὄνομα</a></span>) (\"name\"). An example of synonyms are the words <i>begin</i>, <i>start</i>, and <i>commence</i>. Words can be synonymous when meant in certain <a href=\"https://en.wikipedia.org/wiki/Word_sense\" title=\"Word sense\">senses</a>, even if they are not synonymous in all of their senses. For example, if we talk about a <i>long time</i> or an <i>extended time</i>, <i>long</i> and <i>extended</i> are synonymous within that <a href=\"https://en.wikipedia.org/wiki/Context_%28language_use%29\" title=\"Context (language use)\">context</a>. Synonyms with exact interchangeability share a <a href=\"https://en.wikipedia.org/wiki/Seme_%28semantics%29\" title=\"Seme (semantics)\">seme</a> or denotational <a href=\"https://en.wikipedia.org/wiki/Sememe\" title=\"Sememe\">sememe</a>, whereas those with inexactly similar meanings share a broader denotational or connotational <a href=\"https://en.wikipedia.org/wiki/Sememe\" title=\"Sememe\">sememe</a> and thus overlap within a <a href=\"https://en.wikipedia.org/wiki/Semantic_field\" title=\"Semantic field\">semantic field</a>. Some academics call the former type <a href=\"https://en.wikipedia.org/wiki/Cognitive_synonymy\" title=\"Cognitive synonymy\">cognitive synonyms</a> to distinguish them from the latter type, which they call near-synonyms.<sup id=\"cite_ref-Stanojevi.C4.87_2009_1-0\" class=\"reference\"><a href=\"https://en.wikipedia.org/wiki/Synonym#cite_note-Stanojevi.C4.87_2009-1\"><span>[</span>1<span>]</span></a></sup></p>\r\n<p>In the figurative sense, two words are sometimes said to be synonymous if they have the same <a href=\"https://en.wikipedia.org/wiki/Connotation\" title=\"Connotation\">connotation</a>:</p>\r\n<blockquote class=\"templatequote\">\r\n<p>...a widespread impression that ... <a href=\"https://en.wikipedia.org/wiki/Hollywood\" title=\"Hollywood\">Hollywood</a> was synonymous with immorality...<sup id=\"cite_ref-2\" class=\"reference\"><a href=\"https://en.wikipedia.org/wiki/Synonym#cite_note-2\"><span>[</span>2<span>]</span></a></sup></p>\r\n<div class=\"templatequotecite\"><cite>— <a href=\"https://en.wikipedia.org/wiki/Doris_Kearns_Goodwin\" title=\"Doris Kearns Goodwin\">Doris Kearns Goodwin</a></cite></div>\r\n</blockquote>\r\n<p><a href=\"https://en.wikipedia.org/wiki/Metonymy\" title=\"Metonymy\">Metonymy</a> can sometimes be a form of synonymy, as when, for example, <i>the <a href=\"https://en.wikipedia.org/wiki/White_House\" title=\"White House\">White House</a></i> is used as a synonym of <i>the administration</i> in referring to the U.S. <a href=\"https://en.wikipedia.org/wiki/Executive_%28government%29\" title=\"Executive (government)\">executive branch</a> under a specific president. Thus a metonym is a type of synonym, and the word <i>metonym</i> is a <a href=\"https://en.wikipedia.org/wiki/Hyponymy_and_hypernymy\" title=\"Hyponymy and hypernymy\">hyponym</a> of the word <i>synonym</i>.</p>\r\n<p>The analysis of synonymy, <a href=\"https://en.wikipedia.org/wiki/Polysemy\" title=\"Polysemy\">polysemy</a>, and <a href=\"https://en.wikipedia.org/wiki/Hyponymy_and_hypernymy\" title=\"Hyponymy and hypernymy\">hyponymy and hypernymy</a> is vital to <a href=\"https://en.wikipedia.org/wiki/Taxonomy_%28general%29\" title=\"Taxonomy (general)\">taxonomy</a> and <a href=\"https://en.wikipedia.org/wiki/Ontology_%28information_science%29\" title=\"Ontology (information science)\">ontology</a> in the <a href=\"https://en.wikipedia.org/wiki/Information_science\" title=\"Information science\">information-science</a> senses of those terms. It has applications in <a href=\"https://en.wikipedia.org/wiki/Pedagogy\" title=\"Pedagogy\">pedagogy</a> and <a href=\"https://en.wikipedia.org/wiki/Machine_learning\" title=\"Machine learning\">machine learning</a>, because they rely on <a href=\"https://en.wikipedia.org/wiki/Word-sense_disambiguation\" title=\"Word-sense disambiguation\">word-sense disambiguation</a> and <a href=\"https://en.wiktionary.org/wiki/schema#Noun\" class=\"extiw\" title=\"wikt:schema\">schema</a>.</p>\r\n<p></p>\r\n<div id=\"toc\" class=\"toc\">\r\n<div id=\"toctitle\">\r\n<h2>Contents</h2>\r\n</div></div><div id=\"toc\" class=\"toc\">\r\n<ul>\r\n<li class=\"toclevel-1 tocsection-1\"><a href=\"https://en.wikipedia.org/wiki/Synonym#Examples\"><span class=\"tocnumber\">1</span> <span class=\"toctext\">Examples</span></a></li>\r\n<li class=\"toclevel-1 tocsection-2\"><a href=\"https://en.wikipedia.org/wiki/Synonym#Related_terms\"><span class=\"tocnumber\">2</span> <span class=\"toctext\">Related terms</span></a></li>\r\n<li class=\"toclevel-1 tocsection-3\"><a href=\"https://en.wikipedia.org/wiki/Synonym#See_also\"><span class=\"tocnumber\">3</span> <span class=\"toctext\">See also</span></a></li>\r\n<li class=\"toclevel-1 tocsection-4\"><a href=\"https://en.wikipedia.org/wiki/Synonym#References\"><span class=\"tocnumber\">4</span> <span class=\"toctext\">References</span></a></li>\r\n<li class=\"toclevel-1 tocsection-5\"><a href=\"https://en.wikipedia.org/wiki/Synonym#External_links\"><span class=\"tocnumber\">5</span> <span class=\"toctext\">External links</span></a></li>\r\n</ul>\r\n</div>\r\n<p></p>\r\n<h2><span class=\"mw-headline\" id=\"Examples\">Examples</span></h2>\r\n<p>Synonyms can be any <a href=\"https://en.wikipedia.org/wiki/Part_of_speech\" title=\"Part of speech\">part of speech</a> (such as <a href=\"https://en.wikipedia.org/wiki/Noun\" title=\"Noun\">nouns</a>, <a href=\"https://en.wikipedia.org/wiki/Verbs\" title=\"Verbs\" class=\"mw-redirect\">verbs</a>, <a href=\"https://en.wikipedia.org/wiki/Adjective\" title=\"Adjective\">adjectives</a>, <a href=\"https://en.wikipedia.org/wiki/Adverb\" title=\"Adverb\">adverbs</a> or <a href=\"https://en.wikipedia.org/wiki/Preposition\" title=\"Preposition\" class=\"mw-redirect\">prepositions</a>), as long as both words belong to the same part of speech. Examples:</p>\r\n<ul>\r\n<li>verb\r\n<ul>\r\n<li><i>buy</i> and <i>purchase</i></li>\r\n</ul>\r\n</li>\r\n<li>adjective\r\n<ul>\r\n<li><i>big</i> and <i>large</i></li>\r\n</ul>\r\n</li>\r\n<li>adverb\r\n<ul>\r\n<li><i>quickly</i> and <i>speedily</i></li>\r\n</ul>\r\n</li>\r\n<li>preposition\r\n<ul>\r\n<li><i>on</i> and <i>upon</i></li>\r\n</ul>\r\n</li>\r\n</ul>\r\n<p>Note that synonyms are defined with respect to certain senses of words; for instance, <i>pupil</i> as the <i>aperture in the iris of the eye</i> is not synonymous with <i>student</i>. Such like, <i>he expired</i> means the same as <i>he died</i>, yet <i>my passport has expired</i> cannot be replaced by <i>my passport has died</i>.</p>\r\n<p>In English, many synonyms emerged in the <a href=\"https://en.wikipedia.org/wiki/Middle_Ages\" title=\"Middle Ages\">Middle Ages</a>, after the <a href=\"https://en.wikipedia.org/wiki/Norman_conquest_of_England\" title=\"Norman conquest of England\">Norman conquest of England</a>. While <a href=\"https://en.wikipedia.org/wiki/England\" title=\"England\">England</a>'s new ruling class spoke <a href=\"https://en.wikipedia.org/wiki/Norman_French\" title=\"Norman French\" class=\"mw-redirect\">Norman French</a>, the lower classes continued to speak <a href=\"https://en.wikipedia.org/wiki/Old_English\" title=\"Old English\">Old English</a> (Anglo-Saxon). Thus, today we have synonyms like the Norman-derived <i>people</i>, <i>liberty</i> and <i>archer</i>, and the Saxon-derived <i>folk</i>, <i>freedom</i> and <i>bowman</i>. For more examples, see the <a href=\"https://en.wikipedia.org/wiki/List_of_Germanic_and_Latinate_equivalents_in_English\" title=\"List of Germanic and Latinate equivalents in English\">list of Germanic and Latinate equivalents in English</a>.</p>\r\n<p>Some <a href=\"https://en.wikipedia.org/wiki/Lexicographer\" title=\"Lexicographer\" class=\"mw-redirect\">lexicographers</a> claim that no synonyms have exactly the same meaning (in all contexts or social levels of language) because <a href=\"https://en.wikipedia.org/wiki/Etymology\" title=\"Etymology\">etymology</a>, <a href=\"https://en.wikipedia.org/wiki/Orthography\" title=\"Orthography\">orthography</a>,\r\n phonic qualities, ambiguous meanings, usage, etc. make them unique.\r\nDifferent words that are similar in meaning usually differ for a reason:\r\n <i>feline</i> is more formal than <i>cat</i>; <i>long</i> and <i>extended</i> are only synonyms in one usage and not in others (for example, a <i>long arm</i> is not the same as an <i>extended arm</i>). Synonyms are also a source of <a href=\"https://en.wikipedia.org/wiki/Euphemism\" title=\"Euphemism\">euphemisms</a>.</p>\r\n<p>The purpose of a <a href=\"https://en.wikipedia.org/wiki/Thesaurus\" title=\"Thesaurus\">thesaurus</a> is to offer the user a listing of similar or related words; these are often, but not always, synonyms.</p>\r\n<h2><span class=\"mw-headline\" id=\"Related_terms\">Related terms</span></h2>\r\n<ul>\r\n<li>The word <i><b>poecilonym</b></i> is a rare synonym of the word <i>synonym</i>. It is not entered in most major dictionaries and is a curiosity or piece of trivia for being an <a href=\"https://en.wikipedia.org/wiki/Autological_word\" title=\"Autological word\">autological word</a> because of its <a href=\"https://en.wikipedia.org/wiki/Meta\" title=\"Meta\">meta</a> quality as a synonym of <i>synonym</i>.</li>\r\n<li><b><a href=\"https://en.wikipedia.org/wiki/Antonym\" title=\"Antonym\" class=\"mw-redirect\">Antonyms</a></b> are words with opposite or nearly opposite meanings. For example: <i>hot</i> ↔ <i>cold</i>, <i>large</i> ↔ <i>small</i>, <i>thick</i> ↔ <i>thin</i>, <i>synonym</i> ↔ <i>antonym</i></li>\r\n<li><b><a href=\"https://en.wikipedia.org/wiki/Hypernym\" title=\"Hypernym\" class=\"mw-redirect\">Hypernyms</a></b> and <b><a href=\"https://en.wikipedia.org/wiki/Hyponym\" title=\"Hyponym\" class=\"mw-redirect\">hyponyms</a></b> are words that refer to, respectively, a general category and a specific instance of that category. For example, <i>vehicle</i> is a hypernym of <i>car</i>, and <i>car</i> is a hyponym of <i>vehicle</i>.</li>\r\n<li><b><a href=\"https://en.wikipedia.org/wiki/Homophone\" title=\"Homophone\">Homophones</a></b> are words that have the same pronunciation, but different meanings. For example, <i>witch</i> and <i>which</i> are homophones in most accents (because they are pronounced the same).</li>\r\n<li><b><a href=\"https://en.wikipedia.org/wiki/Homograph\" title=\"Homograph\">Homographs</a></b> are words that have the same spelling, but have different pronunciations. For example, one can <i>record</i> a song or keep a <i>record</i> of documents.</li>\r\n<li><b><a href=\"https://en.wikipedia.org/wiki/Homonym\" title=\"Homonym\">Homonyms</a></b> are words that have the same pronunciation and spelling, but have different meanings. For example, <i>rose</i> (a type of flower) and <i>rose</i> (past tense of <i>rise</i>) are homonyms.</li>\r\n</ul>\r\n<h2><span class=\"mw-headline\" id=\"See_also\">See also</span></h2>\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/-onym\" title=\"-onym\">-onym</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Synonym_ring\" title=\"Synonym ring\">Synonym ring</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Cognitive_synonymy\" title=\"Cognitive synonymy\">Cognitive synonymy</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Elegant_variation\" title=\"Elegant variation\">Elegant variation</a>, the gratuitous use of a synonym in prose</li>\r\n</ul>\r\n<h2><span class=\"mw-headline\" id=\"References\">References</span></h2>\r\n<div class=\"reflist\" style=\"list-style-type: decimal;\">\r\n<ol class=\"references\">\r\n<li id=\"cite_note-Stanojevi.C4.87_2009-1\"></li></ol></div><li id=\"cite_note-Stanojevi.C4.87_2009-1\"> <span class=\"reference-text\"><cite id=\"CITEREFStanojevi.C4.872009\" class=\"citation\">Stanojević, Maja (2009), <a rel=\"nofollow\" class=\"external text\" href=\"http://facta.junis.ni.ac.rs/lal/lal200902/lal200902-05.pdf\">\"Cognitive synonymy: a general overview\"</a> <span style=\"font-size:85%;\">(PDF)</span>, <i>Facta Universitatis, Linguistics and Literature series</i> <b>7</b> (2): 193–200.</cite><span title=\"ctx_ver=Z39.88-2004&amp;rfr_id=info%3Asid%2Fen.wikipedia.org%3ASynonym&amp;rft.atitle=Cognitive+synonymy%3A+a+general+overview&amp;rft.aufirst=Maja&amp;rft.aulast=Stanojevi%C4%87&amp;rft.date=2009&amp;rft.genre=article&amp;rft_id=http%3A%2F%2Ffacta.junis.ni.ac.rs%2Flal%2Flal200902%2Flal200902-05.pdf&amp;rft.issue=2&amp;rft.jtitle=Facta+Universitatis%2C+Linguistics+and+Literature+series&amp;rft.pages=193-200&amp;rft_val_fmt=info%3Aofi%2Ffmt%3Akev%3Amtx%3Ajournal&amp;rft.volume=7\" class=\"Z3988\"></span></span></li>\r\n<li id=\"cite_note-2\"></li><div class=\"reflist\" style=\"list-style-type: decimal;\"><ol class=\"references\"><li value=\"2\" id=\"cite_note-2\"> <span class=\"reference-text\"><cite class=\"citation book\"><a rel=\"nofollow\" class=\"external text\" href=\"http://books.google.com/books?id=bGSSl9I4maEC&amp;pg=PA370&amp;dq=a+widespr…n%20that...%20Hollywood%20was%20synonymous%20with%20immorality&amp;f=false\"><i>The Fitzgeralds and the Kennedys</i></a>. Macmillan. 1991. p. 370. <a href=\"https://en.wikipedia.org/wiki/International_Standard_Book_Number\" title=\"International Standard Book Number\">ISBN</a> <a href=\"https://en.wikipedia.org/wiki/Special:BookSources/9780312063542\" title=\"Special:BookSources/9780312063542\">9780312063542</a><span class=\"reference-accessdate\">. Retrieved <span class=\"nowrap\">2014-05-27</span></span>.</cite><span title=\"ctx_ver=Z39.88-2004&amp;rfr_id=info%3Asid%2Fen.wikipedia.org%3ASynonym&amp;rft.btitle=The+Fitzgeralds+and+the+Kennedys&amp;rft.date=1991&amp;rft.genre=book&amp;rft_id=http%3A%2F%2Fbooks.google.com%2Fbooks%3Fid%3DbGSSl9I4maEC%26pg%3DPA370%26dq%3Da%2Bwidespread%2Bimpression%2Bthat...%2BHollywood%2Bwas%2Bsynonymous%2Bwith%2Bimmorality%26hl%3Den%26sa%3DX%26ei%3DIRSEU7roL--_sQSb7IGgCA%26ved%3D0CCsQ6AEwAA%23v%3Donepage%26q%3Da%2520widespread%2520impression%2520that...%2520Hollywood%2520was%2520synonymous%2520with%2520immorality%26f%3Dfalse&amp;rft.isbn=9780312063542&amp;rft.pages=370&amp;rft.pub=Macmillan&amp;rft_val_fmt=info%3Aofi%2Ffmt%3Akev%3Amtx%3Abook\" class=\"Z3988\"></span></span></li>\r\n</ol>\r\n</div>\r\n<h2><span class=\"mw-headline\" id=\"External_links\">External links</span></h2><div id=\"content\" class=\"mw-body\" role=\"main\"><div id=\"bodyContent\" class=\"mw-body-content\"><div id=\"mw-content-text\" dir=\"ltr\" class=\"mw-content-ltr\" lang=\"en\">\r\n<table class=\"mbox-small plainlinks sistersitebox\" style=\"border:1px solid #aaa;background-color:#f9f9f9\">\r\n<tbody><tr>\r\n<td class=\"mbox-image\"><img alt=\"\" src=\"https://upload.wikimedia.org/wikipedia/commons/thumb/f/f8/Wiktionary-logo-en.svg/37px-Wiktionary-logo-en.svg.png\" srcset=\"//upload.wikimedia.org/wikipedia/commons/thumb/f/f8/Wiktionary-logo-en.svg/55px-Wiktionary-logo-en.svg.png 1.5x, //upload.wikimedia.org/wikipedia/commons/thumb/f/f8/Wiktionary-logo-en.svg/73px-Wiktionary-logo-en.svg.png 2x\" data-file-width=\"1000\" data-file-height=\"1089\" height=\"40\" width=\"37\"></td>\r\n<td class=\"mbox-text plainlist\">Look up <i><b><a href=\"https://en.wiktionary.org/wiki/Special:Search/synonym\" class=\"extiw\" title=\"wiktionary:Special:Search/synonym\">synonym</a></b></i> in Wiktionary, the free dictionary.</td>\r\n</tr>\r\n</tbody></table>\r\n<p>Tools which graph words relations :</p>\r\n<ul>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://graphwords.com/\">Graph Words</a> - Online tool for visualization word relations</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.synonyms.net/\">Synonyms.net</a>\r\n - Online reference resource that provides instant synonyms and antonyms\r\n definitions including visualizations, voice pronunciations and\r\ntranslations</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://dico.isc.cnrs.fr/en/index.html\">English/French Semantic Atlas</a>\r\n - Graph words relations in English, French and gives cross\r\nrepresentations for translations - offers 500 searches per user per day.</li>\r\n</ul>\r\n<p>Plain words synonyms finder :</p>\r\n<ul>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.synonym-finder.com/\">Synonym Finder</a> - Synonym finder including hypernyms in search result</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.how-to-say.net\">how to say</a> - Online Synonym finder</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://synonymosum.com/\">Synonymosum</a> - Online Synonym Dictionary - words and their synonyms</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://trovami.altervista.org/sinonimi/en\">Thesaurus</a> - Online synonyms in English, Italian, French and German</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://synonyms.woxikon.com/\">Woxikon Synonyms</a> - Over 1 million synonyms - English, German, Spanish, French, Italian, Portuguese, Swedish and Dutch</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.thefreedictionary.com/\">Freedictionary.com</a> Free online English thesaurus and dictionary containing synonyms, related words, antonyms, definitions, idioms and more</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.powerthesaurus.org/\">Power Thesaurus</a> - Thesaurus with synonyms ordered by rating</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.findmewords.com/synonyms.html\">FindMeWords Synonyms</a> - Online Synonym Dictionary with definitions</li>\r\n</ul>\r\n<table class=\"navbox\" style=\"border-spacing:0\">\r\n<tbody><tr>\r\n<td style=\"padding:2px\">\r\n<table id=\"collapsibleTable0\" class=\"nowraplinks collapsible autocollapse navbox-inner\" style=\"border-spacing:0;background:transparent;color:inherit\">\r\n<tbody><tr>\r\n<th scope=\"col\" class=\"navbox-title\" colspan=\"2\"><span class=\"collapseButton\">[<a href=\"https://en.wikipedia.org/wiki/Synonym#\" id=\"collapseButton0\">hide</a>]</span>\r\n<div class=\"plainlinks hlist navbar mini\">\r\n<ul>\r\n<li class=\"nv-view\"><a href=\"https://en.wikipedia.org/wiki/Template:Lexicology\" title=\"Template:Lexicology\"><abbr title=\"View this template\" style=\";;background:none transparent;border:none;\">v</abbr></a></li>\r\n<li class=\"nv-talk\"><a href=\"https://en.wikipedia.org/wiki/Template_talk:Lexicology\" title=\"Template talk:Lexicology\"><abbr title=\"Discuss this template\" style=\";;background:none transparent;border:none;\">t</abbr></a></li>\r\n<li class=\"nv-edit\"><a class=\"external text\" href=\"https://en.wikipedia.org/w/index.php?title=Template:Lexicology&amp;action=edit\"><abbr title=\"Edit this template\" style=\";;background:none transparent;border:none;\">e</abbr></a></li>\r\n</ul>\r\n</div>\r\n<div style=\"font-size:114%\"><a href=\"https://en.wikipedia.org/wiki/Lexicology\" title=\"Lexicology\">Lexicology</a></div>\r\n</th>\r\n</tr>\r\n<tr style=\"height:2px\">\r\n<td colspan=\"2\"></td>\r\n</tr>\r\n<tr>\r\n<th scope=\"row\" class=\"navbox-group\">Major terms</th>\r\n<td class=\"navbox-list navbox-odd hlist\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexical_item\" title=\"Lexical item\">Lexical item</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexicon\" title=\"Lexicon\">Lexicon</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexis_%28linguistics%29\" title=\"Lexis (linguistics)\">Lexis</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Word\" title=\"Word\">Word</a></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n<tr style=\"height:2px\">\r\n<td colspan=\"2\"></td>\r\n</tr>\r\n<tr>\r\n<th scope=\"row\" class=\"navbox-group\">Elements</th>\r\n<td class=\"navbox-list navbox-even hlist\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Cherology\" title=\"Cherology\">Chereme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Glyph\" title=\"Glyph\">Glyphs</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Grapheme\" title=\"Grapheme\">Grapheme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lemma_%28psycholinguistics%29\" title=\"Lemma (psycholinguistics)\">Lemma</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexeme\" title=\"Lexeme\">Lexeme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Meronymy\" title=\"Meronymy\">Meronymy</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Morpheme\" title=\"Morpheme\">Morpheme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Phoneme\" title=\"Phoneme\">Phoneme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Seme_%28semantics%29\" title=\"Seme (semantics)\">Seme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Sememe\" title=\"Sememe\">Sememe</a></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n<tr style=\"height:2px\">\r\n<td colspan=\"2\"></td>\r\n</tr>\r\n<tr>\r\n<th scope=\"row\" class=\"navbox-group\"><a href=\"https://en.wikipedia.org/wiki/Semantics\" title=\"Semantics\">Semantic relations</a></th>\r\n<td class=\"navbox-list navbox-odd hlist\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Opposite_%28semantics%29\" title=\"Opposite (semantics)\">Antonymy</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Holonymy\" title=\"Holonymy\">Holonymy</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Hyponymy_and_hypernymy\" title=\"Hyponymy and hypernymy\">Hyponymy and hypernymy</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Idiom\" title=\"Idiom\">Idiom</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexical_semantics\" title=\"Lexical semantics\">Lexical semantics</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Semantic_network\" title=\"Semantic network\">Semantic network</a></li>\r\n<li><strong class=\"selflink\">Synonym</strong></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Troponymy\" title=\"Troponymy\">Troponymy</a></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n<tr style=\"height:2px\">\r\n<td colspan=\"2\"></td>\r\n</tr>\r\n<tr>\r\n<th scope=\"row\" class=\"navbox-group\">Functions</th>\r\n<td class=\"navbox-list navbox-even hlist\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Function_word\" title=\"Function word\">Function word</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Headword\" title=\"Headword\">Headword</a></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n<tr style=\"height:2px\">\r\n<td colspan=\"2\"></td>\r\n</tr>\r\n<tr>\r\n<th scope=\"row\" class=\"navbox-group\">Fields</th>\r\n<td class=\"navbox-list navbox-odd hlist\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Controlled_vocabulary\" title=\"Controlled vocabulary\">Controlled vocabulary</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/English_lexicology_and_lexicography\" title=\"English lexicology and lexicography\">English lexicology and lexicography</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/International_scientific_vocabulary\" title=\"International scientific vocabulary\">International scientific vocabulary</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexicographic_error\" title=\"Lexicographic error\">Lexicographic error</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexicographic_information_cost\" title=\"Lexicographic information cost\">Lexicographic information cost</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Linguistic_prescription\" title=\"Linguistic prescription\">Linguistic prescription</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Morphology_%28linguistics%29\" title=\"Morphology (linguistics)\">Morphology</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Specialised_lexicography\" title=\"Specialised lexicography\">Specialised lexicography</a></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n</tbody></table>\r\n</td>\r\n</tr>\r\n</tbody></table>\r\n<table class=\"navbox\" style=\"border-spacing:0\">\r\n<tbody><tr>\r\n<td style=\"padding:2px\">\r\n<table class=\"nowraplinks hlist navbox-inner\" style=\"border-spacing:0;background:transparent;color:inherit\">\r\n<tbody><tr>\r\n<th scope=\"row\" class=\"navbox-group\"><a href=\"https://en.wikipedia.org/wiki/Help:Authority_control\" title=\"Help:Authority control\">Authority control</a></th>\r\n<td class=\"navbox-list navbox-odd\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Library_of_Congress_Control_Number\" title=\"Library of Congress Control Number\">LCCN</a>: <span class=\"uid\"><a rel=\"nofollow\" class=\"external text\" href=\"http://id.loc.gov/authorities/subjects/sh85131642\">sh85131642</a></span></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Integrated_Authority_File\" title=\"Integrated Authority File\">GND</a>: <span class=\"uid\"><a rel=\"nofollow\" class=\"external text\" href=\"http://d-nb.info/gnd/4058765-4\">4058765-4</a></span></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n</tbody></table>\r\n</td>\r\n</tr>\r\n</tbody></table>\r\n\r\n\r\n\r\n\r\n\r\n\r\n\r\n</div>\r\n        <div id=\"catlinks\" class=\"catlinks\"><div id=\"mw-normal-catlinks\" class=\"mw-normal-catlinks\"><a href=\"https://en.wikipedia.org/wiki/Help:Category\" title=\"Help:Category\">Categories</a>: <ul><li><a href=\"https://en.wikipedia.org/wiki/Category:Lexical_semantics\" title=\"Category:Lexical semantics\">Lexical semantics</a></li><li><a href=\"https://en.wikipedia.org/wiki/Category:Types_of_words\" title=\"Category:Types of words\">Types of words</a></li></ul></div></div>        <div class=\"visualClear\"></div>\r\n              </div>\r\n    </div>\r\n    <div id=\"mw-navigation\">\r\n      <h2>Navigation menu</h2>\r\n\r\n      <div id=\"mw-head\">\r\n                  <div id=\"p-personal\" role=\"navigation\" class=\"\" aria-labelledby=\"p-personal-label\">\r\n\r\n            <ul>\r\n              <li id=\"pt-createaccount\"><a href=\"https://en.wikipedia.org/w/index.php?title=Special:UserLogin&amp;returnto=Synonym&amp;type=signup\" title=\"You are encouraged to create an account and log in; however, it is not mandatory\">Create account</a></li><li id=\"pt-userpage\"><span style=\"color: gray;\">Not logged in</span></li><li id=\"pt-mytalk\"><a accesskey=\"n\" title=\"Your talk page [Alt+Shift+n]\" href=\"https://en.wikipedia.org/wiki/Special:MyTalk\">Talk</a></li><li id=\"pt-mycontris\"><a accesskey=\"y\" title=\"A list of your contributions [Alt+Shift+y]\" href=\"https://en.wikipedia.org/wiki/Special:MyContributions\">Contributions</a></li><li id=\"pt-login\"><a href=\"https://en.wikipedia.org/w/index.php?title=Special:UserLogin&amp;returnto=Synonym\" title=\"You're encouraged to log in; however, it's not mandatory. [Alt+Shift+o]\" accesskey=\"o\">Log in</a></li>            </ul>\r\n          </div>\r\n                  <div id=\"left-navigation\">\r\n                    <div id=\"p-namespaces\" role=\"navigation\" class=\"vectorTabs\" aria-labelledby=\"p-namespaces-label\">\r\n\r\n            <ul>\r\n                              <li id=\"ca-nstab-main\" class=\"selected\"><span><a href=\"https://en.wikipedia.org/wiki/Synonym\" title=\"View the content page [Alt+Shift+c]\" accesskey=\"c\">Article</a></span></li>\r\n                              <li id=\"ca-talk\"><span><a href=\"https://en.wikipedia.org/wiki/Talk:Synonym\" title=\"Discussion about the content page [Alt+Shift+t]\" accesskey=\"t\" rel=\"discussion\">Talk</a></span></li>\r\n                          </ul>\r\n          </div>\r\n\r\n                  </div>\r\n        <div id=\"right-navigation\">\r\n                    <div id=\"p-views\" role=\"navigation\" class=\"vectorTabs\" aria-labelledby=\"p-views-label\">\r\n\r\n            <ul>\r\n                              <li id=\"ca-view\" class=\"selected\"><span><a href=\"https://en.wikipedia.org/wiki/Synonym\">Read</a></span></li>\r\n                              <li id=\"ca-edit\"><span><a href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;action=edit\" title=\"Edit this page [Alt+Shift+e]\" accesskey=\"e\">Edit</a></span></li>\r\n                              <li id=\"ca-history\" class=\"collapsible\"><span><a href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;action=history\" title=\"Past revisions of this page [Alt+Shift+h]\" accesskey=\"h\">View history</a></span></li>\r\n                          </ul>\r\n          </div>\r\n\r\n                    <div id=\"p-search\" role=\"search\">\r\n\r\n\r\n            <form action=\"/w/index.php\" id=\"searchform\">\r\n              <div id=\"simpleSearch\">\r\n              <input autocomplete=\"off\" tabindex=\"1\" name=\"search\" placeholder=\"Search\" title=\"Search Wikipedia [Alt+Shift+f]\" accesskey=\"f\" id=\"searchInput\" type=\"search\"></div></form></div></div></div></div><div id=\"mw-navigation\"><div id=\"mw-head\"><div id=\"right-navigation\"><div id=\"p-search\" role=\"search\"><form action=\"/w/index.php\" id=\"searchform\"><div id=\"simpleSearch\">              </div>\r\n            </form>\r\n          </div>\r\n                  </div>\r\n      </div>\r\n      <div id=\"mw-panel\">\r\n        <div id=\"p-logo\" role=\"banner\"><a class=\"mw-wiki-logo\" href=\"https://en.wikipedia.org/wiki/Main_Page\" title=\"Visit the main page\"></a></div>\r\n            <div class=\"portal\" role=\"navigation\" id=\"p-navigation\" aria-labelledby=\"p-navigation-label\">\r\n\r\n\r\n      <div class=\"body\">\r\n                  <ul>\r\n            <li id=\"n-mainpage-description\"><a href=\"https://en.wikipedia.org/wiki/Main_Page\" title=\"Visit the main page [Alt+Shift+z]\" accesskey=\"z\">Main page</a></li><li id=\"n-contents\"><a href=\"https://en.wikipedia.org/wiki/Portal:Contents\" title=\"Guides to browsing Wikipedia\">Contents</a></li><li id=\"n-featuredcontent\"><a href=\"https://en.wikipedia.org/wiki/Portal:Featured_content\" title=\"Featured content – the best of Wikipedia\">Featured content</a></li><li id=\"n-currentevents\"><a href=\"https://en.wikipedia.org/wiki/Portal:Current_events\" title=\"Find background information on current events\">Current events</a></li><li id=\"n-randompage\"><a href=\"https://en.wikipedia.org/wiki/Special:Random\" title=\"Load a random article [Alt+Shift+x]\" accesskey=\"x\">Random article</a></li><li id=\"n-sitesupport\"><a href=\"https://donate.wikimedia.org/wiki/Special:FundraiserRedirector?utm_source=d…mp;utm_medium=sidebar&amp;utm_campaign=C13_en.wikipedia.org&amp;uselang=en\" title=\"Support us\">Donate to Wikipedia</a></li><li id=\"n-shoplink\"><a href=\"https://shop.wikimedia.org\" title=\"Visit the Wikipedia store\">Wikipedia store</a></li>          </ul>\r\n              </div>\r\n    </div>\r\n      <div class=\"portal\" role=\"navigation\" id=\"p-interaction\" aria-labelledby=\"p-interaction-label\">\r\n      <h3 id=\"p-interaction-label\">Interaction</h3>\r\n\r\n      <div class=\"body\">\r\n                  <ul>\r\n            <li id=\"n-help\"><a href=\"https://en.wikipedia.org/wiki/Help:Contents\" title=\"Guidance on how to use and edit Wikipedia\">Help</a></li><li id=\"n-aboutsite\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:About\" title=\"Find out about Wikipedia\">About Wikipedia</a></li><li id=\"n-portal\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:Community_portal\" title=\"About the project, what you can do, where to find things\">Community portal</a></li><li id=\"n-recentchanges\"><a href=\"https://en.wikipedia.org/wiki/Special:RecentChanges\" title=\"A list of recent changes in the wiki [Alt+Shift+r]\" accesskey=\"r\">Recent changes</a></li><li id=\"n-contactpage\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:Contact_us\" title=\"How to contact Wikipedia\">Contact page</a></li>          </ul>\r\n              </div>\r\n    </div>\r\n      <div class=\"portal\" role=\"navigation\" id=\"p-tb\" aria-labelledby=\"p-tb-label\">\r\n      <h3 id=\"p-tb-label\">Tools</h3>\r\n\r\n      <div class=\"body\">\r\n                  <ul>\r\n            <li id=\"t-whatlinkshere\"><a href=\"https://en.wikipedia.org/wiki/Special:WhatLinksHere/Synonym\" title=\"List of all English Wikipedia pages containing links to this page [Alt+Shift+j]\" accesskey=\"j\">What links here</a></li><li id=\"t-recentchangeslinked\"><a href=\"https://en.wikipedia.org/wiki/Special:RecentChangesLinked/Synonym\" title=\"Recent changes in pages linked from this page [Alt+Shift+k]\" accesskey=\"k\">Related changes</a></li><li id=\"t-upload\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:File_Upload_Wizard\" title=\"Upload files [Alt+Shift+u]\" accesskey=\"u\">Upload file</a></li><li id=\"t-specialpages\"><a href=\"https://en.wikipedia.org/wiki/Special:SpecialPages\" title=\"A list of all special pages [Alt+Shift+q]\" accesskey=\"q\">Special pages</a></li><li id=\"t-permalink\"><a href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;oldid=690006505\" title=\"Permanent link to this revision of the page\">Permanent link</a></li><li id=\"t-info\"><a href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;action=info\" title=\"More information about this page\">Page information</a></li><li id=\"t-wikibase\"><a href=\"https://www.wikidata.org/wiki/Q42106\" title=\"Link to connected data repository item [Alt+Shift+g]\" accesskey=\"g\">Wikidata item</a></li><li id=\"t-cite\"><a href=\"https://en.wikipedia.org/w/index.php?title=Special:CiteThisPage&amp;page=Synonym&amp;id=690006505\" title=\"Information on how to cite this page\">Cite this page</a></li>          </ul>\r\n              </div>\r\n    </div>\r\n      <div class=\"portal\" role=\"navigation\" id=\"p-coll-print_export\" aria-labelledby=\"p-coll-print_export-label\">\r\n      <h3 id=\"p-coll-print_export-label\">Print/export</h3>\r\n\r\n      <div class=\"body\">\r\n                  <ul>\r\n            <li id=\"coll-create_a_book\"><a href=\"https://en.wikipedia.org/w/index.php?title=Special:Book&amp;bookcmd=book_creator&amp;referer=Synonym\">Create a book</a></li><li id=\"coll-download-as-rdf2latex\"><a href=\"https://en.wikipedia.org/w/index.php?title=Special:Book&amp;bookcmd=render_…itle=Synonym&amp;returnto=Synonym&amp;oldid=690006505&amp;writer=rdf2latex\">Download as PDF</a></li><li id=\"t-print\"><a href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;printable=yes\" title=\"Printable version of this page [Alt+Shift+p]\" accesskey=\"p\">Printable version</a></li>          </ul>\r\n              </div>\r\n    </div>\r\n      <div class=\"portal\" role=\"navigation\" id=\"p-lang\" aria-labelledby=\"p-lang-label\"><span aria-haspopup=\"true\" role=\"button\" tabindex=\"0\" title=\"Language settings\" class=\"uls-settings-trigger\"></span>\r\n      <h3 id=\"p-lang-label\">Languages</h3>\r\n\r\n      <div class=\"body\">\r\n                  <ul>\r\n            <li class=\"interlanguage-link interwiki-af\"><a href=\"https://af.wikipedia.org/wiki/Sinoniem\" title=\"Sinoniem – Afrikaans\" hreflang=\"af\" lang=\"af\">Afrikaans</a></li><li class=\"interlanguage-link interwiki-ar\"><a href=\"https://ar.wikipedia.org/wiki/%D8%AA%D8%B1%D8%A7%D8%AF%D9%81\" title=\"ترادف – Arabic\" hreflang=\"ar\" lang=\"ar\">العربية</a></li><li class=\"interlanguage-link interwiki-az\"><a href=\"https://az.wikipedia.org/wiki/Sinoniml%C9%99r\" title=\"Sinonimlər – Azerbaijani\" hreflang=\"az\" lang=\"az\">Azərbaycanca</a></li><li class=\"interlanguage-link interwiki-be\"><a href=\"https://be.wikipedia.org/wiki/%D0%A1%D1%96%D0%BD%D0%BE%D0%BD%D1%96%D0%BC%D1%8B\" title=\"Сінонімы – Belarusian\" hreflang=\"be\" lang=\"be\">Беларуская</a></li><li class=\"interlanguage-link interwiki-be-x-old\"><a href=\"https://be-x-old.wikipedia.org/wiki/%D0%A1%D1%8B%D0%BD%D0%BE%D0%BD%D1%96%D0%BC%D1%8B\" title=\"Сынонімы – беларуская (тарашкевіца)‎\" hreflang=\"be-x-old\" lang=\"be-x-old\">Беларуская (тарашкевіца)‎</a></li><li class=\"interlanguage-link interwiki-bg\"><a href=\"https://bg.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC\" title=\"Синоним – Bulgarian\" hreflang=\"bg\" lang=\"bg\">Български</a></li><li class=\"interlanguage-link interwiki-bar\"><a href=\"https://bar.wikipedia.org/wiki/Synonym\" title=\"Synonym – Bavarian\" hreflang=\"bar\" lang=\"bar\">Boarisch</a></li><li class=\"interlanguage-link interwiki-ca\"><a href=\"https://ca.wikipedia.org/wiki/Sin%C3%B2nim\" title=\"Sinònim – Catalan\" hreflang=\"ca\" lang=\"ca\">Català</a></li><li class=\"interlanguage-link interwiki-cv\"><a href=\"https://cv.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC%D1%81%D0%B5%D0%BC\" title=\"Синонимсем – Chuvash\" hreflang=\"cv\" lang=\"cv\">Чӑвашла</a></li><li class=\"interlanguage-link interwiki-cs\"><a href=\"https://cs.wikipedia.org/wiki/Synonymum\" title=\"Synonymum – Czech\" hreflang=\"cs\" lang=\"cs\">Čeština</a></li><li class=\"interlanguage-link interwiki-cy\"><a href=\"https://cy.wikipedia.org/wiki/Cyfystyr\" title=\"Cyfystyr – Welsh\" hreflang=\"cy\" lang=\"cy\">Cymraeg</a></li><li class=\"interlanguage-link interwiki-da\"><a href=\"https://da.wikipedia.org/wiki/Synonym\" title=\"Synonym – Danish\" hreflang=\"da\" lang=\"da\">Dansk</a></li><li class=\"interlanguage-link interwiki-de\"><a href=\"https://de.wikipedia.org/wiki/Synonym\" title=\"Synonym – German\" hreflang=\"de\" lang=\"de\">Deutsch</a></li><li class=\"interlanguage-link interwiki-et\"><a href=\"https://et.wikipedia.org/wiki/S%C3%BCnon%C3%BC%C3%BCm\" title=\"Sünonüüm – Estonian\" hreflang=\"et\" lang=\"et\">Eesti</a></li><li class=\"interlanguage-link interwiki-el\"><a href=\"https://el.wikipedia.org/wiki/%CE%A3%CF%85%CE%BD%CF%8E%CE%BD%CF%85%CE%BC%CE%BF\" title=\"Συνώνυμο – Greek\" hreflang=\"el\" lang=\"el\">Ελληνικά</a></li><li class=\"interlanguage-link interwiki-es\"><a href=\"https://es.wikipedia.org/wiki/Sinonimia_%28sem%C3%A1ntica%29\" title=\"Sinonimia (semántica) – Spanish\" hreflang=\"es\" lang=\"es\">Español</a></li><li class=\"interlanguage-link interwiki-eo\"><a href=\"https://eo.wikipedia.org/wiki/Sinonimo\" title=\"Sinonimo – Esperanto\" hreflang=\"eo\" lang=\"eo\">Esperanto</a></li><li class=\"interlanguage-link interwiki-eu\"><a href=\"https://eu.wikipedia.org/wiki/Sinonimo\" title=\"Sinonimo – Basque\" hreflang=\"eu\" lang=\"eu\">Euskara</a></li><li class=\"interlanguage-link interwiki-fa\"><a href=\"https://fa.wikipedia.org/wiki/%D9%85%D8%AA%D8%B1%D8%A7%D8%AF%D9%81\" title=\"مترادف – Persian\" hreflang=\"fa\" lang=\"fa\">فارسی</a></li><li class=\"interlanguage-link interwiki-fo\"><a href=\"https://fo.wikipedia.org/wiki/Samheiti\" title=\"Samheiti – Faroese\" hreflang=\"fo\" lang=\"fo\">Føroyskt</a></li><li class=\"interlanguage-link interwiki-fr\"><a href=\"https://fr.wikipedia.org/wiki/Synonymie\" title=\"Synonymie – French\" hreflang=\"fr\" lang=\"fr\">Français</a></li><li class=\"interlanguage-link interwiki-ga\"><a href=\"https://ga.wikipedia.org/wiki/Comhchiallach\" title=\"Comhchiallach – Irish\" hreflang=\"ga\" lang=\"ga\">Gaeilge</a></li><li class=\"interlanguage-link interwiki-gl\"><a href=\"https://gl.wikipedia.org/wiki/Sinonimia\" title=\"Sinonimia – Galician\" hreflang=\"gl\" lang=\"gl\">Galego</a></li><li class=\"interlanguage-link interwiki-ko\"><a href=\"https://ko.wikipedia.org/wiki/%EB%8F%99%EC%9D%98%EC%96%B4\" title=\"동의어 – Korean\" hreflang=\"ko\" lang=\"ko\">한국어</a></li><li class=\"interlanguage-link interwiki-hy\"><a href=\"https://hy.wikipedia.org/wiki/%D5%80%D5%B8%D5%B4%D5%A1%D5%B6%D5%AB%D5%B7\" title=\"Հոմանիշ – Armenian\" hreflang=\"hy\" lang=\"hy\">Հայերեն</a></li><li class=\"interlanguage-link interwiki-hi\"><a href=\"https://hi.wikipedia.org/wiki/%E0%A4%AA%E0%A4%B0%E0%A5%8D%E0%A4%AF%E0%A4%BE%E0%A4%AF%E0%A4%B5%E0%A4%BE%E0%A4%9A%E0%A5%80\" title=\"पर्यायवाची – Hindi\" hreflang=\"hi\" lang=\"hi\">हिन्दी</a></li><li class=\"interlanguage-link interwiki-hr\"><a href=\"https://hr.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Croatian\" hreflang=\"hr\" lang=\"hr\">Hrvatski</a></li><li class=\"interlanguage-link interwiki-io\"><a href=\"https://io.wikipedia.org/wiki/Sinonimo\" title=\"Sinonimo – Ido\" hreflang=\"io\" lang=\"io\">Ido</a></li><li class=\"interlanguage-link interwiki-id\"><a href=\"https://id.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Indonesian\" hreflang=\"id\" lang=\"id\">Bahasa Indonesia</a></li><li class=\"interlanguage-link interwiki-ia\"><a href=\"https://ia.wikipedia.org/wiki/Synonymo\" title=\"Synonymo – Interlingua\" hreflang=\"ia\" lang=\"ia\">Interlingua</a></li><li class=\"interlanguage-link interwiki-is\"><a href=\"https://is.wikipedia.org/wiki/Samheiti\" title=\"Samheiti – Icelandic\" hreflang=\"is\" lang=\"is\">Íslenska</a></li><li class=\"interlanguage-link interwiki-it\"><a href=\"https://it.wikipedia.org/wiki/Sinonimia\" title=\"Sinonimia – Italian\" hreflang=\"it\" lang=\"it\">Italiano</a></li><li class=\"interlanguage-link interwiki-he\"><a href=\"https://he.wikipedia.org/wiki/%D7%9E%D7%99%D7%9C%D7%94_%D7%A0%D7%A8%D7%93%D7%A4%D7%AA\" title=\"מילה נרדפת – Hebrew\" hreflang=\"he\" lang=\"he\">עברית</a></li><li class=\"interlanguage-link interwiki-jv\"><a href=\"https://jv.wikipedia.org/wiki/Dasanama\" title=\"Dasanama – Javanese\" hreflang=\"jv\" lang=\"jv\">Basa Jawa</a></li><li class=\"interlanguage-link interwiki-ka\"><a href=\"https://ka.wikipedia.org/wiki/%E1%83%A1%E1%83%98%E1%83%9C%E1%83%9D%E1%83%9C%E1%83%98%E1%83%9B%E1%83%94%E1%83%91%E1%83%98\" title=\"სინონიმები – Georgian\" hreflang=\"ka\" lang=\"ka\">ქართული</a></li><li class=\"interlanguage-link interwiki-kk\"><a href=\"https://kk.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC\" title=\"Синоним – Kazakh\" hreflang=\"kk\" lang=\"kk\">Қазақша</a></li><li class=\"interlanguage-link interwiki-ku\"><a href=\"https://ku.wikipedia.org/wiki/Hevwate\" title=\"Hevwate – Kurdish\" hreflang=\"ku\" lang=\"ku\">Kurdî</a></li><li class=\"interlanguage-link interwiki-la\"><a href=\"https://la.wikipedia.org/wiki/Synonymia\" title=\"Synonymia – Latin\" hreflang=\"la\" lang=\"la\">Latina</a></li><li class=\"interlanguage-link interwiki-lv\"><a href=\"https://lv.wikipedia.org/wiki/Sinon%C4%ABms\" title=\"Sinonīms – Latvian\" hreflang=\"lv\" lang=\"lv\">Latviešu</a></li><li class=\"interlanguage-link interwiki-lb\"><a href=\"https://lb.wikipedia.org/wiki/Synonymie\" title=\"Synonymie – Luxembourgish\" hreflang=\"lb\" lang=\"lb\">Lëtzebuergesch</a></li><li class=\"interlanguage-link interwiki-lt\"><a href=\"https://lt.wikipedia.org/wiki/Sinonimas\" title=\"Sinonimas – Lithuanian\" hreflang=\"lt\" lang=\"lt\">Lietuvių</a></li><li class=\"interlanguage-link interwiki-hu\"><a href=\"https://hu.wikipedia.org/wiki/Szinon%C3%ADmia\" title=\"Szinonímia – Hungarian\" hreflang=\"hu\" lang=\"hu\">Magyar</a></li><li class=\"interlanguage-link interwiki-mk\"><a href=\"https://mk.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC\" title=\"Синоним – Macedonian\" hreflang=\"mk\" lang=\"mk\">Македонски</a></li><li class=\"interlanguage-link interwiki-ml\"><a href=\"https://ml.wikipedia.org/wiki/%E0%B4%AA%E0%B4%B0%E0%B5%8D%E0%B4%AF%E0%B4%BE%E0%B4%AF%E0%B4%AA%E0%B4%A6%E0%B4%82\" title=\"പര്യായപദം – Malayalam\" hreflang=\"ml\" lang=\"ml\">മലയാളം</a></li><li class=\"interlanguage-link interwiki-ms\"><a href=\"https://ms.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Malay\" hreflang=\"ms\" lang=\"ms\">Bahasa Melayu</a></li><li class=\"interlanguage-link interwiki-nl\"><a href=\"https://nl.wikipedia.org/wiki/Synoniem_%28taalkunde%29\" title=\"Synoniem (taalkunde) – Dutch\" hreflang=\"nl\" lang=\"nl\">Nederlands</a></li><li class=\"interlanguage-link interwiki-ja\"><a href=\"https://ja.wikipedia.org/wiki/%E9%A1%9E%E7%BE%A9%E8%AA%9E\" title=\"類義語 – Japanese\" hreflang=\"ja\" lang=\"ja\">日本語</a></li><li class=\"interlanguage-link interwiki-ce\"><a href=\"https://ce.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC%D0%B0%D1%88\" title=\"Синонимаш – Chechen\" hreflang=\"ce\" lang=\"ce\">Нохчийн</a></li><li class=\"interlanguage-link interwiki-no\"><a href=\"https://no.wikipedia.org/wiki/Synonym\" title=\"Synonym – Norwegian\" hreflang=\"no\" lang=\"no\">Norsk bokmål</a></li><li class=\"interlanguage-link interwiki-nn\"><a href=\"https://nn.wikipedia.org/wiki/Synonym\" title=\"Synonym – Norwegian Nynorsk\" hreflang=\"nn\" lang=\"nn\">Norsk nynorsk</a></li><li class=\"interlanguage-link interwiki-uz\"><a href=\"https://uz.wikipedia.org/wiki/Sinonimlar\" title=\"Sinonimlar – Uzbek\" hreflang=\"uz\" lang=\"uz\">Oʻzbekcha/ўзбекча</a></li><li class=\"interlanguage-link interwiki-pfl\"><a href=\"https://pfl.wikipedia.org/wiki/Wort:Synonyme\" title=\"Wort:Synonyme – Palatine German\" hreflang=\"pfl\" lang=\"pfl\">Pälzisch</a></li><li class=\"interlanguage-link interwiki-pl\"><a href=\"https://pl.wikipedia.org/wiki/Synonim\" title=\"Synonim – Polish\" hreflang=\"pl\" lang=\"pl\">Polski</a></li><li class=\"interlanguage-link interwiki-pt\"><a href=\"https://pt.wikipedia.org/wiki/Sin%C3%B4nimo\" title=\"Sinônimo – Portuguese\" hreflang=\"pt\" lang=\"pt\">Português</a></li><li class=\"interlanguage-link interwiki-ro\"><a href=\"https://ro.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Romanian\" hreflang=\"ro\" lang=\"ro\">Română</a></li><li class=\"interlanguage-link interwiki-qu\"><a href=\"https://qu.wikipedia.org/wiki/Kaqlla_simi\" title=\"Kaqlla simi – Quechua\" hreflang=\"qu\" lang=\"qu\">Runa Simi</a></li><li class=\"interlanguage-link interwiki-ru\"><a href=\"https://ru.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC%D1%8B\" title=\"Синонимы – Russian\" hreflang=\"ru\" lang=\"ru\">Русский</a></li><li class=\"interlanguage-link interwiki-sco\"><a href=\"https://sco.wikipedia.org/wiki/Synonym\" title=\"Synonym – Scots\" hreflang=\"sco\" lang=\"sco\">Scots</a></li><li class=\"interlanguage-link interwiki-sq\"><a href=\"https://sq.wikipedia.org/wiki/Sinonimi\" title=\"Sinonimi – Albanian\" hreflang=\"sq\" lang=\"sq\">Shqip</a></li><li class=\"interlanguage-link interwiki-simple\"><a href=\"https://simple.wikipedia.org/wiki/Synonym\" title=\"Synonym – Simple English\" hreflang=\"simple\" lang=\"simple\">Simple English</a></li><li class=\"interlanguage-link interwiki-sk\"><a href=\"https://sk.wikipedia.org/wiki/Synonymum\" title=\"Synonymum – Slovak\" hreflang=\"sk\" lang=\"sk\">Slovenčina</a></li><li class=\"interlanguage-link interwiki-sl\"><a href=\"https://sl.wikipedia.org/wiki/Sopomenka\" title=\"Sopomenka – Slovenian\" hreflang=\"sl\" lang=\"sl\">Slovenščina</a></li><li class=\"interlanguage-link interwiki-sr\"><a href=\"https://sr.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC\" title=\"Синоним – Serbian\" hreflang=\"sr\" lang=\"sr\">Српски / srpski</a></li><li class=\"interlanguage-link interwiki-sh\"><a href=\"https://sh.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Serbo-Croatian\" hreflang=\"sh\" lang=\"sh\">Srpskohrvatski / српскохрватски</a></li><li class=\"interlanguage-link interwiki-su\"><a href=\"https://su.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Sundanese\" hreflang=\"su\" lang=\"su\">Basa Sunda</a></li><li class=\"interlanguage-link interwiki-fi\"><a href=\"https://fi.wikipedia.org/wiki/Synonymia\" title=\"Synonymia – Finnish\" hreflang=\"fi\" lang=\"fi\">Suomi</a></li><li class=\"interlanguage-link interwiki-sv\"><a href=\"https://sv.wikipedia.org/wiki/Synonym\" title=\"Synonym – Swedish\" hreflang=\"sv\" lang=\"sv\">Svenska</a></li><li class=\"interlanguage-link interwiki-ta\"><a href=\"https://ta.wikipedia.org/wiki/%E0%AE%92%E0%AE%A4%E0%AF%8D%E0%AE%A4%E0%AE%9A%E0%AF%8A%E0%AE%B2%E0%AF%8D\" title=\"ஒத்தசொல் – Tamil\" hreflang=\"ta\" lang=\"ta\">தமிழ்</a></li><li class=\"interlanguage-link interwiki-tt\"><a href=\"https://tt.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC\" title=\"Синоним – Tatar\" hreflang=\"tt\" lang=\"tt\">Татарча/tatarça</a></li><li class=\"interlanguage-link interwiki-te\"><a href=\"https://te.wikipedia.org/wiki/%E0%B0%AA%E0%B0%B0%E0%B1%8D%E0%B0%AF%E0%B0%BE%E0%B0%AF%E0%B0%AA%E0%B0%A6%E0%B0%82\" title=\"పర్యాయపదం – Telugu\" hreflang=\"te\" lang=\"te\">తెలుగు</a></li><li class=\"interlanguage-link interwiki-tr\"><a href=\"https://tr.wikipedia.org/wiki/E%C5%9Fanlaml%C4%B1\" title=\"Eşanlamlı – Turkish\" hreflang=\"tr\" lang=\"tr\">Türkçe</a></li><li class=\"interlanguage-link interwiki-uk\"><a href=\"https://uk.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D1%96%D0%BC\" title=\"Синонім – Ukrainian\" hreflang=\"uk\" lang=\"uk\">Українська</a></li><li class=\"interlanguage-link interwiki-ur\"><a href=\"https://ur.wikipedia.org/wiki/%D9%85%D8%AA%D8%B1%D8%A7%D8%AF%D9%81\" title=\"مترادف – Urdu\" hreflang=\"ur\" lang=\"ur\">اردو</a></li><li class=\"interlanguage-link interwiki-vi\"><a href=\"https://vi.wikipedia.org/wiki/T%E1%BB%AB_%C4%91%E1%BB%93ng_ngh%C4%A9a\" title=\"Từ đồng nghĩa – Vietnamese\" hreflang=\"vi\" lang=\"vi\">Tiếng Việt</a></li><li class=\"interlanguage-link interwiki-wa\"><a href=\"https://wa.wikipedia.org/wiki/Sinonimeye\" title=\"Sinonimeye – Walloon\" hreflang=\"wa\" lang=\"wa\">Walon</a></li><li class=\"interlanguage-link interwiki-zh-yue\"><a href=\"https://zh-yue.wikipedia.org/wiki/%E5%90%8C%E7%BE%A9%E8%A9%9E\" title=\"同義詞 – Cantonese\" hreflang=\"zh-yue\" lang=\"zh-yue\">粵語</a></li><li class=\"interlanguage-link interwiki-zh\"><a href=\"https://zh.wikipedia.org/wiki/%E5%90%8C%E4%B9%89%E8%AF%8D\" title=\"同义词 – Chinese\" hreflang=\"zh\" lang=\"zh\">中文</a></li>          </ul>\r\n        <div class=\"after-portlet after-portlet-lang\"><span class=\"wb-langlinks-edit wb-langlinks-link\"><a href=\"https://www.wikidata.org/wiki/Q42106#sitelinks-wikipedia\" title=\"Edit interlanguage links\" class=\"wbc-editpage\">Edit links</a></span></div>      </div>\r\n    </div>\r\n        </div>\r\n    </div>\r\n    <div id=\"footer\" role=\"contentinfo\">\r\n              <ul id=\"footer-info\">\r\n                      <li id=\"footer-info-lastmod\"> This page was last modified on 10 November 2015, at 18:44.</li>\r\n                      <li id=\"footer-info-copyright\">Text is available under the <a rel=\"license\" href=\"https://en.wikipedia.org/wiki/Wikipedia:Text_of_Creative_Commons_Attribution-ShareAlike_3.0_Unported_License\">Creative Commons Attribution-ShareAlike License</a>;\r\nadditional terms may apply.  By using this site, you agree to the <a href=\"https://wikimediafoundation.org/wiki/Terms_of_Use\">Terms of Use</a> and <a href=\"https://wikimediafoundation.org/wiki/Privacy_policy\">Privacy Policy</a>. Wikipedia® is a registered trademark of the <a href=\"https://www.wikimediafoundation.org/\">Wikimedia Foundation, Inc.</a>, a non-profit organization.</li>\r\n                  </ul>\r\n              <ul id=\"footer-places\">\r\n                      <li id=\"footer-places-privacy\"><a href=\"https://wikimediafoundation.org/wiki/Privacy_policy\" title=\"wmf:Privacy policy\">Privacy policy</a></li>\r\n                      <li id=\"footer-places-about\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:About\" title=\"Wikipedia:About\">About Wikipedia</a></li>\r\n                      <li id=\"footer-places-disclaimer\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:General_disclaimer\" title=\"Wikipedia:General disclaimer\">Disclaimers</a></li>\r\n                      <li id=\"footer-places-contact\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:Contact_us\">Contact Wikipedia</a></li>\r\n                      <li id=\"footer-places-developers\"><a href=\"https://www.mediawiki.org/wiki/Special:MyLanguage/How_to_contribute\">Developers</a></li>\r\n                      <li id=\"footer-places-mobileview\"><a href=\"https://en.m.wikipedia.org/w/index.php?title=Synonym&amp;mobileaction=toggle_view_mobile\" class=\"noprint stopMobileRedirectToggle\">Mobile view</a></li>\r\n                  </ul>\r\n                    <ul id=\"footer-icons\" class=\"noprint\">\r\n                      <li id=\"footer-copyrightico\">\r\n              <a href=\"https://wikimediafoundation.org/\"><img src=\"https://en.wikipedia.org/static/images/wikimedia-button.png\" srcset=\"/static/images/wikimedia-button-1.5x.png 1.5x, /static/images/wikimedia-button-2x.png 2x\" alt=\"Wikimedia Foundation\" height=\"31\" width=\"88\"></a>            </li>\r\n                      <li id=\"footer-poweredbyico\">\r\n              <a href=\"https://www.mediawiki.org/\"><img src=\"https://en.wikipedia.org/static/1.27.0-wmf.7/resources/assets/poweredby_mediawiki_88x31.png\" alt=\"Powered by MediaWiki\" srcset=\"/static/1.27.0-wmf.7/resources/assets/poweredby_mediawiki_132x47.png 1.5x, /static/1.27.0-wmf.7/resources/assets/poweredby_mediawiki_176x62.png 2x\" height=\"31\" width=\"88\"></a>            </li>\r\n                  </ul>\r\n            <div style=\"clear:both\"></div>\r\n    </div>\r\n";

var BrowserOSXFirefoxPlainTextFixture = "<html><head><meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"></head><body>XXX</body></html>\r\n";

var BrowserOSXFirefoxAnnotatedTextFixture = "<html><head><meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"></head><body>X<a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" title=\"United States Senate election in Illinois, 2004\">X</a>X</body></html>\r\n";

var BrowserOSXFirefoxTwoParagraphsFixture = "<html><head><meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"></head><body><p>AAA</p><p>BBB</p></body></html>";

var BrowserWindowsFirefoxPlainTextFixture = "<html><body>\r\n<!--StartFragment-->XXX<!--EndFragment-->\r\n</body>\r\n</html>";

var BrowserWindowsFirefoxAnnotatedTextFixture = "<html><body>\r\n<!--StartFragment-->X<a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" title=\"United States Senate election in Illinois, 2004\">X</a>X<!--EndFragment-->\r\n</body>\r\n</html>";

var BrowserWindowsFirefoxTwoParagraphsFixture = "<html><body>\r\n<!--StartFragment--><p>AAA</p><p>BBB</p><!--EndFragment-->\r\n</body>\r\n</html>";

var BrowserWindowsEdgePlainTextFixture = "<HTML class=\"client-js ve-not-available\" lang=\"en\" dir=\"ltr\"><HEAD> <TITLE>Barack Obama - Wikipedia, the free encyclopedia</TITLE>\r\n</HEAD><BODY class=\"mediawiki ltr sitedir-ltr ns-0 ns-subject page-Barack_Obama skin-vector action-view\"><DIV class=\"mw-body\" id=\"content\" role=\"main\"><DIV class=\"mw-body-content\" id=\"bodyContent\"><DIV class=\"mw-content-ltr\" id=\"mw-content-text\" lang=\"en\" dir=\"ltr\"><P><!--StartFragment-->XXX<!--EndFragment--></P></DIV></DIV></DIV></BODY></HTML>";

var BrowserWindowsEdgeAnnotatedTextFixture = "<HTML class=\"client-js ve-not-available\" lang=\"en\" dir=\"ltr\"><HEAD> <TITLE>Barack Obama - Wikipedia, the free encyclopedia</TITLE>\r\n</HEAD><BODY class=\"mediawiki ltr sitedir-ltr ns-0 ns-subject page-Barack_Obama skin-vector action-view\"><DIV class=\"mw-body\" id=\"content\" role=\"main\"><DIV class=\"mw-body-content\" id=\"bodyContent\"><DIV class=\"mw-content-ltr\" id=\"mw-content-text\" lang=\"en\" dir=\"ltr\"><P><!--StartFragment-->X<A title=\"United States Senate election in Illinois, 2004\" href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\">X</A>X<!--EndFragment--></P></DIV></DIV></DIV></BODY></HTML>";

var BrowserWindowsEdgeTwoParagraphsFixture = "<HTML class=\"client-js ve-not-available\" lang=\"en\" dir=\"ltr\"><HEAD> <TITLE>Barack Obama - Wikipedia, the free encyclopedia</TITLE>\r\n</HEAD><BODY class=\"mediawiki ltr sitedir-ltr ns-0 ns-subject page-Barack_Obama skin-vector action-view\"><DIV class=\"mw-body\" id=\"content\" role=\"main\"><DIV class=\"mw-body-content\" id=\"bodyContent\"><DIV class=\"mw-content-ltr\" id=\"mw-content-text\" lang=\"en\" dir=\"ltr\"><!--StartFragment--><P>AAA</P><P>BBB</P><!--EndFragment--></DIV></DIV></DIV></BODY></HTML>";

var GDocsOSXLinuxChromePlainTextFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><meta charset=\"utf-8\"><b style=\"font-weight:normal;\" id=\"docs-internal-guid-5bea85da-43d8-f5c6-8657-4da8de7325c0\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">XXX</span></b>";

var GDocsOSXLinuxChromeAnnotatedTextFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><meta charset=\"utf-8\"><b style=\"font-weight:normal;\" id=\"docs-internal-guid-5bea85da-43db-ac2a-ad43-e477b9e4cca9\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">X</span><a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" style=\"text-decoration:none;\"><span style=\"font-size:14px;font-family:Arial;color:#0b0080;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">X</span></a><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">X</span></b>";

var GDocsOSXLinuxChromeTwoParagraphsFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><meta charset=\"utf-8\"><b style=\"font-weight:normal;\" id=\"docs-internal-guid-5bea85da-43dc-fb06-e327-00c1c6576cf7\"><p dir=\"ltr\" style=\"line-height:1.8327272415161155;margin-top:6pt;margin-bottom:6pt;\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">AAA</span></p><p dir=\"ltr\" style=\"line-height:1.8327272415161155;margin-top:6pt;margin-bottom:6pt;\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">BBB</span></p></b><br class=\"Apple-interchange-newline\">";

var GDocsLinuxFirefoxPlainTextFixture = "<meta charset=\"utf-8\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;\" id=\"docs-internal-guid-af43648c-43e0-a996-6b08-68384b28b868\">XXX</span>";

var GDocsLinuxFirefoxAnnotatedTextFixture = "<meta charset=\"utf-8\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;\" id=\"docs-internal-guid-af43648c-43e2-b6e6-7296-c5f0e1288023\">X</span><a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" style=\"text-decoration:none;\"><span style=\"font-size:14px;font-family:Arial;color:#0b0080;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;\">X</span></a><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;\">X</span>";

var GDocsOSXFirefoxPlainTextFixture = "<html><head><meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"></head><body><meta charset=\"utf-8\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;\" id=\"docs-internal-guid-e0b4b2ba-4421-0487-8eea-908ebdaae584\">XXX</span></body></html>";

var LibreOfficeOSXPlainTextFixture = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\r\n<html>\r\n<head>\r\n  <meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"/>\r\n  <title></title>\r\n  <meta name=\"generator\" content=\"LibreOffice 5.0.3.2 (MacOSX)\"/>\r\n  <style type=\"text/css\">\r\n    @page { margin: 0.79in }\r\n    p { margin-bottom: 0.08in }\r\n    a:link { so-language: zxx }\r\n  </style>\r\n</head>\r\n<body lang=\"en\" dir=\"ltr\">\r\n<p style=\"margin-bottom: 0in\">XXX</p>\r\n</body>\r\n</html>";

var LibreOfficeOSXAnnotatedTextFixture = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\r\n<html>\r\n<head>\r\n  <meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"/>\r\n  <title></title>\r\n  <meta name=\"generator\" content=\"LibreOffice 5.0.3.2 (MacOSX)\"/>\r\n  <style type=\"text/css\">\r\n    @page { margin: 0.79in }\r\n    p { margin-bottom: 0.08in }\r\n    a:link { so-language: zxx }\r\n  </style>\r\n</head>\r\n<body lang=\"en\" dir=\"ltr\">\r\n<p style=\"margin-bottom: 0in\">X<a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\">X</a>X</p>\r\n</body>\r\n</html>";

var LibreOfficeOSXTwoParagraphsFixture = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\r\n<html>\r\n<head>\r\n  <meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"/>\r\n  <title></title>\r\n  <meta name=\"generator\" content=\"LibreOffice 5.0.3.2 (MacOSX)\"/>\r\n  <style type=\"text/css\">\r\n    @page { margin: 0.79in }\r\n    p { margin-bottom: 0.08in }\r\n    a:link { so-language: zxx }\r\n  </style>\r\n</head>\r\n<body lang=\"en\" dir=\"ltr\">\r\n<p style=\"margin-bottom: 0in\">AAA</p>\r\n<p style=\"margin-bottom: 0in\">BBB</p>\r\n</body>\r\n</html>";

var MSW11OSXPlainTextFixture = "<html xmlns:o=\"urn:schemas-microsoft-com:office:office\"\r\nxmlns:w=\"urn:schemas-microsoft-com:office:word\"\r\nxmlns:m=\"http://schemas.microsoft.com/office/2004/12/omml\"\r\nxmlns=\"http://www.w3.org/TR/REC-html40\">\r\n\r\n<head>\r\n</head>\r\n\r\n<body bgcolor=white lang=EN-US style='tab-interval:36.0pt'>\r\n<!--StartFragment--><span lang=EN style='font-size:12.0pt;font-family:Cambria;\r\nmso-ascii-theme-font:minor-latin;mso-fareast-font-family:\"ＭＳ 明朝\";mso-fareast-theme-font:\r\nminor-fareast;mso-hansi-theme-font:minor-latin;mso-bidi-font-family:\"Times New Roman\";\r\nmso-bidi-theme-font:minor-bidi;mso-ansi-language:EN;mso-fareast-language:EN-US;\r\nmso-bidi-language:AR-SA'>XXX</span><!--EndFragment-->\r\n</body>\r\n\r\n</html>";

var MSW11OSXAnnotatedTextFixture = "<html xmlns:o=\"urn:schemas-microsoft-com:office:office\"\r\nxmlns:w=\"urn:schemas-microsoft-com:office:word\"\r\nxmlns:m=\"http://schemas.microsoft.com/office/2004/12/omml\"\r\nxmlns=\"http://www.w3.org/TR/REC-html40\">\r\n\r\n<head>\r\n</head>\r\n\r\n<body bgcolor=white lang=EN-US link=blue vlink=purple style='tab-interval:36.0pt'>\r\n<!--StartFragment--><span lang=EN style='font-size:12.0pt;font-family:Cambria;\r\nmso-ascii-theme-font:minor-latin;mso-fareast-font-family:\"ＭＳ 明朝\";mso-fareast-theme-font:\r\nminor-fareast;mso-hansi-theme-font:minor-latin;mso-bidi-font-family:\"Times New Roman\";\r\nmso-bidi-theme-font:minor-bidi;mso-ansi-language:EN;mso-fareast-language:EN-US;\r\nmso-bidi-language:AR-SA'>X<a\r\nhref=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\">X</a>X</span><!--EndFragment-->\r\n</body>\r\n\r\n</html>\r\n";

var MSW11OSXTwoParagraphsFixture = "<html xmlns:o=\"urn:schemas-microsoft-com:office:office\"\r\nxmlns:w=\"urn:schemas-microsoft-com:office:word\"\r\nxmlns:m=\"http://schemas.microsoft.com/office/2004/12/omml\"\r\nxmlns=\"http://www.w3.org/TR/REC-html40\">\r\n\r\n<head>\r\n</head>\r\n\r\n<body bgcolor=white lang=EN-US link=blue vlink=purple style='tab-interval:36.0pt'>\r\n<!--StartFragment-->\r\n\r\n<p style='margin-bottom:0cm;margin-bottom:.0001pt'><span lang=EN\r\nstyle='mso-ansi-language:EN'>AAA</p>\r\n\r\n<p style='margin-bottom:0cm;margin-bottom:.0001pt'><span lang=EN\r\nstyle='mso-ansi-language:EN'>BBB</p>\r\n\r\n<!--EndFragment-->\r\n</body>\r\n\r\n</html>";

const ParagraphHTMLConverter = substance.ParagraphPackage.ParagraphHTMLConverter;
const HeadingHTMLConverter = substance.HeadingPackage.HeadingHTMLConverter;
const StrongHTMLConverter = substance.StrongPackage.StrongHTMLConverter;
const EmphasisHTMLConverter = substance.EmphasisPackage.EmphasisHTMLConverter;
const LinkHTMLConverter = substance.LinkPackage.LinkHTMLConverter;
const CodeblockHTMLConverter = substance.CodeblockPackage.CodeblockHTMLConverter;

ClipboardTests();

if (substance.platform.inBrowser) {
  ClipboardTests('memory');
}

function ClipboardTests(memory) {

  const test = substanceTest.module('Clipboard' + (memory ? ' [memory]' : ''), {
    before: function() {
      if (memory) substance.platform.inBrowser = false;
    },
    after: function() {
      substance.platform._reset();
    }
  });

  test("Copying HTML, and plain text", function(t) {
    let { editorSession, clipboard } = _fixture(t, simple);
    editorSession.setSelection({ type: 'property', path: ['p1', 'content'], startOffset: 0, endOffset: 5 });
    let event = new ClipboardEvent();
    clipboard.onCopy(event);
    let clipboardData = event.clipboardData;
    t.notNil(clipboardData.data['text/plain'], "Clipboard should contain plain text data.");
    t.notNil(clipboardData.data['text/html'], "Clipboard should contain HTML data.");
    let htmlDoc = substance.DefaultDOMElement.parseHTML(clipboardData.data['text/html']);
    let body = htmlDoc.find('body');
    t.notNil(body, 'The copied HTML should always be a full HTML document string, containing a body element.');
    t.end();
  });

  test("Copying a property selection", function(t) {
    let { editorSession, clipboard } = _fixture(t, simple);
    editorSession.setSelection({ type: 'property', path: ['p1', 'content'], startOffset: 0, endOffset: 5 });
    let TEXT = '01234';
    let event = new ClipboardEvent();
    clipboard.onCopy(event);

    let clipboardData = event.clipboardData;
    t.equal(clipboardData.data['text/plain'], TEXT, "Plain text should be correct.");

    let htmlDoc = substance.DefaultDOMElement.parseHTML(clipboardData.data['text/html']);
    let body = htmlDoc.find('body');
    t.equal(body.text(), TEXT, "HTML text should be correct.");
    t.end();
  });

  test("Copying a container selection", function(t) {
    let { editorSession, clipboard } = _fixture(t, simple);
    editorSession.setSelection({
      type: 'container',
      containerId: 'body',
      startPath: ['p1', 'content'],
      startOffset: 1,
      endPath: ['p3', 'content'],
      endOffset: 5
    });
    let TEXT = [
      '123456789',
      '0123456789',
      '01234'
    ];

    let event = new ClipboardEvent();
    clipboard.onCopy(event);

    let clipboardData = event.clipboardData;
    t.equal(clipboardData.data['text/plain'], TEXT.join('\n'), "Plain text should be correct.");

    let htmlDoc = substance.DefaultDOMElement.parseHTML(clipboardData.data['text/html']);
    let elements = htmlDoc.find('body').getChildren();
    t.equal(elements.length, 3, "HTML should consist of three elements.");
    let p1 = elements[0];
    t.equal(p1.attr('data-id'), 'p1', "First element should have correct data-id.");
    t.equal(p1.text(), TEXT[0], "First element should have correct text content.");
    let p2 = elements[1];
    t.equal(p2.attr('data-id'), 'p2', "Second element should have correct data-id.");
    t.equal(p2.text(), TEXT[1], "Second element should have correct text content.");
    let p3 = elements[2];
    t.equal(p3.attr('data-id'), 'p3', "Third element should have correct data-id.");
    t.equal(p3.text(), TEXT[2], "Third element should have correct text content.");
    t.end();
  });

  test("Pasting text into ContainerEditor using 'text/plain'.", function(t) {
    let { editorSession, clipboard, doc } = _fixture(t, simple);
    editorSession.setSelection({
      type: 'property',
      path: ['p1', 'content'],
      startOffset: 1,
      containerId: 'body'
    });
    let event = new ClipboardEvent();
    event.clipboardData.setData('text/plain', 'XXX');
    clipboard.onPaste(event);
    t.equal(doc.get(['p1', 'content']), '0XXX123456789', "Plain text should be correct.");
    t.end();
  });

  test("Pasting without any data given.", function(t) {
    let { editorSession, clipboard, doc } = _fixture(t, simple);
    editorSession.setSelection({
      type: 'property',
      path: ['p1', 'content'],
      startOffset: 1,
      containerId: 'body'
    });
    let event = new ClipboardEvent();
    clipboard.onPaste(event);
    t.equal(doc.get(['p1', 'content']), '0123456789', "Text should be still the same.");
    t.end();
  });

  test("Pasting text into ContainerEditor using 'text/html'.", function(t) {
    let { editorSession, clipboard, doc } = _fixture(t, simple);
    editorSession.setSelection({
      type: 'property',
      path: ['p1', 'content'],
      startOffset: 1,
      containerId: 'body'
    });
    let TEXT = 'XXX';
    let event = new ClipboardEvent();
    event.clipboardData.setData('text/plain', TEXT);
    event.clipboardData.setData('text/html', TEXT);
    clipboard.onPaste(event);
    t.equal(doc.get(['p1', 'content']), '0XXX123456789', "Plain text should be correct.");
    t.end();
  });

  // this test revealed #700: the problem was that in source code there where
  // `"` and `'` characters which did not survive the way through HTML correctly
  test("Copy and Pasting source code.", function(t) {
    let { editorSession, clipboard, doc } = _fixture(t, simple);
    let body = doc.get('body');
    let cb = doc.create({
      type: 'codeblock',
      id: 'cb1',
      content: [
        "function hello_world() {",
        "  alert('Hello World!');",
        "}"
      ].join("\n")
    });
    body.show(cb, body.getPosition('p1')+1);
    editorSession.setSelection(doc.createSelection({
      type: 'container',
      startPath: ['p1', 'content'],
      startOffset: 1,
      endPath: ['p2', 'content'],
      endOffset: 1,
      containerId: 'body',
    }));
    let event = new ClipboardEvent();
    clipboard.onCut(event);
    let cb1 = doc.get('cb1');
    t.isNil(cb1, "Codeblock should have been cutted.");
    clipboard.onPaste(event);
    cb1 = doc.get('cb1');
    t.notNil(cb1, "Codeblock should have been pasted.");
    t.deepEqual(cb1.toJSON(), cb.toJSON(), "Codeblock should have been pasted correctly.");
    t.end();
  });

  test("Browser - Chrome (OSX/Linux) - Plain Text", function(t) {
    _plainTextTest(t, BrowserLinuxPLainTextFixture);
  });

  test("Browser - Chrome (OSX/Linux) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserLinuxAnnotatedTextFixture);
  });

  test("Browser - Chrome (OSX/Linux) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserLinuxTwoParagraphsFixture);
  });

  test("Browser - Chrome (Windows) - Plain Text", function(t) {
    _plainTextTest(t, BrowserWindowsPlainTextFixture, 'forceWindows');
  });

  test("Browser - Chrome (Windows) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserWindowsAnnotatedTextFixture, 'forceWindows');
  });

  test("Browser - Chrome (Windows) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserWindowsTwoParagraphsFixture, 'forceWindows');
  });

  test("Browser - Firefox (Linux) - Plain Text", function(t) {
    _plainTextTest(t, BrowserLinuxFirefoxPlainTextFixture);
  });

  test("Browser - Firefox (Linux) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserLinuxFirefoxAnnotatedTextFixture);
  });

  test("Browser - Firefox (Linux) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserLinuxFirefoxTwoParagraphsFixture);
  });

  test("Browser - Firefox (Linux) - Whole Page", function(t) {
    let html = BrowserLinuxFirefoxWholePageFixture;
    _fixtureTest(t, html, function(doc, clipboard) {
      let event = new ClipboardEvent();
      event.clipboardData.setData('text/plain', 'XXX');
      event.clipboardData.setData('text/html', html);
      clipboard.onPaste(event);
      // make sure HTML paste succeeded, by checking against the result of plain text insertion
      t.notOk(doc.get('p1').getText() === '0XXX123456789', "HTML conversion and paste should have been successful (not fall back to plain-text).");
      t.ok(doc.get('body').getLength() > 30, 'There should be a lot of paragraphs');
      t.end();
    });
  });

  test("Browser - Firefox (OSX) - Plain Text", function(t) {
    _plainTextTest(t, BrowserOSXFirefoxPlainTextFixture);
  });

  test("Browser - Firefox (OSX) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserOSXFirefoxAnnotatedTextFixture);
  });

  test("Browser - Firefox (OSX) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserOSXFirefoxTwoParagraphsFixture);
  });

  test("Browser - Firefox (Windows) - Plain Text", function(t) {
    _plainTextTest(t, BrowserWindowsFirefoxPlainTextFixture, 'forceWindows');
  });

  test("Browser - Firefox (Windows) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserWindowsFirefoxAnnotatedTextFixture, 'forceWindows');
  });

  test("Browser - Firefox (Windows) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserWindowsFirefoxTwoParagraphsFixture, 'forceWindows');
  });

  test("Browser - Edge (Windows) - Plain Text", function(t) {
    _plainTextTest(t, BrowserWindowsEdgePlainTextFixture, 'forceWindows');
  });

  test("Browser - Edge (Windows) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserWindowsEdgeAnnotatedTextFixture, 'forceWindows');
  });

  test("Browser - Edge (Windows) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserWindowsEdgeTwoParagraphsFixture, 'forceWindows');
  });

  test("GoogleDocs - Chrome (OSX/Linux) - Plain Text", function(t) {
    _plainTextTest(t, GDocsOSXLinuxChromePlainTextFixture);
  });

  test("GoogleDocs - Chrome (OSX/Linux) - Annotated Text", function(t) {
    _annotatedTextTest(t, GDocsOSXLinuxChromeAnnotatedTextFixture);
  });

  test("GoogleDocs - Chrome (OSX/Linux) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, GDocsOSXLinuxChromeTwoParagraphsFixture);
  });

  test("GoogleDocs - Firefox (Linux) - Plain Text", function(t) {
    _plainTextTest(t, GDocsLinuxFirefoxPlainTextFixture);
  });

  test("GoogleDocs - Firefox (Linux) - Annotated Text", function(t) {
    _annotatedTextTest(t, GDocsLinuxFirefoxAnnotatedTextFixture);
  });

  test("GoogleDocs - Firefox (OSX) - Plain Text", function(t) {
    _plainTextTest(t, GDocsOSXFirefoxPlainTextFixture);
  });

  test("LibreOffice (OSX/Linux) - Plain Text", function(t) {
    _plainTextTest(t, LibreOfficeOSXPlainTextFixture);
  });

  test("LibreOffice (OSX/Linux) - Annotated Text", function(t) {
    _annotatedTextTest(t, LibreOfficeOSXAnnotatedTextFixture);
  });

  test("LibreOffice (OSX/Linux) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, LibreOfficeOSXTwoParagraphsFixture);
  });

  test("Microsoft Word 11 (OSX) - Plain Text", function(t) {
    _plainTextTest(t, MSW11OSXPlainTextFixture);
  });

  test("Microsoft Word 11 (OSX) - Annotated Text", function(t) {
    _annotatedTextTest(t, MSW11OSXAnnotatedTextFixture);
  });

  test("Microsoft Word 11 (OSX) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, MSW11OSXTwoParagraphsFixture);
  });

}


let converterRegistry = new substance.Registry({
  "html": new substance.Registry({
    "paragraph": ParagraphHTMLConverter,
    "heading": HeadingHTMLConverter,
    "strong": StrongHTMLConverter,
    "emphasis": EmphasisHTMLConverter,
    "link": LinkHTMLConverter,
    "codeblock": CodeblockHTMLConverter,
  })
});

let clipboardConfig = {
  converterRegistry: converterRegistry
};

class ClipboardEventData {
  constructor() {
    this.data = {};
  }

  getData(format) {
    return this.data[format]
  }

  setData(format, data) {
    this.data[format] = data;
  }

  get types() {
    return Object.keys(this.data)
  }
}

class ClipboardEvent {
  constructor() {
    this.clipboardData = new ClipboardEventData();
  }
  preventDefault() {}
  stopPropagation() {}
}

function _fixture(t, seed) {
  let { editorSession, doc } = setupEditor(t, seed);
  let clipboard = new substance.Clipboard(editorSession, clipboardConfig);
  return { editorSession, doc, clipboard }
}

function _fixtureTest(t, html, impl, forceWindows) {
  let { editorSession, clipboard, doc } = _fixture(t, simple);
  if (forceWindows) {
    // NOTE: faking 'Windows' mode in importer so that
    // the correct implementation will be used
    clipboard.htmlImporter._isWindows = true;
  }
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    containerId: 'body'
  });
  impl(doc, clipboard);
}

function _plainTextTest(t, html, forceWindows) {
  _fixtureTest(t, html, function(doc, clipboard) {
    let event = new ClipboardEvent();
    event.clipboardData.setData('text/plain', '');
    event.clipboardData.setData('text/html', html);
    clipboard.onPaste(event);
    t.equal(doc.get(['p1', 'content']), '0XXX123456789', "Content should have been pasted correctly.");
    t.end();
  }, forceWindows);
}

function _annotatedTextTest(t, html, forceWindows) {
  _fixtureTest(t, html, function(doc, clipboard) {
    let event = new ClipboardEvent();
    event.clipboardData.setData('text/plain', '');
    event.clipboardData.setData('text/html', html);
    clipboard.onPaste(event);
    t.equal(doc.get(['p1', 'content']), '0XXX123456789', "Content should have been pasted correctly.");
    let annotations = doc.getIndex('annotations').get(['p1', 'content']);
    t.equal(annotations.length, 1, "There should be one annotation on the property now.");
    let anno = annotations[0];
    t.equal(anno.type, 'link', "The annotation should be a link.");
    t.end();
  }, forceWindows);
}

function _twoParagraphsTest(t, html, forceWindows) {
  _fixtureTest(t, html, function(doc, clipboard) {
    let event = new ClipboardEvent();
    event.clipboardData.setData('text/plain', '');
    event.clipboardData.setData('text/html', html);
    clipboard.onPaste(event);
    let body = doc.get('body');
    let p1 = body.getChildAt(0);
    t.equal(p1.content, '0AAA', "First paragraph should be truncated.");
    let p2 = body.getChildAt(1);
    t.equal(p2.content, 'BBB', "Second paragraph should contain 'BBB'.");
    let p3 = body.getChildAt(2);
    t.equal(p3.content, '123456789', "Remainder of original p1 should go into forth paragraph.");
    t.end();
  }, forceWindows);
}

let EXAMPLE_CHANGE = {
  ops: [{some: 'operation'}],
  info: {
    userId: 'testuser'
  }
};

/*
  Util for seeding
*/
function _addChange(store, change) {
  return function(cb) {
    store.addChange('test-doc', change, cb);
  }
}

function testChangeStore(createEmptyChangeStore, test) {

  /*
    Add change
  */

  test('Add a first change', (t) => {
    let changeStore = createEmptyChangeStore();
    changeStore.addChange('test-doc', EXAMPLE_CHANGE, (err, version) => {
      t.notOk(err, 'Should not error');
      t.equal(version, 1, 'Initial version should be one');
      t.end();
    });
  });

  test('Add a second change', (t) => {
    let changeStore = createEmptyChangeStore();
    substance.async.series([
      _addChange(changeStore, EXAMPLE_CHANGE)
    ], () => {
      changeStore.addChange('test-doc', EXAMPLE_CHANGE, (err, version) => {
        t.notOk(err, 'Should not error');
        t.equal(version, 2, 'Incremened version should be two');
        t.end();
      });
    });
  });

  /*
    Get changes
  */

  test('Get all changes', (t) => {
    let changeStore = createEmptyChangeStore();
    substance.async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
      _addChange(changeStore, EXAMPLE_CHANGE)
    ], () => {
      changeStore.getChanges('test-doc', (err, changes, version) => {
        t.equal(changes.length, 2, 'There should be two changes in the db');
        t.equal(version, 2, 'Latest version should be 2');
        t.end();
      });
    });
  });

  test('Get changes since version 1 (2 changes stored)', (t) => {
    let changeStore = createEmptyChangeStore();
    substance.async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
      _addChange(changeStore, EXAMPLE_CHANGE)
    ], () => {
      changeStore.getChanges('test-doc', 1, (err, changes, version) => {
        t.equal(changes.length, 1, 'There should be two changes since version 1');
        t.equal(version, 2, 'Doc version should be 2');
        t.end();
      });
    });
  });

  test('Get changes since version 1 up to version 2 (3 changes stored)', (t) => {
    let changeStore = createEmptyChangeStore();
    substance.async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
      _addChange(changeStore, EXAMPLE_CHANGE),
      _addChange(changeStore, EXAMPLE_CHANGE)
    ], () => {
      changeStore.getChanges('test-doc', 1, 2, (err, changes, version) => {
        t.equal(changes.length, 1, 'There should be two changes since version 1');
        t.equal(version, 3, 'Latest version should be 3');
        t.end();
      });
    });
  });

  test('Should return no changes if sinceVersion = actual version', (t) => {
    let changeStore = createEmptyChangeStore();
    substance.async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
    ], () => {
      changeStore.getChanges('test-doc', 1, (err, changes, version) => {
        t.notOk(err, 'Should not error');
        t.equal(changes.length, 0, 'Should have zero changes');
        t.equal(version, 1, 'Document version should be 1');
        t.end();
      });
    });
  });

  /*
    Get version
  */

  test('Return version of test-doc', function(t) {
    let changeStore = createEmptyChangeStore();
    substance.async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
    ], () => {
      changeStore.getVersion('test-doc', (err, version) => {
        t.notOk(err, 'Should not error');
        t.equal(version, 1, 'Document version should equal 1');
        t.end();
      });
    });
  });

  test("Return version=0 if no changes are found", (t) => {
    let changeStore = createEmptyChangeStore();
    changeStore.getVersion('not-existing-doc', (err, version) => {
      t.notOk(err, 'Should not error');
      t.equal(version, 0, 'Document version should equal 0');
      t.end();
    });
  });

  /*
    Delete changes
  */

  test('Delete all changes', (t) => {
    let changeStore = createEmptyChangeStore();
    substance.async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
      _addChange(changeStore, EXAMPLE_CHANGE)
    ], () => {
      changeStore.deleteChanges('test-doc', function(err, changeCount) {
        t.notOk(err, 'Should not error');
        t.equal(changeCount, 2, 'There should be 2 deleted change');
        changeStore.getChanges('test-doc', (err, changes, version) => {
          t.notOk(err, 'Should not error');
          t.equal(changes.length, 0, 'There should not be changes anymore');
          t.equal(version, 0, 'Document version should be 0');
          t.end();
        });
      });
    });
  });

  test("Delete changes of not existing doc", (t) => {
    let changeStore = createEmptyChangeStore();
    changeStore.deleteChanges('not-existing-doc', (err, changeCount) => {
      t.notOk(err, 'Should not error');
      t.equal(changeCount, 0, 'There should be 0 deleted changes');
      t.end();
    });
  });
}

const test$3 = substanceTest.module('collab/ChangeStore');

function createChangeStore() {
  return new substance.ChangeStore()
}

// Runs the offical backend test suite
testChangeStore(createChangeStore, test$3);

// NOTE: this is essentially a copy of ../../collab/computeSnapshot
function computeSnapshot(jsonDoc, changeset) {
  jsonDoc = substance.cloneDeep(jsonDoc);
  let nodes = jsonDoc.nodes;
  changeset.forEach((change) => {
    change.ops.forEach((opData) => {
      try {
        let op = substance.ObjectOperation.fromJSON(opData);
        op.apply(nodes);
      } catch (err) {
        console.error(err, opData);
      }
    });
  });
  return jsonDoc
}

/*
  Create a changeset

  Allows one or more functions to be passed, which will each represent a change
  in the result.
*/
function createChangeset(doc, fns) {
  if (!doc._isDocument) {
    throw new Error('Illegal arguments')
  }
  if (!Array.isArray(fns)) {
    fns = [ fns ];
  }

  let txDoc = new substance.TransactionDocument(doc);
  let tx = new substance.EditingInterface(txDoc);
  let opCount = 0;
  let changes = [];

  fns.forEach((fn) => {
    fn(tx);
    let opsForChange = txDoc.ops.slice(opCount);
    changes.push(new substance.DocumentChange(opsForChange, {}, {}).toJSON());
    opCount = txDoc.ops.length;
  });
  txDoc.dispose();
  return changes
}

var twoParagraphs = function(tx) {
  var body = tx.get('body');
  tx.create({
    id: 'p1',
    type: 'paragraph',
    content: '0123456789'
  });
  body.show('p1');

  tx.create({
    id: 'p2',
    type: 'paragraph',
    content: '0123456789'
  });
  body.show('p2');
};

const EMPTY_DOC = {nodes: {}};

/*
  Some transforms for paragraph (p1)
*/
function insertText(pos, text) {
  return function(tx) {
    tx.setSelection({ type: 'property', path: ['p1', 'content'], startOffset: pos });
    tx.insertText(text);
  }
}

/*
  'test-doc' with numChanges changes and available snapshots

  @param {Number} numChanges number of available changes
  @param {Number[]} snapshots an array of version numbers
*/
function makeStoresFixture(numChanges, snapshots) {
  // available snapshots e.g. [1,2,5]
  snapshots = snapshots || [];

  let changeFns = [];
  if (numChanges > 0) changeFns.push(twoParagraphs);
  for(var i=1; i < numChanges; i++) {
    changeFns.push(insertText(3, 'foo'));
  }
  let changes = createChangeset(createTestArticle(), changeFns);
  let changeStore = new substance.ChangeStore({
    'test-doc': changes
  });
  let snapshotSeed = {'test-doc': {}};
  snapshots.forEach((version) => {
    snapshotSeed['test-doc'][version] = computeSnapshot(
      EMPTY_DOC,
      changes.slice(0,version)
    );
  });
  let snapshotStore = new substance.SnapshotStore(snapshotSeed);
  return {
    changeStore: changeStore,
    snapshotStore: snapshotStore
  }
}

/* eslint-disable consistent-return */
const test$4 = substanceTest.module('collab/DocumentEngine');

/*
  TODO: Test documentEngine.deleteDocument, which removes clears all changes
        and snapshots
*/

test$4('Should allow creation of a new document', function(t) {
  let documentEngine = _fixture$1(0); // 0 changes, no snapshots
  let initialChange = _getChange(1);

  documentEngine.createDocument('test-doc', initialChange, function(err, newVersion) {
    t.notOk(err, 'Should not error');
    t.equal(newVersion, 1, 'Initial version should be 1');
    documentEngine.getDocument('test-doc', (err, snapshot, headVersion) => {
      t.ok(snapshot, 'Should have a snaphot');
      t.equal(headVersion, 1, 'Head version should be 1');
      t.end();
    });
  });
});

test$4('Should allow adding a new change', function(t) {
  let documentEngine = _fixture$1(2); // 0 changes, no snapshots
  let newChange = _getChange(3);

  documentEngine.addChange('test-doc', newChange, function(err, newVersion) {
    t.notOk(err, 'Should not error');
    t.equal(newVersion, 3, 'New version should be 3');
    documentEngine.getDocument('test-doc', (err, snapshot, headVersion) => {
      t.ok(snapshot, 'Should have a snaphot');
      t.equal(headVersion, 3, 'Latest snapshot version should be 3');
      t.end();
    });
  });
});


test$4('Should be able to retrieve changes', function(t) {
  let documentEngine = _fixture$1(2); // 0 changes, no snapshots
  documentEngine.getChanges('test-doc', function(err, changes) {
    t.equal(changes.length, 2, 'Should have two changes');
    t.end();
  });
});

/*
  'test-doc' with numChanges changes and available snapshots

  @param {Number} numChanges number of available changes
  @param {Number[]} snapshots an array of version numbers
*/
function _fixture$1(numChanges, snapshots) {
  let stores = makeStoresFixture(numChanges, snapshots);
  let documentEngine = new substance.DocumentEngine({
    changeStore: stores.changeStore,
    snapshotStore: stores.snapshotStore,
    snapshotFrequency: 1 // store snapshots for each version
  });
  return documentEngine
}

/*
  Compute a valid change
*/
function _getChange(changeNum) {
  let stores = makeStoresFixture(changeNum);
  return stores.changeStore._changes['test-doc'][changeNum-1]
}

/* eslint-disable consistent-return */
const test$5 = substanceTest.module('collab/SnapshotEngine');

function _fixture$2(numChanges, snapshots) {
  let stores = makeStoresFixture(numChanges, snapshots);
  return new substance.SnapshotEngine(stores)
}

test$5('Should error when getting snapshot that does not exist', function(t) {
  let snapshotEngine = _fixture$2(0); // 0 changes, no snapshots
  snapshotEngine.getSnapshot('test-doc', 1, function(err, snapshot) {
    t.ok(err, 'There should be an error');
    t.notOk(snapshot);
    t.end();
  });
});

test$5('Get existing snapshot (straight)', function(t) {
  let snapshotEngine = _fixture$2(1, [1]); // 1 change, snapshot for v1
  snapshotEngine.getSnapshot('test-doc', 1, function(err, snapshot) {
    t.notOk(err, 'There should be no error');
    t.ok(snapshot);
    t.end();
  });
});

test$5('Get existing snapshot (computed from scratch)', function(t) {
  // SnapshotEngine must compute snapshot on demand
  let snapshotEngine = _fixture$2(2, [1]); // 1 change, snapshot for v1
  snapshotEngine.getSnapshot('test-doc', 2, function(err, snapshot) {
    t.notOk(err, 'There should be no error');
    t.ok(snapshot);
    t.end();
  });
});

test$5('Get existing snapshot (computed)', function(t) {
  // SnapshotEngine takes existing snapshot v1 and fetches remaining
  // changes to compute v2.
  let snapshotEngine = _fixture$2(2, [1]); // 1 change, snapshot for v1
  snapshotEngine.getSnapshot('test-doc', 2, function(err, snapshot) {
    t.notOk(err, 'There should be no error');
    t.ok(snapshot);
    t.end();
  });
});

/*
  Run the functions in the tasks collection in series.

  NOTE: You can not access results of the executed functions
*/
function series(tasks, cb, i) {
  i = i || 0;
  tasks[i](function(err) {
    // Always stop execution on error
    if (err) return cb(err)
    if (i === tasks.length-1) {
      cb(...arguments); // we are done
    } else {
      series(tasks, cb, i + 1);
    }
  });
}

let EXAMPLE_SNAPSHOT = {test: 'test'};

function testSnapshotStore(createEmptySnapshotStore, test) {

  /*
    Save snapshot
  */
  test('Save a snapshot', (t) => {
    let snapshotStore = createEmptySnapshotStore();
    snapshotStore.saveSnapshot('test-doc', 1, EXAMPLE_SNAPSHOT, (err, snapshot) => {
      t.notOk(err, 'should not error');
      t.ok(snapshot, 'stored snapshot entry expected');
      t.end();
    });
  });

  /*
    Get snapshot
  */
  test('Retrieve snapshot for test-doc', (t) => {
    let snapshotStore = createEmptySnapshotStore();

    function _create(cb) {
      snapshotStore.saveSnapshot('test-doc', 3, EXAMPLE_SNAPSHOT, cb);
    }

    function _get(cb) {
      snapshotStore.getSnapshot('test-doc', 3, cb);
    }

    function _verify(err, snapshot) {
      t.notOk(err, 'should not error');
      t.equal(snapshot, EXAMPLE_SNAPSHOT);
      t.end();
    }

    series([_create, _get], _verify);
  });

  /*
    Get versions
  */
  test('Get all available versions for a document', (t) => {
    let snapshotStore = createEmptySnapshotStore();

    function _createV1(cb) {
      snapshotStore.saveSnapshot('test-doc', 1, EXAMPLE_SNAPSHOT, cb);
    }

    function _createV3(cb) {
      snapshotStore.saveSnapshot('test-doc', 3, EXAMPLE_SNAPSHOT, cb);
    }

    function _getVersions(cb) {
      snapshotStore.getVersions('test-doc', cb);
    }

    function _verify(err, versions) {
      t.notOk(err, 'should not error');
      t.deepEqual(versions, ['1', '3']);
      t.end();
    }

    series([_createV1, _createV3, _getVersions], _verify);
  });

  /*
    Delete snapshot
  */
  test('Delete snapshot', (t) => {
    let snapshotStore = createEmptySnapshotStore();

    function _create(cb) {
      snapshotStore.saveSnapshot('test-doc', 4, EXAMPLE_SNAPSHOT, cb);
    }

    function _delete(cb) {
      snapshotStore.deleteSnapshot('test-doc', 4, cb);
    }

    function _verify(err) {
      t.notOk(err, 'should not error');
      snapshotStore.getSnapshot('test-doc', 4, (err, snapshot) => {
        t.notOk(snapshot, 'snapshot should be undefined');
        t.notOk(err, 'should not error');
        t.end();
      });
    }

    series([_create, _delete], _verify);
  });
}

const test$6 = substanceTest.module('collab/SnapshotStore');

function createEmptySnapshotStore() {
  return new substance.SnapshotStore()
}

// Runs the offical backend test suite
testSnapshotStore(createEmptySnapshotStore, test$6);

class TestComponent extends substance.Component {

  constructor(...args) {
    super(...args);
    this._enableSpies();
  }

  _enableSpies() {
    ['didMount','didUpdate','dispose','shouldRerender','render'].forEach((name) => {
      substanceTest.spy(this, name);
    });
  }

  _disableSpies() {
    ['didMount','didUpdate','dispose','shouldRerender','render'].forEach((name) => {
      this[name].restore();
    });
  }
}

TestComponent.create = function(renderFunc, props) {
  const comp = new TestComponent(null, props);
  if (renderFunc) {
    comp.render = renderFunc;
  }
  if (props) {
    comp.setProps(props);
  } else {
    comp.rerender();
  }
  return comp
};

class SimpleComponent extends TestComponent {

  render($$) {
    var el = $$('div').addClass('simple-component');
    if (this.props.children) {
      el.append(this.props.children);
    }
    return el
  }
}

TestComponent.Simple = SimpleComponent;

const Simple = TestComponent.Simple;

// regular rendering using default DOM elements
ComponentTests();

// RenderingEngine in debug mode
ComponentTests('debug');

// in the browser do an extra run on memory DOM elements
if (substance.platform.inBrowser) {
  ComponentTests(false, 'memory');
}

function ComponentTests(debug, memory) {

  const test = substanceTest.module('Component' + (debug ? ' [debug]' : '') + (memory ? ' [memory]' : ''), {
    before: function(t) {
      substance.substanceGlobals.DEBUG_RENDERING = Boolean(debug);
      if (memory) substance.platform.inBrowser = false;
      t._document = substance.DefaultDOMElement.createDocument('html');
    },
    after: function() {
      substance.platform._reset();
    }
  });

  test("Throw error when render method is not returning an element", function(t) {
    class MyComponent extends TestComponent {
      render() {}
    }
    t.throws(function() {
      MyComponent.render();
    }, "Should throw an exception when render does not return an element");
    t.end();
  });

  test("Mounting a component", function(t) {
    // Mounting a detached element
    let doc = t._document.createDocument('html');
    let el = doc.createElement('div');
    let comp = Simple.mount(el);
    t.equal(comp.didMount.callCount, 0, "didMount must not be called when mounting to detached elements");
    // Mounting an attached element
    comp = Simple.mount(doc.firstChild);
    t.equal(comp.didMount.callCount, 1, "didMount should have been called");
    t.end();
  });

  test("Render an HTML element", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div')
    });
    t.equal(comp.el.tagName, 'div', 'Element should be a "div".');
    comp = TestComponent.create(function($$) {
      return $$('span')
    });
    t.equal(comp.el.tagName, 'span', 'Element should be a "span".');
    t.end();
  });

  test("Render an element with attributes", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').attr('data-id', 'foo')
    });
    t.equal(comp.el.attr('data-id'), 'foo', 'Element should be have data-id="foo".');
    t.end();
  });

  test("Render an element with css styles", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').css('width', '100px')
    });
    t.equal(comp.el.css('width'), '100px', 'Element should have a css width of 100px.');
    t.end();
  });

  test("Render an element with classes", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').addClass('test')
    });
    t.ok(comp.el.hasClass('test'), 'Element should have class "test".');
    t.end();
  });

  test("Render an element with value", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('input').attr('type', 'text').val('foo')
    });
    t.equal(comp.el.val(), 'foo', 'Value should be set.');
    t.end();
  });

  test("Render an element with plain text", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').text('foo')
    });
    t.equal(comp.el.textContent, 'foo','textContent should be set.');
    t.end();
  });

  test("Render an element with custom html", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').html('Hello <b>World</b>')
    });
    // ATTENTION: it is important to call find() on the element API
    // not on the Component API, as Component#find will only provide
    // elements which are Component instance.
    let b = comp.el.find('b');
    t.notNil(b, 'Element should have rendered HTML as content.');
    t.equal(b.textContent, 'World','Rendered element should have right content.');
    t.end();
  });

  test("Rendering an element with HTML attributes etc.", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div')
        .addClass('foo')
        .attr('data-id', 'foo')
        .htmlProp('type', 'foo')
    });
    t.equal(comp.el.attr('data-id'), 'foo', 'Element should have data-id="foo".');
    t.ok(comp.el.hasClass('foo'), 'Element should have class "foo".');
    t.equal(comp.el.getProperty('type'), 'foo', 'Element should have type "foo".');
    t.end();
  });

  test("Rendering an input element with value", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('input').attr('type', 'text').val('foo')
    });
    t.equal(comp.el.val(), 'foo', 'Input field should have value "foo".');
    t.end();
  });

  test("Render a component", function(t) {
    let comp = Simple.render();
    t.equal(comp.el.tagName.toLowerCase(), 'div', 'Element should be a "div".');
    t.ok(comp.el.hasClass('simple-component'), 'Element should have class "simple-component".');
    t.end();
  });

  test("Rerender on setProps()", function(t) {
    let comp = Simple.render({ foo: 'bar '});
    comp.shouldRerender.reset();
    comp.render.reset();
    comp.setProps({ foo: 'baz' });
    t.ok(comp.shouldRerender.callCount > 0, "Component should have been asked whether to rerender.");
    t.ok(comp.render.callCount > 0, "Component should have been rerendered.");
    t.end();
  });

  test("Rerendering triggers didUpdate()", function(t) {
    let comp = Simple.render({ foo: 'bar '});
    substanceTest.spy(comp, 'didUpdate');
    comp.rerender();
    t.ok(comp.didUpdate.callCount === 1, "didUpdate() should have been called once.");
    t.end();
  });

  test("Setting props triggers willReceiveProps()", function(t) {
    let comp = Simple.render({ foo: 'bar '});
    substanceTest.spy(comp, 'willReceiveProps');
    comp.setProps({ foo: 'baz' });
    t.ok(comp.willReceiveProps.callCount === 1, "willReceiveProps() should have been called once.");
    t.end();
  });

  test("Rerender on setState()", function(t) {
    let comp = Simple.render();
    comp.shouldRerender.reset();
    comp.render.reset();
    comp.setState({ foo: 'baz' });
    t.ok(comp.shouldRerender.callCount > 0, "Component should have been asked whether to rerender.");
    t.ok(comp.render.callCount > 0, "Component should have been rerendered.");
    t.end();
  });

  test("Setting state triggers willUpdateState()", function(t) {
    let comp = Simple.render();
    substanceTest.spy(comp, 'willUpdateState');
    comp.setState({ foo: 'baz' });
    t.ok(comp.willUpdateState.callCount === 1, "willUpdateState() should have been called once.");
    t.end();
  });

  test("Trigger didUpdate() when state or props have changed even with shouldRerender() = false", function(t) {
    class A extends substance.Component {
      shouldRerender() {
        return false
      }
      render($$) {
        return $$('div')
      }
    }
    let comp = A.render();
    substanceTest.spy(comp, 'didUpdate');
    // component will not rerender but still should trigger didUpdate()
    comp.setProps({foo: 'bar'});
    t.ok(comp.didUpdate.callCount === 1, "comp.didUpdate() should have been called once.");
    comp.didUpdate.reset();
    comp.setState({foo: 'bar'});
    t.ok(comp.didUpdate.callCount === 1, "comp.didUpdate() should have been called once.");
    t.end();
  });

  test("Dependency-Injection", function(t) {
    class Parent extends substance.Component {
      getChildContext() {
        let childContext = {};
        if (this.props.name) {
          childContext[this.props.name] = this.props.name;
        }
        return childContext
      }
      render($$) {
        let el = $$('div');
        // direct child
        el.append($$(Child).ref('a'));
        // indirect child
        el.append($$('div').append(
          $$(Child).ref('b')
        ));
        el.append(
          $$(Wrapper, {
            name:'bar',
            // ingested grandchild
            children: [
              $$(Child).ref('c')
            ]
          })
        );
        return el
      }
    }
    class Child extends substance.Component {
      render($$) {
        return $$('div')
      }
    }
    class Wrapper extends substance.Component {
      render($$) {
        return $$('div').append(this.props.children)
      }
    }

    Wrapper.prototype.getChildContext = Parent.prototype.getChildContext;

    let comp = Parent.render({name: 'foo'});
    let a = comp.refs.a;
    let b = comp.refs.b;
    let c = comp.refs.c;
    t.notNil(a.context.foo, "'a' should have a property 'foo' in its context");
    t.isNil(a.context.bar, ".. but not 'bar'");
    t.notNil(b.context.foo, "'b' should have a property 'foo' in its context");
    t.isNil(b.context.bar, ".. but not 'bar'");
    t.notNil(c.context.foo, "'c' should have a property 'foo' in its context");
    t.notNil(c.context.bar, ".. and also 'bar'");
    t.end();
  });

  /* ##################### Rerendering ##########################*/

  test("Rerendering varying content", function(t) {
    class TestComponent$$1 extends substance.Component {
      getInitialState() {
        return { mode: 0 }
      }
      render($$) {
        let el = $$('div');
        if (this.state.mode === 0) {
          el.append(
            "Foo",
            $$('br')
          );
        } else {
          el.append(
            "Bar",
            $$('span'),
            "Baz",
            $$('br')
          );
        }
        return el
      }
    }
    let comp = TestComponent$$1.render();
    let childNodes = comp.el.getChildNodes();
    t.equal(childNodes.length, 2, '# Component should have two children in mode 0');
    t.ok(childNodes[0].isTextNode(), '__first should be a TextNode');
    t.equal(childNodes[0].textContent, 'Foo', '____with proper text content');
    t.equal(childNodes[1].tagName, 'br', '__and second should be a <br>');

    comp.setState({ mode: 1 });
    childNodes = comp.el.getChildNodes();
    t.equal(childNodes.length, 4, '# Component should have 4 children in mode 1');
    t.ok(childNodes[0].isTextNode(), '__first should be a TextNode');
    t.equal(childNodes[0].textContent, 'Bar', '____with proper text content');
    t.equal(childNodes[1].tagName, 'span', '__second should be <span>');
    t.ok(childNodes[2].isTextNode(), '__third should be a TextNode');
    t.equal(childNodes[2].textContent, 'Baz', '____with proper text content');
    t.equal(childNodes[3].tagName, 'br', '__and last should be a <br>');
    t.end();
  });

  // events are not supported by cheerio
  test("Rendering an element with click handler", function(t) {

    class ClickableComponent extends substance.Component {
      constructor(...args) {
        super(...args);
        this.value = 0;
      }
      render($$) {
        let el = $$('a').append('Click me');
        if (this.props.method === 'instance') {
          el.on('click', this.onClick);
        } else if (this.props.method === 'anonymous') {
          el.on('click', () => {
            this.value += 10;
          });
        }
        return el
      }
      onClick() {
        this.value += 1;
      }
    }

    // first render without a click handler
    let comp = ClickableComponent.render();

    comp.click();
    t.equal(comp.value, 0, 'Handler should not have been triggered');

    comp.value = 0;
    comp.setProps({method: 'instance'});
    comp.click();
    t.equal(comp.value, 1, 'Instance method should have been triggered');
    comp.rerender();
    comp.click();
    t.equal(comp.value, 2, 'Rerendering should not add multiple listeners.');

    comp.value = 0;
    comp.setProps({method: 'anonymous'});
    comp.click();
    t.equal(comp.value, 10, 'Anonymous handler should have been triggered');
    comp.rerender();
    comp.click();
    t.equal(comp.value, 20, 'Rerendering should not add multiple listeners.');
    t.end();
  });

  test("Rendering an element with once-click handler", function(t) {
    class ClickableComponent extends substance.Component {
      constructor(...args) {
        super(...args);
        this.clicks = 0;
      }
      render($$) {
        return $$('a').append('Click me')
          .on('click', this.onClick, this, { once: true })
      }
      onClick() {
        this.clicks += 1;
      }
    }

    let comp = ClickableComponent.render();
    comp.click();
    t.equal(comp.clicks, 1, 'Handler should have been triggered');
    comp.click();
    t.equal(comp.clicks, 1, 'Handler should not have been triggered again');
    t.end();
  });

  /* ##################### Nested Elements/Components ##########################*/

  test("Render children elements", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').addClass('parent')
        .append($$('div').addClass('child1'))
        .append($$('div').addClass('child2'))
    });
    t.equal(comp.el.getChildCount(), 2, 'Component should have two children.');
    t.ok(comp.el.hasClass('parent'), 'Element should have class "parent".');
    t.ok(comp.el.getChildAt(0).hasClass('child1'), 'First child should have class "child1".');
    t.ok(comp.el.getChildAt(1).hasClass('child2'), 'Second child should have class "child2".');
    t.end();
  });

  test("Render children components", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$(Simple, {
          children: ['a']
        }),
        $$(Simple, {
          children: ['b']
        })
      )
    });
    t.equal(comp.getChildCount(), 2, "Component should have two children");
    let first = comp.getChildAt(0);
    let second = comp.getChildAt(1);
    t.ok(first instanceof Simple, 'First child should be a Simple');
    t.equal(first.el.textContent, 'a', '.. and should have text "a".');
    t.ok(second instanceof Simple, 'Second child should be a Simple');
    t.equal(second.el.textContent, 'b', '.. and should have text "b".');
    t.end();
  });

  test("Render grandchildren elements", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$('div').addClass('child').append(
          $$('div').addClass('a'),
          $$('div').addClass('b')
        )
      )
    });
    t.equal(comp.getChildCount(), 1, "Component should have 1 child");
    let child = comp.getChildAt(0);
    t.equal(child.getChildCount(), 2, ".. and two grandchildren");
    let first = child.getChildAt(0);
    let second = child.getChildAt(1);
    t.ok(first.el.hasClass('a'), 'First should have class "a".');
    t.ok(second.el.hasClass('b'), 'Second should have class "b".');
    t.end();
  });


  test("Render nested elements passed via props", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$(Simple, {
          children: [
            $$('div').addClass('a'),
            $$('div').addClass('b')
          ]
        })
      )
    });
    t.equal(comp.getChildCount(), 1, "Component should have 1 child");
    let child = comp.getChildAt(0);
    t.equal(child.getChildCount(), 2, ".. and two grandchildren");
    let first = child.getChildAt(0);
    let second = child.getChildAt(1);
    t.ok(first.el.hasClass('a'), 'First grandchild should have class "a".');
    t.ok(second.el.hasClass('b'), 'Second grandchild should have class "b".');
    t.end();
  });

  test("Call didMount once when mounted", function(t) {
    class Parent extends TestComponent {
      render($$) {
        return $$('div')
          .append($$(Child, {loading: true}).ref('child'))
      }
      didMount() {
        this.refs.child.setProps({ loading: false });
      }
    }
    class Child extends TestComponent {
      render($$) {
        if (this.props.loading) {
          return $$('div').append('Loading...')
        } else {
          return $$('div').append(
            $$(Simple).ref('child')
          )
        }
      }
    }

    let comp = Parent.mount(getMountPoint(t));
    let childComp = comp.refs.child;
    let grandChildComp = childComp.refs.child;
    t.equal(childComp.didMount.callCount, 1, "Child's didMount should have been called.");
    t.notNil(grandChildComp, 'Grandchild should have been rendered');
    // t.equal(grandChildComp.didMount.callCount, 1, "Grandchild's didMount should have been called too.")

    comp.empty();
    t.equal(childComp.dispose.callCount, 1, "Child's dispose should have been called once.");
    t.equal(grandChildComp.dispose.callCount, 1, "Grandchild's dispose should have been called once.");
    t.end();
  });

  test('Propagating properties to nested components', function(t) {
    class ItemComponent extends TestComponent {
      render($$) {
        return $$('div').append(this.props.name)
      }
    }
    class CompositeComponent extends TestComponent {
      render($$) {
        let el = $$('div').addClass('composite-component');
        for (let i = 0; i < this.props.items.length; i++) {
          let item = this.props.items[i];
          el.append($$(ItemComponent, item));
        }
        return el
      }
    }

    let comp = CompositeComponent.render({
      items: [ {name: 'A'}, {name: 'B'} ]
    });
    t.equal(comp.getChildCount(), 2, 'Component should have two children.');
    t.equal(comp.getChildAt(0).el.textContent, 'A', 'First child should have text A');
    t.equal(comp.getChildAt(1).el.textContent, 'B', 'First child should have text B');

    // Now we are gonna set new props
    comp.setProps({
      items: [ {name: 'X'}, {name: 'Y'} ]
    });
    t.equal(comp.getChildCount(), 2, 'Component should have two children.');
    t.equal(comp.getChildAt(0).el.textContent, 'X', 'First child should have text X');
    t.equal(comp.getChildAt(1).el.textContent, 'Y', 'First child should have text Y');
    t.end();
  });

  test("Special nesting situation", function(t) {
    // problem was observed in TOCPanel where components (tocEntry) are ingested via dependency-injection
    // and appended to a 'div' element (tocEntries) which then was ingested into a ScrollPane.
    // The order of _capturing must be determined correctly, i.e. first the ScrollPane needs to
    // be captured, so that the parent of the 'div' element (tocEntries) is known.
    // only then the tocEntry components can be captured.
    class Parent extends TestComponent {
      render($$) {
        let el = $$('div');
        // grandchildren wrapped into a 'div' element
        let grandchildren = $$('div').append(
          $$(GrandChild, { name: 'foo' }).ref('foo'),
          $$(GrandChild, { name: 'bar' }).ref('bar')
        );
        el.append(
          // grandchildren wrapper ingested into Child component
          $$(Child, {
            children: grandchildren
          })
        );
        return el
      }
    }
    class Child extends TestComponent {
      render($$) {
        return $$('div').append(this.props.children)
      }
    }
    class GrandChild extends TestComponent {
      render($$) {
        return $$('div').append(this.props.name)
      }
    }

    let comp = Parent.render();
    let foo = comp.refs.foo;
    let bar = comp.refs.bar;
    t.notNil(foo, "Component should have a ref 'foo'.");
    t.equal(foo.el.textContent, 'foo', "foo should have textContent 'foo'");
    t.notNil(bar, "Component should have a ref 'bar'.");
    t.equal(bar.el.textContent, 'bar', "bar should have textContent 'bar'");
    t.end();
  });

  test("Special nesting situation II", function(t) {
    class Parent extends substance.Component {
      render($$) {
        return $$('div').addClass('parent').append(
          $$(Child, {
            children: [
              $$('div').addClass('grandchild-container').append(
                $$(Grandchild).ref('grandchild')
              )
            ]
          }).ref('child')
        )
      }
    }
    class Child extends substance.Component {
      render($$) {
        let el = $$('div').addClass('child').append(
          this.props.children
        );
        return el
      }
    }
    class Grandchild extends substance.Component {
      render($$) {
        return $$('div').addClass('grandchild')
      }
    }

    let comp = Parent.render();
    let child = comp.refs.child;
    let grandchild = comp.refs.grandchild;
    t.notNil(child, "Child should be referenced.");
    t.notNil(grandchild, "Grandchild should be referenced.");
    comp.rerender();
    t.ok(child === comp.refs.child, "Child should have been retained.");
    t.ok(grandchild === comp.refs.grandchild, "Grandchild should have been retained.");
    t.end();
  });

  // TODO: this test reveals that our rendering algorithm is not able
  // to preserve elements when ref'd components are passed down via props.
  // In such cases, the parent already
  test("Implicit retaining in 3-level nesting situation", function(t) {
    class Parent extends substance.Component {
      render($$) {
        // Ideally, the 'wrapper' element and Child component would be preserved automatically
        // because of the ref'd component 'grandchild' passed via props.
        // However, ATM the rendering algorithm does not 'know' about the existence
        // of the ref'd component when rendering the top-level component.
        // This would be revealed during descent when the Child component
        // is rendered. A chicken egg problem: to decide to preserve the wrapper we need
        // to have it rendered already. We need to rethink the rendering algorithm.
        // For now, we need to ref the component which we pass the ref'd component into.
        return $$('div').addClass('parent').append(
          $$('div').addClass('wrapper').append(
            $$(Child, {
              children: [
                $$(Grandchild).ref('grandchild')
              ]
            })
            // disable the next line to reveal the problem
            .ref('child')
          )
        )
      }
    }
    class Child extends substance.Component {
      render($$) {
        let el = $$('div').addClass('child').append(
          this.props.children
        );
        return el
      }
    }
    class Grandchild extends substance.Component {
      render($$) {
        return $$('div').addClass('grandchild')
      }
    }
    let comp = Parent.render();
    let wrapper = comp.find('.wrapper');
    comp.rerender();
    let wrapper2 = comp.find('.wrapper');
    t.ok(wrapper.el === wrapper2.el, "wrapper element should have been retained.");
    t.end();
  });

  test("Edge case: unused children", function(t) {
    class Parent extends substance.Component {
      render($$) {
        return $$('div').append(
          $$(Child, {
            // TODO: should this element be created at all?
            children: [$$('div').ref('unused')]
          })
        )
      }
    }
    class Child extends substance.Component {
      render($$) {
        return $$('div')
      }
    }
    let comp = Parent.render();
    t.equal(comp.el.getChildCount(), 1, "Should have 1 child");
    t.equal(comp.el.textContent, '', "textContent should be empty");
    t.end();
  });

  test("Providing a ref'd child", function(t) {
    class Parent extends substance.Component {
      render($$) {
        return $$('div').append(
          $$(Child, {
            children: [$$(Grandchild).ref('grandchild')]
          })
        )
      }
    }
    class Child extends substance.Component {
      render($$) {
        return $$('div').append(this.props.children)
      }
    }
    class Grandchild extends substance.Component {
      render($$) {
        return $$('div')
      }
    }
    let parent = Parent.render();
    t.equal(parent.getChildCount(), 1, "Should have 1 child");
    let child = parent.getChildAt(0);
    t.equal(child.getChildCount(), 1, "Should have 1 grandchild");
    let grandchild = child.getChildAt(0);
    t.ok(parent.refs.grandchild === grandchild, "Grandchild should be the same as the referenced component.");
    t.ok(child.props.children[0].getComponent() === grandchild, "Grandchild should be accessible via props of child.");
    t.end();
  });

  test("Implicit retaining should not override higher-level rules", function(t) {
    // If a child component has refs, itself should not be retained without
    // being ref'd by the parent
    class Parent extends substance.Component {
      render($$) {
        // Child is not ref'd: this means the parent is not interested in keeping
        // this instance on rerender
        return $$('div').addClass('parent').append($$(Child))
      }
    }
    class Child extends substance.Component {
      render($$) {
        // 'foo' is ref'd, so it should be retained when rerendering on this level
        let el = $$('div').addClass('child').append(
          $$('div').addClass('foo').ref('foo')
        );
        return el
      }
    }
    let comp = Parent.render();
    let child = comp.find('.child');
    t.notNil(child, "Child should exist.");
    let foo = child.refs.foo;
    child.rerender();
    t.ok(child.refs.foo === foo, "'foo' should have been retained.");
    comp.rerender();
    let child2 = comp.find('.child');
    t.ok(child !== child2, "Child should have been renewed.");
    t.ok(foo !== child2.refs.foo, "'foo' should be different as well.");
    t.end();
  });

  test("Eventlisteners on child element", function(t) {
    class Parent extends substance.Component {
      render($$) {
        return $$('div').append($$(Child).ref('child'))
      }
    }
    class Child extends substance.Component {
      constructor(...args) {
        super(...args);
        this.clicks = 0;
      }
      render($$) {
        return $$('a').append('Click me').on('click', this.onClick)
      }
      onClick() {
        this.clicks++;
      }
    }

    let comp = Parent.render();
    let child = comp.refs.child;
    child.click();
    t.equal(child.clicks, 1, 'Handler should have been triggered');
    comp.rerender();
    child.clicks = 0;
    child.click();
    t.equal(child.clicks, 1, 'Handler should have been triggered');
    t.end();
  });


  /* ##################### Refs: Preserving Components ##########################*/

  test("Children without a ref are not retained", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$(Simple)
      )
    });
    let child = comp.getChildAt(0);
    comp.rerender();
    let newChild = comp.getChildAt(0);
    // as we did not apply a ref, the component should get rerendered from scratch
    t.ok(newChild !== child, 'Child component should have been renewed.');
    t.ok(newChild.el !== child.el, 'Child element should have been renewed.');
    t.end();
  });

  test("A ref must be unique in owner scope (fail on inadvertent reuse)", function(t) {
    class MyComponent extends TestComponent {
      render($$) {
        return $$('div')
          .append($$('div').ref('foo'))
          .append($$('div').ref('foo'))
      }
    }
    t.throws(function() {
      MyComponent.render();
    }, "Should throw an exception when a reference id is used multiple times");
    t.end();
  });

  test("Render a child element with ref", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').addClass('parent')
        .append($$('div').addClass('child').ref('foo'))
    });
    t.notNil(comp.refs.foo, 'Component should have a ref "foo".');
    t.ok(comp.refs.foo.hasClass('child'), 'Referenced component should have class "child".');
    // check that the instance is retained after rerender
    let child = comp.refs.foo;
    let el = child.getNativeElement();
    comp.rerender();
    t.ok(comp.refs.foo === child, 'Child element should have been preserved.');
    t.ok(comp.refs.foo.getNativeElement() === el, 'Native element should be the same.');
    t.end();
  });

  test("Render a child component with ref", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$(Simple).ref('foo')
      )
    });
    let child = comp.refs.foo;
    let el = child.getNativeElement();
    comp.rerender();
    t.ok(comp.refs.foo === child, 'Child component should have been preserved.');
    t.ok(comp.refs.foo.getNativeElement() === el, 'Native element should be the same.');
    t.end();
  });

  test("Rerendering a child component with ref triggers didUpdate()", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$(Simple).ref('foo')
      )
    });
    let child = comp.refs.foo;
    substanceTest.spy(child, 'didUpdate');
    comp.rerender();
    t.ok(child.didUpdate.callCount === 1, "child.didUpdate() should have been called once.");
    t.end();
  });

  test("Trigger didUpdate() on children even when shouldRerender()=false", function(t) {
    class Child extends substance.Component {
      shouldRerender() {
        return false
      }
    }
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        // change prop randomly
        $$(Child, {foo: Date.now()}).ref('foo')
      )
    });
    let child = comp.refs.foo;
    substanceTest.spy(child, 'didUpdate');
    comp.rerender();
    t.ok(child.didUpdate.callCount === 1, "child.didUpdate() should have been called once.");
    t.end();
  });

  test("didUpdate() provides old props and old state", function(t) {
    let oldProps = null;
    let oldState = null;
    class MyComponent extends substance.Component {
      getInitialState() {
        return {
          val: 1
        }
      }
      didUpdate(_oldProps, _oldState) {
        oldProps = _oldProps;
        oldState = _oldState;
      }
    }
    let comp = MyComponent.mount({
      val: 'a'
    }, getMountPoint(t));
    comp.setState({ val: 2 });
    t.notNil(oldProps, 'old props should have been provided');
    t.equal(oldProps.val, 'a', 'old props should contain original value');
    t.notNil(oldState, 'old state should have been provided');
    t.equal(oldState.val, 1, 'old state should contain original value');
    oldProps = null;
    oldState = null;
    comp.setProps({ val: 'b' });
    t.notNil(oldProps, 'old props should have been provided');
    t.equal(oldProps.val, 'a', 'old props should contain original value');
    t.notNil(oldState, 'old state should have been provided');
    t.equal(oldState.val, 2, 'old state should contain original value');
    t.end();
  });

  test("Refs on grandchild elements.", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$('div').append(
          $$('div').ref(this.props.grandChildRef) // eslint-disable-line no-invalid-this
        )
      )
    }, { grandChildRef: "foo"});

    t.notNil(comp.refs.foo, "Ref 'foo' should be set.");
    let foo = comp.refs.foo;
    comp.rerender();
    t.ok(foo === comp.refs.foo, "Referenced grandchild should have been retained.");
    substanceTest.spy(foo, 'dispose');
    comp.setProps({ grandChildRef: "bar" });
    t.ok(foo.dispose.callCount > 0, "Former grandchild should have been disposed.");
    t.notNil(comp.refs.bar, "Ref 'bar' should be set.");
    t.ok(foo !== comp.refs.bar, "Grandchild should have been recreated.");
    t.end();
  });

  // it happened, that a grandchild component with ref was not preserved
  test("Ref on grandchild component.", function(t) {
    class Grandchild extends TestComponent {
      render($$) {
        return $$('div').append(this.props.foo)
      }
    }

    let comp = TestComponent.create(function($$) {
      let el = $$('div');
      el.append(
        $$('div').append(
          // generating a random property making sure the grandchild gets rerendered
          $$(Grandchild, { foo: String(Date.now()) }).ref('grandchild')
        )
      );
      return el
    });
    t.notNil(comp.refs.grandchild, "Ref 'grandchild' should be set.");
    let grandchild = comp.refs.grandchild;
    comp.rerender();
    t.notNil(comp.refs.grandchild, "Ref 'grandchild' should be set.");
    t.ok(comp.refs.grandchild === grandchild, "'grandchild' should be the same");
    t.end();
  });

  test("Retain refs owned by parent but nested in child component.", function(t) {
    // Note: the child component does not know that there is a ref
    // set by the parent. Still, the component should be retained on rerender
    class Child extends TestComponent {
      render($$) {
        return $$('div').append(
          $$('div').append(this.props.children)
        )
      }
    }

    let comp = TestComponent.create(function($$) {
      let el = $$('div');
      el.append(
        $$('div').append(
          // generating a random property making sure the grandchild gets rerendered
          $$(Child).ref('child').append(
            $$('div').append('foo').ref('grandchild')
          )
        )
      );
      return el
    });
    t.notNil(comp.refs.grandchild, "Ref 'grandchild' should be set.");
    let grandchild = comp.refs.grandchild;
    comp.rerender();
    t.ok(comp.refs.grandchild === grandchild, "'grandchild' should be the same");
    t.end();
  });

  test("Should wipe a referenced component when class changes", function(t) {
    class ComponentA extends TestComponent {
      render($$) {
        return $$('div').addClass('component-a')
      }
    }
    class ComponentB extends TestComponent {
      render($$) {
        return $$('div').addClass('component-b')
      }
    }
    class MainComponent extends TestComponent {
      render($$) {
        let el = $$('div').addClass('context');
        let ComponentClass;
        if (this.props.context ==='A') {
          ComponentClass = ComponentA;
        } else {
          ComponentClass = ComponentB;
        }
        el.append($$(ComponentClass).ref('context'));
        return el
      }
    }

    let comp = MainComponent.render({context: 'A'});
    t.ok(comp.refs.context instanceof ComponentA, 'Context should be of instance ComponentA');
    comp.setProps({context: 'B'});
    t.ok(comp.refs.context instanceof ComponentB, 'Context should be of instance ComponentB');
    t.end();
  });

  test('Should store refs always on owners', function(t) {
    class MyComponent extends TestComponent {
      render($$) {
        return $$('div').append(
          $$(Simple).append(
            $$('div').ref('helloComp')
          ).ref('simpleComp')
        )
      }
    }
    let comp = MyComponent.render(MyComponent);
    t.ok(comp.refs.helloComp, 'There should stil be a ref to the helloComp element/component');
    t.end();
  });

  test("Implicitly retain elements when grandchild elements have refs.", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$('div').append(
          $$('div').ref(this.props.grandChildRef) // eslint-disable-line no-invalid-this
        )
      )
    }, { grandChildRef: "foo"});

    let child = comp.getChildAt(0);
    comp.rerender();
    t.ok(child === comp.getChildAt(0), "Child should be retained.");
    t.ok(child.el === comp.getChildAt(0).el, "Child element should be retained.");
    t.end();
  });

  test("Implicitly retain elements when passing grandchild with ref.", function(t) {
    class Child extends TestComponent {
      render($$) {
        return $$('div').append(
          $$('div').append(this.props.children)
        )
      }
    }

    let comp = TestComponent.create(function($$) {
      let grandchild = $$('div').ref('grandchild');
      return $$('div').append(
        $$(Child).append(grandchild)
      )
    });

    let child = comp.getChildAt(0);
    comp.rerender();
    t.ok(child === comp.getChildAt(0), "Child should be retained.");
    t.ok(child.el === comp.getChildAt(0).el, "Child element should be retained.");
    t.end();
  });

  // In ScrollPane we provied a link into the Srollbar which accesses
  // ScrollPanes ref in didUpdate()
  // This is working fine when didUpdate() is called at the right time,
  // i.e., when ScrollPane has been rendered already
  test("Everthing should be rendered when didUpdate() is triggered.", function(t) {
    let parentIsUpdated = false;
    class Parent extends substance.Component {
      render($$) {
        return $$('div').append(
          $$(Child, {parent: this}).ref('child')
        )
      }
    }
    class Child extends substance.Component {
      render($$) {
        return $$('div')
      }
      didUpdate() {
        if (this.props.parent.el) {
          parentIsUpdated = true;
        }
      }
    }

    let comp = Parent.render();
    // didUpdate() should not have been called (no rerender)
    t.notOk(parentIsUpdated, 'Initially child.didUpdate() should not have been called.');
    comp.rerender();
    t.ok(parentIsUpdated, 'After rerender child.didUpdate() should have access to parent.el');
    t.end();
  });

  /* ##################### Integration tests / Issues ##########################*/

  test('Preserve components when ref matches and rerender when props changed', function(t) {

    class ItemComponent extends TestComponent {
      shouldRerender(nextProps) {
        return !substance.isEqual(nextProps, this.props)
      }
      render($$) {
        return $$('div').append(this.props.name)
      }
    }
    class CompositeComponent extends TestComponent {
      render($$) {
        let el = $$('div').addClass('composite-component');
        this.props.items.forEach(function(item) {
          el.append($$(ItemComponent, {name: item.name}).ref(item.ref));
        });
        return el
      }
    }

    // Initial mount
    let comp = CompositeComponent.render({
      items: [
        {ref: 'a', name: 'A'},
        {ref: 'b', name: 'B'},
        {ref: 'c', name: 'C'}
      ]
    });

    let a = comp.refs.a;
    let b = comp.refs.b;
    let c = comp.refs.c;

    let childNodes = comp.childNodes;
    t.equal(childNodes.length, 3, 'Component should have 3 children.');
    t.equal(childNodes[0].textContent, 'A', '.. first child should have text A');
    t.equal(childNodes[1].textContent, 'B', '.. second child should have text B');
    t.equal(childNodes[2].textContent, 'C', '.. third child should have text C');

    comp.refs.a.render.reset();
    comp.refs.b.render.reset();

    // Props update that preserves some of our components, drops some others
    // and adds some new
    comp.setProps({
      items: [
        {ref: 'a', name: 'X'}, // preserved (props changed)
        {ref: 'd', name: 'Y'}, // new
        {ref: 'b', name: 'B'}, // preserved (same props)
        {ref: 'e', name: 'Z'}  // new
      ]
    });

    childNodes = comp.childNodes;
    t.equal(childNodes.length, 4, 'Component should now have 4 children.');
    // a and b should have been preserved
    t.equal(a, comp.refs.a, '.. a should be the same instance');
    t.equal(b, comp.refs.b, '.. b should be the same component instance');
    // c should be gone
    t.equal(c.dispose.callCount, 1, '.. c should have been unmounted');
    // a should have been rerendered (different props) while b should not (same props)
    t.ok(a.render.callCount > 0, '.. Component a should have been rerendered');
    t.equal(b.render.callCount, 0, '.. Component b should not have been rerendered');
    // check content
    t.equal(childNodes[0].textContent, 'X', '.. first child should have text X');
    t.equal(childNodes[1].textContent, 'Y', '.. second child should have text Y');
    t.equal(childNodes[2].textContent, 'B', '.. third child should have text Y');
    t.equal(childNodes[3].textContent, 'Z', '.. fourth child should have text Z');
    t.end();
  });

  // Note: this is more of an integration test, but I did not manage to isolate the error
  // maybe the solution gets us closer to what actually went wrong.
  // TODO: try to split into useful smaller pieces.
  test("Unspecific integration test:  ref'd component must be retained", function(t) {
    class ComponentWithRefs extends substance.Component {
      getInitialState() {
        return {contextId: 'hello'}
      }
      render($$) {
        let el = $$('div').addClass('lc-lens lc-writer sc-controller');

        let workspace = $$('div').ref('workspace').addClass('le-workspace');

        workspace.append(
          // Main (left column)
          $$('div').ref('main').addClass("le-main").append(
            $$(Simple).ref('toolbar').append($$(Simple)),

            $$(Simple).ref('contentPanel').append(
              $$(Simple).ref('coverEditor'),

              // The full fledged document (ContainerEditor)
              $$("div").ref('content').addClass('document-content').append(
                $$(Simple, {
                }).ref('mainEditor')
              ),
              $$(Simple).ref('bib')
            )
          )
        );

        // Context section (right column)
        workspace.append(
          $$(Simple, {
          }).ref(this.state.contextId)
        );

        el.append(workspace);

        // Status bar
        el.append(
          $$(Simple, {}).ref('statusBar')
        );
        return el
      }
    }

    let comp = ComponentWithRefs.render();
    t.ok(comp.refs.contentPanel, 'There should be a ref to the contentPanel component');
    comp.setState({contextId: 'foo'});
    t.ok(comp.refs.contentPanel, 'There should stil be a ref to the contentPanel component');
    comp.setState({contextId: 'bar'});
    t.ok(comp.refs.contentPanel, 'There should stil be a ref to the contentPanel component');
    comp.setState({contextId: 'baz'});
    t.ok(comp.refs.contentPanel, 'There should stil be a ref to the contentPanel component');
    t.end();
  });

  test("#312: refs should be bound to the owner, not to the parent.", function(t) {
    class Child extends TestComponent {
      render($$) {
        return $$('div').append(this.props.children)
      }
    }
    class Parent extends TestComponent {
      render($$) {
        let el = $$('div');
        el.append(
          $$(Child).append(
            $$('div').ref('foo').append('foo')
          )
        );
        return el
      }
    }
    let comp = Parent.render();
    t.notNil(comp.refs.foo, 'Ref should be bound to owner.');
    t.equal(comp.refs.foo.text(), 'foo', 'Ref should point to the right component.');
    t.end();
  });

  test('#635: Relocating a preserved component', function(t) {
    class Parent extends TestComponent {
      render($$) {
        let el = $$('div');
        el.append('X');
        if (this.props.nested) {
          el.append(
            $$('div').append(
              $$(Simple).ref('foo').append('Y')
            )
          );
        } else {
          el.append(
            $$(Simple).ref('foo').append('Y')
          );
        }
        el.append('Z');
        return el
      }
    }
    let comp = Parent.render();
    t.equal(comp.refs.foo.getParent(), comp, "First 'foo' should be direct child.");
    t.equal(comp.el.textContent, 'XYZ', "... and content should be correct.");
    comp.setProps({ nested: true });
    t.equal(comp.refs.foo.getParent().getParent(), comp, "Then 'foo' should be grand-child.");
    t.equal(comp.el.textContent, 'XYZ', "... and content should be correct.");
    comp.setProps({});
    t.equal(comp.refs.foo.getParent(), comp, "At last 'foo' should be direct child again.");
    t.equal(comp.el.textContent, 'XYZ', "... and content should be correct.");
    t.end();
  });

  test("Combine props and children via append", function(t) {
    class Toolbar extends TestComponent {
      render($$) {
        let el = $$('div').append(
          $$(Simple, this.props.strong).append('Strong')
        );
        return el
      }
    }

    let toolState = {strong: {active: true}};
    Toolbar.render(toolState);
    // the original toolState object should not have been changed
    t.ok(substance.isEqual(toolState, {strong: {active: true}}), "props object should not have been touched");
    t.end();
  });

  test("Pass-through props and add children via append", function(t) {
    class MyComponent extends TestComponent {
      render($$) {
        let el = $$('div').append(
          $$(Simple, this.props).append('Child 1')
        );
        return el
      }
    }
    let props = {foo: 'bar'};
    let comp = MyComponent.render(props);
    let simple = comp.getChildAt(0);
    t.notNil(simple, 'Should have a child component.');
    t.equal(simple.props.foo, props.foo, '.. with props past through');
    t.equal(simple.props.children.length, 1, '.. with props.children having one element');
    t.equal(simple.textContent, 'Child 1', '.. with correct text content');
    t.end();
  });
}

const test$7 = substanceTest.module('ContainerAddress');

test$7("[0,1] is before [1,0]", function(t) {
  var first = new substance.ContainerAddress(0, 1);
  var second = new substance.ContainerAddress(1, 0);
  t.ok(first.isBefore(second, 'strict'));
  t.ok(first.isBefore(second));
  t.end();
});

test$7("[1,0] is not before [0,1]", function(t) {
  var first = new substance.ContainerAddress(1, 0);
  var second = new substance.ContainerAddress(0, 1);
  t.notOk(first.isBefore(second, 'strict'));
  t.notOk(first.isBefore(second));
  t.end();
});

test$7("[0,0] is before [0,1]", function(t) {
  var first = new substance.ContainerAddress(0, 0);
  var second = new substance.ContainerAddress(0, 1);
  t.ok(first.isBefore(second, 'strict'));
  t.ok(first.isBefore(second));
  t.end();
});

test$7("[0,1] is not before [0,0]", function(t) {
  var first = new substance.ContainerAddress(0, 1);
  var second = new substance.ContainerAddress(0, 0);
  t.notOk(first.isBefore(second, 'strict'));
  t.notOk(first.isBefore(second));
  t.end();
});

test$7("[0,1] is not-strictly before [0,1]", function(t) {
  var first = new substance.ContainerAddress(0, 1);
  var second = new substance.ContainerAddress(0, 1);
  t.notOk(first.isBefore(second, 'strict'));
  t.ok(first.isBefore(second));
  t.end();
});

test$7("[0,1] is equal to [0,1]", function(t) {
  var first = new substance.ContainerAddress(0, 1);
  var second = new substance.ContainerAddress(0, 1);
  t.ok(first.isEqual(second));
  t.end();
});

const test$8 = substanceTest.module('ContainerSelection');

test$8("Creating a ContainerSelection", function(t) {
  var sel = new substance.ContainerSelection('body',['p1', 'content'], 1, ['p2', 'content'], 2);
  t.ok(sel.isContainerSelection(), 'Should be a container selection.');
  t.ok(!sel.isNull(), 'Should not be null.');
  t.deepEqual(sel.start.path, ['p1', 'content'], 'startPath should be correct.');
  t.equal(sel.start.offset, 1, 'startOffset should be correct.');
  t.deepEqual(sel.end.path, ['p2', 'content'], 'endPath should be correct.');
  t.equal(sel.end.offset, 2, 'endOffset should be correct.');
  t.ok(!sel.isReverse(), 'Selection should not be reverse');
  t.end();
});

test$8("Collapsed ContainerSelection", function(t) {
  var sel = new substance.ContainerSelection('body', ['p1', 'content'], 1, ['p1', 'content'], 1);
  t.ok(sel.isContainerSelection(), 'Should be a container selection.');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed.');
  t.end();
});

test$8("Reverse ContainerSelection", function(t) {
  var sel = new substance.ContainerSelection('body', ['p1', 'content'], 1, ['p1', 'content'], 3, true);
  t.ok(sel.isReverse(), 'Selection should be reverse.');
  t.end();
});

test$8("isInsideOf: strictly inside other", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 4,
    endPath: ['p3', 'content'],
    endOffset: 5,
  });
  t.ok(sel.isInsideOf(other), 'should be inside');
  t.ok(sel.isInsideOf(other, true), 'should be strictly inside');
  t.end();
});

test$8("isInsideOf: not-strictly inside other", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 5,
  });
  t.ok(sel.isInsideOf(other), 'should be inside');
  t.notOk(sel.isInsideOf(other, true), 'should not be strictly inside');
  t.end();
});

test$8("isInsideOf: inside a PropertySelection", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 4,
    endPath: ['p1', 'content'],
    endOffset: 6,
  });
  var other = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 8,
  });
  t.ok(sel.isInsideOf(other), 'should be inside');
  t.end();
});

test$8("isInsideOf: not inside", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 5,
  });
  // wrapping
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 4,
    endPath: ['p3', 'content'],
    endOffset: 6,
  });
  t.notOk(sel.isInsideOf(other), 'should not be inside');
  // left-boundary not inside
  sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 2,
  });
  t.notOk(sel.isInsideOf(other), 'should not be inside');
  // right-boundary not inside
  sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 7,
  });
  t.notOk(sel.isInsideOf(other), 'should not be inside');

  var nullSel = doc.createSelection(null);
  t.notOk(sel.isInsideOf(nullSel), 'should not be inside null selection');
  t.end();
});

test$8("overlaps with other ContainerSelection", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 4,
    endPath: ['p3', 'content'],
    endOffset: 6,
  });
  // equal
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 4,
    endPath: ['p3', 'content'],
    endOffset: 6,
  });
  t.ok(sel.overlaps(other));
  // inside
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 5,
  });
  t.ok(sel.overlaps(other));
  // contained
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 2,
    endPath: ['p3', 'content'],
    endOffset: 8,
  });
  t.ok(sel.overlaps(other));
  // left
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 2,
    endPath: ['p2', 'content'],
    endOffset: 1,
  });
  t.ok(sel.overlaps(other));
  // right
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 2,
    endPath: ['p3', 'content'],
    endOffset: 8,
  });
  t.ok(sel.overlaps(other));
  t.end();
});


test$8("Collapsing to the left", function(t) {
  var sel = new substance.ContainerSelection('body', ['p1', 'content'], 1, ['p3', 'content'], 3);
  sel = sel.collapse('left');
  t.ok(sel.isCollapsed(), 'should be collapsed');
  t.deepEqual(sel.start.path, ['p1', 'content']);
  t.equal(sel.start.offset, 1);
  t.end();
});

test$8("Collapsing to the right", function(t) {
  var sel = new substance.ContainerSelection('body', ['p1', 'content'], 1, ['p3', 'content'], 3);
  sel = sel.collapse('right');
  t.ok(sel.isCollapsed(), 'should be collapsed');
  t.deepEqual(sel.start.path, ['p3', 'content']);
  t.equal(sel.start.offset, 3);
  t.end();
});

test$8("Expanding: other is inside", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 6,
    endOffset: 8
  });
  var newSel = sel.expand(other);
  t.ok(newSel.equals(sel), "Selection should not have changed.");
  t.end();
});

test$8("Expand: is inside other", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 3,
    endPath: ['p3', 'content'],
    endOffset: 5,
  });
  var newSel = sel.expand(other);
  t.ok(newSel.equals(other));
  t.end();
});

test$8("Expand right", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 6,
  });
  var newSel = sel.expand(other);
  t.ok(newSel.start.equals(sel.start));
  t.ok(newSel.end.equals(other.end));
  t.end();
});

test$8("Expand left", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 2,
  });
  var newSel = sel.expand(other);
  t.ok(newSel.start.equals(other.start));
  t.ok(newSel.end.equals(sel.end));
  t.end();
});

test$8("Expand left with PropertySelection", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  var newSel = sel.expand(other);
  t.ok(newSel.start.equals(other.start));
  t.ok(newSel.end.equals(sel.end));
  t.end();
});

test$8("Expand right with PropertySelection", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'property',
    path: ['p3', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  var newSel = sel.expand(other);
  t.ok(newSel.start.equals(sel.start));
  t.ok(newSel.end.equals(other.end));
  t.end();
});

test$8("Truncate with other ContainerSelection", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  // left side overlapping
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 1,
  });
  var newSel = sel.truncateWith(other);
  t.ok(newSel.start.equals(other.end));
  t.ok(newSel.end.equals(sel.end));
  // right side overlapping
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 8,
  });
  newSel = sel.truncateWith(other);
  t.ok(newSel.start.equals(sel.start));
  t.ok(newSel.end.equals(other.start));
  // equal
  newSel = sel.truncateWith(sel);
  t.ok(newSel.isNull());
  // wrapping
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 8,
  });
  newSel = sel.truncateWith(other);
  t.ok(newSel.isNull());
  // left side aligned
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p2', 'content'],
    endOffset: 1,
  });
  newSel = sel.truncateWith(other);
  t.ok(newSel.start.equals(other.end));
  t.ok(newSel.end.equals(sel.end));
  // right side aligned
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  newSel = sel.truncateWith(other);
  t.ok(newSel.start.equals(sel.start));
  t.ok(newSel.end.equals(other.start));
  t.end();
});

test$8("containsNode: inner node", function(t) {
  var doc = createTestArticle(simple);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 3
  });
  t.ok(sel.containsNode('p2'), 'Should contain p2.');
  t.end();
});

test$8("containsNode: outer nodes", function(t) {
  var doc = createTestArticle(simple);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 3
  });
  t.notOk(sel.containsNode('p1'), 'Should contain p1.');
  t.notOk(sel.containsNode('p4'), 'Should contain p4.');
  t.end();
});

test$8("containsNode: start/end is nodeFragment", function(t) {
  var doc = createTestArticle(simple);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1'],
    startOffset: 0,
    endPath: ['p1'],
    endOffset: 1
  });
  t.ok(sel.containsNode('p1'), 'Should contain node.');
  sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1'],
    startOffset: 0,
    endPath: ['p2'],
    endOffset: 1
  });
  t.ok(sel.containsNode('p2'), 'Should contain node.');
  t.end();
});

test$8("containsNode: with partial node fragment", function(t) {
  var doc = createTestArticle(simple);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1'],
    startOffset: 1,
    endPath: ['p2'],
    endOffset: 1
  });
  t.notOk(sel.containsNode('p1'), 'Should not contain node.');
  sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1'],
    startOffset: 0,
    endPath: ['p2'],
    endOffset: 0
  });
  t.notOk(sel.containsNode('p2'), 'Should not contain node.');
  t.end();
});

if (substance.platform.inBrowser) {
  DOMElementTests('BrowserDOMElement');
}

DOMElementTests('MemoryDOMElement');

function DOMElementTests(impl) {

  const test = substanceTest.module('DefaultDOMElement ('+impl+')', {
    before: () => {
      if (impl === 'MemoryDOMElement') substance.platform.inBrowser = false;
    },
    after: () => {
      substance.platform._reset();
    }
  });

  test("Wrapping a native element", (t) => {
    let doc;
    if (impl === 'BrowserDOMElement') {
      doc = window.document;
    } else {
      doc = substance.MemoryDOMElement.createDocument('html');
    }
    let nativeEl = doc.createElement('div');
    let el = substance.DefaultDOMElement.wrap(nativeEl);
    t.ok(el._isDOMElement, 'native element should be wrapped');
    let el2 = substance.DefaultDOMElement.wrap(nativeEl);
    t.ok(el === el2, 'wrapping an element twice should give the same DOMElement instance');
    t.throws(() => {
      substance.DefaultDOMElement.wrap(undefined);
    }, 'Should throw when trying to wrap nil');
    t.end();
  });

  test("Creating an HTML document", (t) => {
    const doc = substance.DefaultDOMElement.createDocument('html');
    t.notNil(doc.find('head'), 'Document should have a <head>');
    t.notNil(doc.find('body'), '.. and a <body>');
    t.equal(doc.getContentType(), 'text/html', '.. content type should be correct');
    t.end();
  });

  test("Creating an XML document", (t) => {
    const doc = substance.DefaultDOMElement.createDocument('xml');
    t.equal(doc.getContentType(), 'application/xml', 'Document should have correct content type');
    t.equal(doc.children.length, 0, '.. and should be empty');
    t.end();
  });

  test("Parsing a full HTML document", function(t) {
    const html = '<html><head><title>TEST</title></head><body>TEST</body></html>';
    const doc = substance.DefaultDOMElement.parseHTML(html);
    const head = doc.find('head');
    t.notNil(head, '<head> should be there');

    const title = head.find('title');
    t.notNil(title, '<head> should contain <title>');
    t.equal(title.text(), 'TEST', '.. with correct text');

    const body = doc.find('body');
    t.notNil(body, 'document should have a <body> element.');
    t.equal(body.text(), 'TEST', '.. with correct text.');
    t.end();
  });

  test("Parsing one HTML element", function(t) {
    const html = '<p>TEST</p>';
    const p = substance.DefaultDOMElement.parseSnippet(html, 'html');
    t.notNil(p, 'HTML should get parsed.');
    t.equal(p.tagName, 'p', '.. providing one <p> element,');
    t.equal(p.text(), 'TEST', '.. with correct content.');
    t.end();
  });

  test("Parsing multiple HTML elements", function(t) {
    const html = '<p>TEST</p><p>TEST2</p>';
    const els = substance.DefaultDOMElement.parseSnippet(html, 'html');
    t.notNil(els, 'HTML should get parsed.');
    t.equal(els.length, 2, '.. Providing 2 elements');
    t.equal(els[0].tagName, 'p', '.. the first a <p>');
    t.equal(els[0].text(), 'TEST', '.. with correct content');
    t.equal(els[1].tagName, 'p', '.. the second a <p>');
    t.equal(els[1].text(), 'TEST2', '.. with correct content');
    t.end();
  });

  test("Parsing annotated HTML text", function(t) {
    const html = '123<b>456</b>789';
    const els = substance.DefaultDOMElement.parseSnippet(html, 'html');
    t.equal(els.length, 3, 'there are three elements');
    t.equal(els[0].nodeType, 'text', 'first is a text node');
    t.equal(els[0].text(), '123', '... it has correct content');
    t.equal(els[1].nodeType, 'element', 'second is an element');
    t.equal(els[1].tagName, 'b', '... it is a <b>');
    t.equal(els[1].text(), '456', '... it has correct content');
    t.equal(els[2].nodeType, 'text', 'third is a text node again');
    t.equal(els[2].text(), '789', '... it has correct content');
    t.end();
  });

  test("Parsing an XML document", function(t) {
    const xml = "<mydoc><myhead><mytitle>TEST</mytitle></myhead><mybody>TEST</mybody></mydoc>";
    const doc = substance.DefaultDOMElement.parseXML(xml);
    const head = doc.find('myhead');
    t.notNil(head, '<myhead> should be there');
    const title = head.find('mytitle');
    t.notNil(title, '<head> should contain <title>');
    t.equal(title.text(), 'TEST', '.. with correct text');
    const body = doc.find('mybody');
    t.notNil(body, 'document should have a <body> element');
    t.equal(body.text(), 'TEST', '.. with correct text');
    t.end();
  });

  test("Parsing an incomplete HTML document", (t) => {
    const doc = substance.DefaultDOMElement.parseHTML('<title>Foo</title><p>TEST</p>');
    const head = doc.find('head');
    const title = head.find('title');
    const body = doc.find('body');
    t.notNil(head, 'document should have a <head>');
    t.notNil(title, '<head> should contain <title>');
    t.equal(title.textContent, 'Foo', 'title should be correct');
    t.equal(body.innerHTML, '<p>TEST</p>', 'body should have contain the remaining content');
    t.end();
  });

  test("Parsing an HTML document with comments", (t) => {
    let doc = substance.DefaultDOMElement.parseHTML('<body><!--TEST--></body>');
    let body = doc.find('body');
    let comment = body.firstChild;
    t.notNil(comment, 'there should a child element');
    t.ok(comment.isCommentNode(), '.. which should be a comment');
    t.equal(comment.textContent, 'TEST', '.. with correct content');
    // multiple lines
    doc = substance.DefaultDOMElement.parseHTML('<body><!--\n  TEST\n  Foo\n  Bar--></body>');
    body = doc.find('body');
    comment = body.firstChild;
    t.equal(comment.textContent, '\n  TEST\n  Foo\n  Bar', 'multi-line comment should work too');
    /*
      NOTE: from looking at the default implementation in htmlparser2 it seems that they would merge subsequent comments
        but actually this is not the case. So, there is a branch in oncomment() which I do not understand
        The same with TextNodes.
        ```
        if(lastTag && lastTag.type === ElementType.Comment){
          lastTag.data += data
        }
        ```
    */
    // doc = DefaultDOMElement.parseHTML('<body><!--FOO--><!--BAR--></body>')
    // body = doc.find('body')
    // t.deepEqual(body.childNodes.map(c=>c.nodeType), ['comment','comment'], 'subsequent comments should not be merged')
    t.end();
  });

  test("Parsing an XML document with CDATA", (t) => {
    let foo = substance.DefaultDOMElement.parseSnippet(`<foo><![CDATA[TEST]]></foo>`, 'xml');
    let cdata = foo.firstChild;
    t.notNil(cdata, 'there should a child element');
    t.equal(cdata.nodeType, 'cdata', '.. which should be CDATA');
    t.equal(cdata.textContent, 'TEST', '.. with correct content');
    t.end();
  });

  test("Parsing void and self-closing elements", (t) => {
    t.throws(() => {
      substance.DefaultDOMElement.parseSnippet(`<foo>`, 'xml');
    }, 'should not allow void elements in XML');
    t.doesNotThrow(() => {
      substance.DefaultDOMElement.parseSnippet(`<foo />`, 'xml');
    }, 'should allow self-closing elements in XML');
    t.doesNotThrow(() => {
      substance.DefaultDOMElement.parseSnippet(`<input>`, 'html');
      substance.DefaultDOMElement.parseSnippet(`<input />`, 'html');
    }, 'in HTML there are some allowed void elements, such as <input>');
    t.end();
  });

}

const test$9 = substanceTest.module('Document');

test$9("Create null selection.", function(t) {
  let doc = createTestArticle(simple);
  let sel = doc.createSelection(null);
  t.ok(sel.isNull(), 'Selection should be null.');
  t.end();
});

test$9("Create collapsed property selection.", function(t) {
  let doc = createTestArticle(simple);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3
  });
  t.ok(sel.isPropertySelection(), 'Selection should be a property selection.');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed.');
  t.deepEqual(sel.start.path, ['p1', 'content'], 'path should be correct.');
  t.deepEqual(sel.start.offset, 3, 'start offset should be correct.');
  t.end();
});

test$9("Create expanded property selection.", function(t) {
  let doc = createTestArticle(simple);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 4
  });
  t.ok(sel.isPropertySelection(), 'Selection should be a property selection.');
  t.notOk(sel.isCollapsed(), 'Selection should not be collapsed.');
  t.deepEqual(sel.start.path, ['p1', 'content'], 'path should be correct.');
  t.deepEqual(sel.start.offset, 1, 'start offset should be correct.');
  t.deepEqual(sel.end.offset, 4, 'end offset should be correct.');
  t.end();
});

test$9("Node.toJSON() should not export undefined optional properties", function(t) {
  let doc = createTestArticle(simple);
  let p = doc.create({
    type: 'paragraph',
    id: 'p',
    content: ''
  });
  t.deepEqual(p.toJSON(), { type: 'paragraph', id: 'p', content: '' }, 'JSON should be correct.');
  t.end();
});

const test$10 = substanceTest.module('DocumentChange');

test$10("hasUpdated()", (t) => {
  let tx = setup(simple);
  let path = ['p1', 'content'];
  // updating a property
  tx.setSelection({
    type: 'property',
    path,
    startOffset: 1
  });
  tx.insertText('X');
  let change = tx.generateChange();
  t.ok(change.hasUpdated(path), 'should tell that property has been updated');
  t.ok(change.hasUpdated(path[0]), 'should tell that node has been updated');
  // deleting a node
  tx.delete('p2');
  change = tx.generateChange();
  t.ok(change.hasDeleted('p2'), 'should tell that node has been deleted');
  // update but then delete a node
  tx.setSelection({
    type: 'property',
    path,
    startOffset: 1
  });
  tx.insertText('X');
  tx.delete('p1');
  change = tx.generateChange();
  t.notOk(change.hasUpdated(path), 'should not tell that property has been updated');
  t.ok(change.hasDeleted('p1'), 'should tell that node has been deleted');
  t.end();
});

test$10("serialize() and deserialize()", function(t) {
  let tx = setup(simple);
  let path = ['p1', 'content'];
  // updating a property
  let selStart = tx.setSelection({
    type: 'property',
    path,
    startOffset: 1
  });
  tx.insertText('X');
  let change = tx.generateChange();
  let selEnd = tx.selection;
  change.before.selection = selStart;
  change.after.selection = selEnd;

  let serialized = change.serialize();
  let deserialized = substance.DocumentChange.deserialize(serialized);

  t.deepEqual(deserialized.ops, change.ops, 'ops should be equal after deserialization');
  t.ok(selStart.equals(deserialized.before.selection), 'selection (before) should have been deserialized too');
  t.ok(selEnd.equals(deserialized.after.selection), 'selection (after) should have been deserialized too');
  t.end();
});

function setup(...seeds) {
  let config = getTestConfig();
  let doc = new substance.Document(config.getSchema());
  doc.create({ id: 'body', type: 'container' });
  seeds.forEach(s=>s(doc, doc.get('body')));
  return new substance.ChangeRecorder(doc)
}

const H1_TEXT = 'h1:abcdefg';

function _h1(doc, body) {
  doc.create({
    type: 'heading',
    id: 'h1',
    content: H1_TEXT,
    level: 1
  });
  body.show('h1');
}





const P1_TEXT = 'p1:abcdef';

function _p1(doc, body) {
  doc.create({
    type: 'paragraph',
    id: 'p1',
    content: P1_TEXT
  });
  body.show('p1');
}

const P2_TEXT = 'p2:ghijk';

function _p2(doc, body) {
  doc.create({
    type: 'paragraph',
    id: 'p2',
    content: P2_TEXT
  });
  body.show('p2');
}

function _empty(doc, body) {
  doc.create({
    type: 'paragraph',
    id: 'empty',
    content: ''
  });
  body.show('empty');
}

function _s1(doc) {
  doc.create({
    type: 'strong',
    id: 's1',
    start: {
      path: ['p1', 'content'],
      offset: 3,
    },
    end: {
      offset: 5
    }
  });
}

function _il1(doc) {
  let tx = new substance.EditingInterface(doc);
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  tx.insertInlineNode({
    type: 'test-inline-node',
    id: 'il1',
    content: 'X'
  });
}

const LI1_TEXT = 'l1-1:abcdef';
const LI2_TEXT = 'l1-2:0123456';
const LI3_TEXT = 'l1-3:ghij';

// list with two items
function _l1(doc, body) {
  doc.create({
    type: 'list',
    id: 'l1'
  });
  body.show('l1');
}

function _l11(doc) {
  const l1 = doc.get('l1');
  doc.create({
    type: 'list-item',
    id: 'l1-1',
    content: LI1_TEXT
  });
  l1.appendItem('l1-1');
}

function _l12(doc) {
  const l1 = doc.get('l1');
  doc.create({
    type: 'list-item',
    id: 'l1-2',
    content: LI2_TEXT
  });
  l1.appendItem('l1-2');
}

function _l13(doc) {
  const l1 = doc.get('l1');
  doc.create({
    type: 'list-item',
    id: 'l1-3',
    content: LI3_TEXT
  });
  l1.appendItem('l1-3');
}

function _l1_empty(doc) {
  const l1 = doc.get('l1');
  doc.create({
    type: 'list-item',
    id: 'l1-empty',
    content: ''
  });
  l1.appendItem('l1-empty');
}

function _li1plus(doc) {
  doc.set(['l1-1', 'level'], 2);
}

const LI21_TEXT = 'l2-1:abcdef';
const LI22_TEXT = 'l2-2:0123456';

function _l2(doc, body) {
  doc.create({
    type: 'list',
    id: 'l2'
  });
  body.show('l2');
}

function _l21(doc) {
  const l2 = doc.get('l2');
  doc.create({
    type: 'list-item',
    id: 'l2-1',
    content: LI21_TEXT
  });
  l2.appendItem('l2-1');
}

function _l22(doc) {
  const l2 = doc.get('l2');
  doc.create({
    type: 'list-item',
    id: 'l2-2',
    content: LI22_TEXT
  });
  l2.appendItem('l2-2');
}



function _block1(doc, body) {
  doc.create({
    type: 'test-block',
    id: 'block1'
  });
  body.show('block1');
}

function _block2(doc, body) {
  doc.create({
    type: 'test-block',
    id: 'block2'
  });
  body.show('block2');
}

const IN1_TITLE = 'TITLE';
const IN1_BODY = 'BODY';
const IN1_CAPTION = 'CAPTION';
function _in1(doc, body) {
  doc.create({
    type: 'structured-node',
    id: 'in1',
    title: IN1_TITLE,
    body: IN1_BODY,
    caption: IN1_CAPTION
  });
  body.show('in1');
}

const T_CONTENT = [['A1', 'B1'], ['A2', 'B2']];

function _t1(doc, body) {
  doc.create({type: 'table-cell', id: 't1_a1', content:T_CONTENT[0][0]});
  doc.create({type: 'table-cell', id: 't1_b1', content:T_CONTENT[0][1]});
  doc.create({type: 'table-cell', id: 't1_a2', content:T_CONTENT[1][0]});
  doc.create({type: 'table-cell', id: 't1_b2', content:T_CONTENT[1][1]});
  body.show(doc.create({
    type: 'table', id: 't1', cells: [['t1_a1', 't1_b1'], ['t1_a2', 't1_b2']]
  }));
}

function _t1_sparse(doc, body) {
  _t1(doc, body);
  doc.set(['t1', 'cells'], [['t1_a1', null], [null, 't1_b2']]);
}

const test$11 = substanceTest.module('documentHelpers');

test$11("Get annotations for selection", (t) => {
  let doc = createTestArticle(simple);
  const path = ["p1", "content"];
  doc.create({ type: 'strong', path, startOffset: 0, endOffset: 1});
  doc.create({ type: 'emphasis', path, startOffset: 2, endOffset: 3});
  doc.create({ type: 'strong', path, startOffset: 5, endOffset: 6});
  // this lies outside of selection
  doc.create({ type: 'strong', path, startOffset: 6, endOffset: 7});
  let sel = doc.createSelection({
    type: "property",
    path,
    startOffset: 0,
    endOffset: 5
  });
  let annos = substance.documentHelpers.getPropertyAnnotationsForSelection(doc, sel);
  t.equal(annos.length, 3, "should return 3 annos for selection");
  // filtered by type
  annos = substance.documentHelpers.getPropertyAnnotationsForSelection(doc, sel, { type: 'emphasis'});
  t.equal(annos.length, 1, "should return one anno");
  t.equal(annos[0].type, 'emphasis', ".. of type emphasis");
  t.end();
});

test$11("Get text for null selection.", (t) => {
  let doc = createTestArticle(simple);
  t.equal(substance.documentHelpers.getTextForSelection(doc, null), "", "Should be empty for null selection.");
  t.equal(substance.documentHelpers.getTextForSelection(doc, doc.createSelection(null)), "", "Should be empty for null selection.");
  t.end();
});

test$11("Get text for property selection", (t) => {
  let doc = createTestArticle(simple);
  let sel = doc.createSelection({
    type: "property",
    path: ["p1", "content"],
    startOffset: 0,
    endOffset: 5
  });
  t.equal(substance.documentHelpers.getTextForSelection(doc, sel), "01234");
  t.end();
});

test$11("Get text for container selection", (t) => {
  let doc = createTestArticle(simple);
  let sel = doc.createSelection({
    type: "container",
    containerId: "body",
    startPath: ["p1", "content"],
    startOffset: 5,
    endPath: ["p2", "content"],
    endOffset: 5
  });
  t.equal(substance.documentHelpers.getTextForSelection(doc, sel), "56789\n01234");
  t.end();
});

test$11("Get change from document", (t) => {
  let doc = createTestArticle(simple);
  doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 0, endOffset: 1});
  let change = substance.documentHelpers.getChangeFromDocument(doc);
  t.equal(change.ops.length, 6, 'There should be 6 operations');
  t.deepEqual(change.ops.map(op => op.type), new Array(6).fill('create'), 'all should be create ops');
  t.deepEqual(change.ops.map(op => op.path[0]), ['p1', 'p2', 'p3', 'p4', 's1', 'body'], '.. in correct order');
  t.end();
});

test$11("deleteNode()", (t) => {
  let doc = createTestArticle(simple);[_l1, _l11].forEach(f=>f(doc, doc.get('body')));
  doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 0, endOffset: 1});
  substance.documentHelpers.deleteNode(doc, doc.get('p1'));
  t.nil(doc.get('p1'), 'node should have been deleted');
  t.nil(doc.get('s1'), 'annotation should have been deleted too');
  substance.documentHelpers.deleteNode(doc, doc.get('l1'));
  t.nil(doc.get('l1'), 'node should have been deleted');
  t.nil(doc.get('l1-1'), 'list item should have been deleted too');
  t.end();
});

test$11("deleteTextRange()", (t) => {
  const path = ['p1', 'content'];
  // anno is after
  let doc = createTestArticle(simple);
  let s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 5, endOffset: 6});
  substance.documentHelpers.deleteTextRange(doc, { path, offset: 0 }, { path, offset: 2 });
  t.deepEqual([s1.start.offset, s1.end.offset], [3,4], 'offsets should have been shifted');
  // anno is inside
  doc = createTestArticle(simple);
  s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 3, endOffset: 4});
  substance.documentHelpers.deleteTextRange(doc, { path, offset: 2 }, { path, offset: 5 });
  t.nil(doc.get('s1'), 'annotation should have been deleted');
  // anno.start between and anno.end after
  doc = createTestArticle(simple);
  s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 3, endOffset: 6});
  substance.documentHelpers.deleteTextRange(doc, { path, offset: 2 }, { path, offset: 5 });
  t.deepEqual([s1.start.offset, s1.end.offset], [2,3], 'offsets should have been updated correctly');
  // anno.start same and anno.end after
  doc = createTestArticle(simple);
  s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 3, endOffset: 6});
  substance.documentHelpers.deleteTextRange(doc, { path, offset: 3 }, { path, offset: 5 });
  t.deepEqual([s1.start.offset, s1.end.offset], [3,4], 'offsets should have been updated correctly');
  // anno.start before and anno.end between
  doc = createTestArticle(simple);
  s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 1, endOffset: 3});
  substance.documentHelpers.deleteTextRange(doc, { path, offset: 2 }, { path, offset: 4 });
  t.deepEqual([s1.start.offset, s1.end.offset], [1,2], 'offsets should have been updated correctly');
  // anno.start before and anno.end after
  doc = createTestArticle(simple);
  s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 1, endOffset: 6});
  substance.documentHelpers.deleteTextRange(doc, { path, offset: 2 }, { path, offset: 4 });
  t.deepEqual([s1.start.offset, s1.end.offset], [1,4], 'offsets should have been updated correctly');
  t.end();
});

test$11("deleteListRange()", (t) => {
  // first and last not entirely selected, and no item in between
  let doc = createTestArticle(simple);[_l1, _l11, _l12, _l13].forEach(f=>f(doc, doc.get('body')));
  let l1 = doc.get('l1');
  substance.documentHelpers.deleteListRange(doc, l1,
    { path: ['l1-1', 'content'], offset: 2 },
    { path: ['l1-2', 'content'], offset: 3 }
  );
  t.nil(doc.get('l1-2'), 'second list item should have been deleted');
  t.equal(doc.get('l1-1').getText(), LI1_TEXT.slice(0,2)+LI2_TEXT.slice(3), 'list items should have been merged');

  // item in between
  doc = createTestArticle(simple)
  ;[_l1, _l11, _l12, _l13].forEach(f=>f(doc, doc.get('body')));
  l1 = doc.get('l1');
  substance.documentHelpers.deleteListRange(doc, l1,
    { path: ['l1-1', 'content'], offset: 2 },
    { path: ['l1-3', 'content'], offset: 3 }
  );
  t.nil(doc.get('l1-2'), 'second list item should have been deleted');
  t.nil(doc.get('l1-3'), 'third list item should have been deleted');
  t.equal(doc.get('l1-1').getText(), LI1_TEXT.slice(0,2)+LI3_TEXT.slice(3), 'list items should have been merged');

  // entirely selected
  doc = createTestArticle(simple)
  ;[_l1, _l11, _l12, _l13].forEach(f=>f(doc, doc.get('body')));
  l1 = doc.get('l1');
  substance.documentHelpers.deleteListRange(doc, l1,
    { path: ['l1-1', 'content'], offset: 0 },
    { path: ['l1-3', 'content'], offset: LI3_TEXT }
  );
  t.nil(doc.get('l1-1'), 'first list item should have been deleted');
  t.nil(doc.get('l1-2'), 'second list item should have been deleted');
  t.nil(doc.get('l1-3'), 'third list item should have been deleted');

  t.end();
});

test$11("isContainerAnnotation()", (t) => {
  let doc = createTestArticle(simple);
  t.ok(substance.documentHelpers.isContainerAnnotation(doc, 'test-container-anno'));
  t.notOk(substance.documentHelpers.isContainerAnnotation(doc, 'strong'));
  t.end();
});

test$11("Get container annotations for property selection", (t) => {
  let doc = createTestArticle(containerAnnoSample);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p3', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  let annos;
  // without options
  annos = substance.documentHelpers.getContainerAnnotationsForSelection(doc, sel, 'body');
  t.equal(annos.length, 1, 'There should be one container anno');
  // filtered by type
  annos = substance.documentHelpers.getContainerAnnotationsForSelection(doc, sel, 'body', {
    type: 'test-container-anno'
  });
  t.equal(annos.length, 1, 'There should be one container anno');
  t.end();
});

if (substance.platform.inBrowser) {
  DOMElementTests$1('BrowserDOMElement');
}

DOMElementTests$1('MemoryDOMElement');

function DOMElementTests$1(impl) {

  const test = substanceTest.module('DOMElement ('+impl+')', {
    before: () => {
      if (impl === 'MemoryDOMElement') substance.platform.inBrowser = false;
    },
    after: () => {
      substance.platform._reset();
    }
  });

  test("getNativeElement()", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    const nativeEl = el.getNativeElement();
    let isNative;
    if (impl === 'BrowserDOMElement') {
      isNative = nativeEl instanceof window.Element;
    } else {
      isNative = nativeEl._isMemoryDOMElement;
    }
    t.ok(isNative, 'should return a native element.');
    t.end();
  });

  test("getNodeType()", function(t) {
    let doc = substance.DefaultDOMElement.createDocument('html');
    t.equal(doc.getNodeType(), 'document', 'document should have correct node type');
    let el = doc.createElement('div');
    t.equal(el.getNodeType(), 'element', 'element should have correct node type');
    t.equal(el.nodeType, 'element', '.. should also work via property getter');
    t.ok(el.isElementNode(), '.. should tell it is an element node');
    el = doc.createTextNode('foo');
    t.equal(el.getNodeType(), 'text', 'textNode should have correct node type');
    t.ok(el.isTextNode(), '.. and should tell it is a text node');
    el = doc.createComment('Hallo World!');
    t.equal(el.getNodeType(), 'comment', 'comment element should have correct node type');
    t.ok(el.isCommentNode(), '.. and should tell it is a comment node');
    el = doc.createProcessingInstruction('foo', 'bar');
    t.equal(el.getNodeType(), 'directive', 'procession instruction element should have correct node type');
    // cdata only allowed in XML documents
    doc = substance.DefaultDOMElement.createDocument('xml');
    el = doc.createCDATASection('XYZ');
    t.equal(el.getNodeType(), 'cdata', 'cdata element should have correct node type');
    t.end();
  });

  test("getTagName()", function(t) {
    const el = substance.DefaultDOMElement.parseSnippet('<p></p>', 'html');
    t.equal(el.getTagName(), 'p', 'should have correct tagName');
    t.equal(el.tagName, 'p', 'should also work with property accessor');
    t.equal(el.nodeName, 'p', 'nodeName is an alias');
    t.end();
  });

  test("getTagName() in HTML is always lower case", function(t) {
    const el = substance.DefaultDOMElement.parseSnippet('<DIV></DIV>', 'html');
    t.equal(el.tagName, 'div', 'should bet lower case');
    t.end();
  });

  test("getTagName() in XML does not change case", function(t) {
    const el = substance.DefaultDOMElement.parseSnippet('<myNode></myNode>', 'xml');
    t.equal(el.tagName, 'myNode', 'case should be changed');
    t.end();
  });

  test("setTagName()", function(t) {
    const el = substance.DefaultDOMElement.parseSnippet('<p></p>', 'html');
    // NOTE: in the browser this will be done by replacing the native element
    el.setTagName('div');
    t.equal(el.getTagName(), 'div', 'should have changed the tagName');
    el.tagName = 'h1';
    t.equal(el.getTagName(), 'h1', 'should also work via property setter');
    t.end();
  });

  test("getId()", function(t) {
    const el = substance.DefaultDOMElement.parseSnippet('<div id="foo"></div>', 'html');
    t.equal(el.getId(), 'foo', 'should have correct id');
    t.equal(el.id, 'foo', 'should also work with property accessor');
    t.end();
  });

  test("setId()", function(t) {
    const el = substance.DefaultDOMElement.parseSnippet('<div id="foo"></div>', 'html');
    el.setId('bar');
    t.equal(el.getId(), 'bar', 'should have changed id');
    el.id = 'baz';
    t.equal(el.getId(), 'baz', 'should also work using property setter');
    t.end();
  });

  test("setTagName", function(t) {
    var el = substance.DefaultDOMElement.parseSnippet('<p class="foo">ABC<b>DEF</b>GHI</p>', 'html');
    var onClick = substanceTest.spy();
    el.on('click', onClick);
    // this call is brutal as a new element needs to be created
    // and all the content and attributes be copied over
    el.setTagName('h1');
    t.equal(el.tagName, 'h1', 'Now the element should be a heading');
    t.equal(el.textContent, 'ABCDEFGHI', '.. its text content should have been preserved');
    t.equal(el.getChildCount(), 3, '.. and its children should still be there');
    t.notNil(el.find('b'), '.. including the <b> element');
    el.click();
    t.equal(onClick.callCount, 1, '.. and even the click handler should still work');
    t.end();
  });

  test("setTagName on XML should create XML elements", function(t) {
    var el = substance.DefaultDOMElement.parseSnippet('<dummy></dummy>', 'xml');
    // this call is brutal as a new element needs to be created
    // and all the content and attributes be copied over
    el.setTagName('foo');
    t.equal(el.getNativeElement().ownerDocument.contentType, 'application/xml', 'Element should still be an XML element');
    el.setInnerHTML('<link>foo</link>');
    // when using an HTML element <link> will get exported as self-closing
    t.equal(el.outerHTML, '<foo><link>foo</link></foo>');
    t.end();
  });

  test("hasClass()", function(t) {
    const el = substance.DefaultDOMElement.parseSnippet('<div class="foo">TEST</div>', 'html');
    t.ok(el.hasClass('foo'), 'should have class "foo".');
    t.end();
  });

  test("addClass()", function(t) {
    let doc = substance.DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.addClass('foo');
    t.ok(el.hasClass('foo'), 'should have class "foo".');
    t.end();
  });

  test("'class' attribute", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.setAttribute('class', 'foo bar');
    t.ok(el.hasClass('foo'), 'should have class "foo" set');
    t.ok(el.hasClass('bar'), 'should have class "bar" set');
    t.end();
  });

  test("removeClass()", function(t) {
    const el = substance.DefaultDOMElement.parseSnippet('<div class="foo">TEST</div>', 'html');
    el.removeClass('foo');
    t.notOk(el.hasClass('foo'), 'class should have been removed');
    t.end();
  });

  test("hasAttribute()", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<div foo="foo">TEST</div>', 'html');
    t.ok(el.hasAttribute('foo'), 'should have attribute "foo"');
    t.end();
  });

  test("getAttribute()", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<div foo="1">TEST</div>', 'html');
    t.equal(el.getAttribute('foo'), "1", 'should have return attribute "foo"');
    t.end();
  });

  test("setAttribute()", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.setAttribute('foo', 1);
    t.equal(el.getAttribute('foo'), "1", 'should have set attribute');
    t.end();
  });

  test("removeAttribute()", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<div foo="1">TEST</div>', 'html');
    el.removeAttribute('foo');
    t.isNil(el.getAttribute('foo'), 'should removed attribute');
    t.end();
  });

  test("attr() - getter", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<div foo="1">TEST</div>', 'html');
    t.equal(el.attr('foo'), "1", 'should return attribute value');
    t.end();
  });

  test("attr() - single attribute", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.attr('foo', 1);
    t.equal(el.getAttribute('foo'), "1", 'should have set attribute');
    t.end();
  });

  test("attr() - multiple attributes", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.attr({
      foo: 1,
      bar: 2
    });
    t.equal(el.getAttribute('foo'), "1", 'should have set attribute');
    t.equal(el.getAttribute('bar'), "2", 'should have set attribute');
    t.end();
  });

  test("removeAttr() - single attribute", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<div foo="1">TEST</div>', 'html');
    el.removeAttr('foo');
    t.isNil(el.getAttribute('foo'), 'should removed attribute');
    t.end();
  });

  test("removeAttr() - multiple attributes", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<div foo="1" bar="2">TEST</div>', 'html');
    el.removeAttr('foo bar');
    t.isNil(el.getAttribute('foo'), 'should removed attribute');
    t.isNil(el.getAttribute('bar'), 'should removed attribute');
    t.end();
  });

  test("getAttributes()", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<div foo="1" bar="2">TEST</div>', 'html');
    const entries = Array.from(el.getAttributes().entries());
    t.deepEqual(entries, [["foo", '1'], ["bar", '2']], 'should return all attributes');
    t.end();
  });

  test("getProperty()", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<input type="text" value="foo">', 'html');
    t.equal(el.getProperty('value'), "foo", 'should have correct value');
    t.end();
  });

  test("getProperty() -- input[type=checkbox]", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<input type="checkbox" checked>', 'html');
    t.equal(el.getProperty('value'), 'on', 'should have correct value');
    t.equal(el.getProperty('checked'), true, 'should have correct "checked" property');
    t.end();
  });

  test("setProperty()", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    const el = doc.createElement('input');
    el.setProperty('value', 1);
    t.equal(el.getProperty('value'), "1", 'should have set property');
    t.end();
  });

  test("setProperty() on XML element is not allowed", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('xml');
    const el = doc.createElement('foo');
    t.throws(()=>{
      el.setProperty('bar', 1);
    }, 'setProperty() on XML element is not allowed');
    t.end();
  });

  test("unsetting a property", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.setProperty('foo', 1);
    el.setProperty('foo', undefined);
    t.isNil(el.getProperty('foo'), 'should have unset property');
    t.end();
  });

  test("htmlProp() -- getter", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<input type="text" value="foo">', 'html');
    t.equal(el.htmlProp('value'), "foo", 'should return value');
    t.end();
  });

  test("htmlProp() -- single property", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    const el = doc.createElement('input');
    el.htmlProp('value', 1);
    t.equal(el.getProperty('value'), "1", 'should have set property');
    t.end();
  });

  test("htmlProp() -- multiple properties", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    const el = doc.createElement('input');
    el.htmlProp({
      'value': 1,
      'foo': 'foo'
    });
    t.equal(el.getProperty('value'), "1", 'should have set property');
    t.equal(el.getProperty('foo'), "foo", 'should have set property');
    t.end();
  });

  test("getValue()", (t) => {
    const el = substance.DefaultDOMElement.parseSnippet('<input type="text" value="foo">', 'html');
    t.equal(el.getValue(), "foo", 'should have correct value');
    t.equal(el.val(), "foo", 'should work via jquery style getter');
    t.equal(el.value, "foo", 'should work via property getter');
    t.end();
  });

  test("setValue()", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    const el = doc.createElement('input');
    el.setValue(1);
    t.equal(el.getValue(), "1", 'should have set value');
    el.value = 2;
    t.equal(el.value, "2", 'should also work via property accessors');
    el.val(3);
    t.equal(el.value, "3", 'should also work via jquery style setter');
    t.end();
  });

  test("getStyle()", (t) => {
    let el = substance.DefaultDOMElement.parseSnippet('<div style="color:blue"></div>', 'html');
    t.equal(el.getStyle('color'), 'blue', 'element should have style');
    t.equal(el.getAttribute('style'), 'color:blue', 'style can be retrieved via "style" attribute');
    t.end();
  });

  test("setStyle()", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    el.setStyle('color', 'blue');
    t.equal(el.getStyle('color'), 'blue', 'element should have style set');
    el.setAttribute('style', 'color:green');
    t.equal(el.getStyle('color'), 'green', 'style can be set via attribute "style" too');
    t.end();
  });

  test("setStyle() -- px styles", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    el.css({
      top: 10,
      bottom: 11,
      left: 12,
      right: 13,
      height: 100,
      width: 50
    });
    t.equal(el.getStyle('top'), '10px', 'px should have been added automatically');
    t.equal(el.getStyle('bottom'), '11px', 'px should have been added automatically');
    t.equal(el.getStyle('left'), '12px', 'px should have been added automatically');
    t.equal(el.getStyle('right'), '13px', 'px should have been added automatically');
    t.equal(el.getStyle('height'), '100px', 'px should have been added automatically');
    t.equal(el.getStyle('width'), '50px', 'px should have been added automatically');
    t.end();
  });

  test("css()", (t) => {
    let el = substance.DefaultDOMElement.parseSnippet('<div style="color:blue"></div>', 'html');
    t.equal(el.css('color'), 'blue', 'can access style via jquery style getter');
    el.css('color', 'green');
    t.equal(el.getStyle('color'), 'green', 'can change style via jquery style setter');
    el.css({
      color: 'red',
      background: 'orange'
    });
    t.deepEqual([el.getStyle('color'), el.getStyle('background')], ['red', 'orange'], 'can change multiple styles at once via jquery style setter');
    t.end();
  });

  test("setTextContent", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    el.setTextContent('foo');
    t.equal(el.getTextContent(), 'foo', 'text content should have been set');
    el.textContent = 'bar';
    t.equal(el.getTextContent(), 'bar', 'should work also via property setter');
    el.text('baz');
    t.equal(el.getTextContent(), 'baz', 'should work also via jquery style setter');
    t.end();
  });

  test("innerHTML", (t) => {
    let el = substance.DefaultDOMElement.parseSnippet('<p>abc<span class="foo" data-foo="foo">TEST</span>def</p>', 'html');
    t.equal(el.innerHTML, 'abc<span class="foo" data-foo="foo">TEST</span>def', 'el.innerHTML should give the HTML of the childNodes');
    el.innerHTML = 'TEST';
    t.equal(el.innerHTML, 'TEST', 'setting innerHTML should replace the content');
    t.end();
  });

  test("innerHTML on XML", (t) => {
    let el = substance.DefaultDOMElement.parseSnippet('<foo>abc<bar bla="foo" blupp="foo">TEST</bar>def</foo>', 'xml');
    t.equal(el.innerHTML, 'abc<bar bla="foo" blupp="foo">TEST</bar>def', 'el.innerHTML should give the XML of the element content');
    el.innerHTML = 'TEST';
    t.equal(el.innerHTML, 'TEST', 'setting innerHTML should replace the content');
    t.end();
  });

  test("html()", (t) => {
    let el = substance.DefaultDOMElement.parseSnippet('<div>TEST</div>', 'html');
    t.equal(el.html(), 'TEST', 'should return innerHTML');
    el.html('<span>FOO</span>');
    t.equal(el.innerHTML, '<span>FOO</span>', 'should have set innerHTML');
    t.end();
  });

  test("outerHTML", function(t) {
    var p = substance.DefaultDOMElement.parseSnippet('<p class="foo">TEST</p>', 'html');
    t.equal(p.outerHTML, '<p class="foo">TEST</p>', 'outerHTML should be complete');
    t.end();
  });

  test("outerHTML of XML with camelCase tagnames", function(t) {
    let el = substance.DefaultDOMElement.parseSnippet('<dummy><myNode></myNode></dummy>', 'xml');
    t.equal(el.outerHTML, '<dummy><myNode/></dummy>', 'XML tags should be serialized preserving case.');
    t.end();
  });

  test("find via tagname", function(t) {
    let el = substance.DefaultDOMElement.parseSnippet('<dummy><foo></foo></dummy>', 'xml');
    let foo = el.find('foo');
    t.notNil(foo, 'Should find a <foo> element');
    t.end();
  });

  test("find multiple elements", function(t) {
    let el = substance.DefaultDOMElement.parseSnippet('<dummy><foo>Bla</foo><foo>Blupp</foo></dummy>', 'xml');
    let foos = el.findAll('foo');
    t.equal(foos.length, 2, 'Should find two <foo> elements');
    t.equal(foos[0].textContent, 'Bla', '... with correct textContent');
    t.equal(foos[1].textContent, 'Blupp', '... with correct textContent');
    t.end();
  });

  test("find via #id", function(t) {
    let el = substance.DefaultDOMElement.parseSnippet('<dummy><bla id="foo"></bla></dummy>', 'xml');
    let foo = el.find('#foo');
    t.notNil(foo, 'Should find a #foo element');
    t.end();
  });

  test("find via attr", function(t) {
    let el = substance.DefaultDOMElement.parseSnippet('<dummy><bla data-id="foo"></bla></dummy>', 'xml');
    let foo = el.find('[data-id=foo]');
    t.notNil(foo, 'Should find a element with data-id=foo');
    t.end();
  });

  test("find in XML with camelCase tagnames", function(t) {
    let el = substance.DefaultDOMElement.parseSnippet('<dummy><myNode myValue="foo"></myNode></dummy>', 'xml');
    t.notNil(el.find('myNode'), 'Should find a <myNode> element');
    t.notNil(el.find('[myValue="foo"]'), 'Should find a [myValue] element');
    t.end();
  });

  test("find in XML with namespace tagnames", function(t) {
    let xmlDoc = substance.DefaultDOMElement.parseXML('<dummy xmlns:foo="foo"><foo:myNode></foo:myNode></dummy>');
    let myNode = xmlDoc.find('myNode');
    t.notNil(myNode, 'Should find a <foo:myNode> element');
    t.equal(myNode.tagName, 'foo:myNode', 'tagName should be qualified');
    t.end();
  });

  test("insertAt", function(t) {
    let dummy = substance.DefaultDOMElement.parseSnippet('<dummy><bla></bla></dummy>', 'xml');
    let doc = dummy.getOwnerDocument();
    let newChild = doc.createElement('blupp');
    dummy.insertAt(1, newChild);
    t.equal(dummy.getChildCount(), 2, 'There should be 2 children now');
    t.ok(dummy.getChildAt(0).is('bla'), '.. the first should be <bla>');
    t.ok(dummy.getChildAt(1).is('blupp'), '.. the second should be <blupp>');
    t.end();
  });

  test("reattaching an element via insertAt", function(t) {
    let dummy = substance.DefaultDOMElement.parseSnippet('<dummy><bla><blupp></blupp></bla></dummy>', 'xml');
    let blupp = dummy.find('blupp');
    dummy.insertAt(0, blupp);
    t.equal(dummy.getChildCount(), 2, 'There should be 2 children now');
    t.ok(dummy.getChildAt(0).is('blupp'), '.. the first should be <blupp>');
    t.ok(dummy.getChildAt(1).is('bla'), '.. the second should be <bla>');
    t.equal(dummy.getChildAt(1).getChildCount(), 0, '.. and <bla> should have no children');
    t.end();
  });

  test("reattaching an element via append", function(t) {
    let dummy = substance.DefaultDOMElement.parseSnippet('<dummy><bla><blupp></blupp></bla></dummy>', 'xml');
    let blupp = dummy.find('blupp');
    dummy.appendChild(blupp);
    t.equal(dummy.getChildCount(), 2, 'There should be 2 children now');
    t.ok(dummy.getChildAt(0).is('bla'), '.. the first should be <bla>');
    t.ok(dummy.getChildAt(1).is('blupp'), '.. the second should be <blupp>');
    t.equal(dummy.getChildAt(0).getChildCount(), 0, '.. and <bla> should have no children');
    t.end();
  });

  test("children should only include elements", function(t) {
    let dummy = substance.DefaultDOMElement.parseSnippet('<dummy>bli<bla></bla>blupp</dummy>', 'xml');
    dummy.children.forEach((c)=>{
      t.ok(c.isElementNode(), 'child should be an element');
    });
    t.end();
  });

  test("event listeners for custom events", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    let event = null;
    let handleFoo = function(e) {
      event = e;
    };
    el.addEventListener('foo', handleFoo);
    t.deepEqual(el.getEventListeners().map(l=>l.eventName), ['foo'], 'Element should have one listener');
    el.emit('foo', { bar: 'bar' });
    t.notNil(event, 'should have handled event');
    t.notNil(event.detail, 'event should have event.detail');
    t.deepEqual(event.detail, { bar: 'bar' });
    el.removeEventListener('foo', handleFoo);
    t.equal(el.getEventListeners().length, 0, 'element should not have a listener anymore');
    event = null;
    el.emit('foo', { baz: 'baz' });
    t.nil(event, 'event handler should not be called this time');
    t.end();
  });

  test("getChildNodes()", (t) => {
    let el = substance.DefaultDOMElement.parseSnippet('<p>Foo<span>Bar</span>Baz</p>', 'html');
    let childNodes = el.childNodes;
    t.deepEqual(childNodes.map(c => c.nodeType), ['text', 'element', 'text'], "should return all childNodes");
    t.end();
  });

  test("getChildren()", (t) => {
    let el = substance.DefaultDOMElement.parseSnippet('<p>Foo<span>Bar</span>Baz</p>', 'html');
    let children = el.children;
    t.deepEqual(children.map(c => c.tagName), ['span'], "should return all children");
    t.end();
  });

  test("firstChild", (t) => {
    let el = substance.DefaultDOMElement.parseSnippet('<p>Foo<span>Bar</span>Baz</p>', 'html');
    let firstChild = el.firstChild;
    t.equal(firstChild.textContent, 'Foo', "should return all firstChild");
    t.end();
  });

  test("lastChild", (t) => {
    let el = substance.DefaultDOMElement.parseSnippet('<p>Foo<span>Bar</span>Baz</p>', 'html');
    let lastChild = el.lastChild;
    t.equal(lastChild.textContent, 'Baz', "should return all lastChild");
    t.end();
  });

  test("is()", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div').attr('data-id', 'foo');
    t.ok(el.is('div[data-id=foo]'));
    t.notOk(el.is('foo'));
    t.end();
  });

  test("getOwnerDocument()", (t) => {
    let doc = substance.DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    t.equal(el.getOwnerDocument(), doc, 'should return the ownerDocument');
    t.equal(el.ownerDocument, doc, 'should also work via property getter');
    t.end();
  });

}

const test$12 = substanceTest.module('DOMImporter');

const pConverter = {
  type: 'paragraph',
  tagName: 'p',
  import(el, node, converter) {
    node.content = converter.annotatedText(el, [node.id, 'content']);
  }
};

test$12("creating a DOMImporter", (t) => {
  const schema = getTestConfig().getSchema();
  // fail without converters
  t.throws(() => {
    new substance.DOMImporter({ schema });
  }, 'should throw if no converters are given');
  // fail without DocumentClass
  t.throws(() => {
    new substance.DOMImporter({ converters: [] });
  }, 'should throw if schema is not given');
  // fail if a converter has no associated type
  t.throws(() => {
    new substance.DOMImporter({ schema, converters: [{
      matchElement() { return true }
    }] });
  }, 'should throw if a converter does not have associated type');
  // fail if a converter has no matcher
  t.throws(() => {
    new substance.DOMImporter({ schema, converters: [{
      type: 'paragraph'
    }] });
  }, 'should throw if a converter does not have a matcher function');
  t.throws(() => {
    new substance.DOMImporter({ schema, converters: [{
      type: 'heading', tagName: 'h1'
    }] });
  }, 'should throw if there is no converter for the defaultTextType');
  t.throws(() => {
    new substance.DOMImporter({ schema, converters: [{
      type: 'foo', tagName: 'h1'
    }] });
  }, 'should throw if a converter is associated with an unknown node type');
  t.end();
});

test$12("default matchElement()", (t) => {
  const testConverter = {
    type: 'paragraph', tagName: 'p'
  };
  const importer = createImporter([testConverter]);
  t.notNil(importer._allConverters[0].matchElement, 'registered converter should have a matchElement() function');
  t.end();
});

test$12("using a Converter class", (t) => {
  class MyConverter {
    get type() { return 'paragraph' }
    get tagName() { return 'p' }
  }
  const importer = createImporter([MyConverter]);
  t.equal(importer._allConverters[0].tagName, 'p', 'there should be one converter registered for <p>');
  t.end();
});

test$12("convertElement() -- block element", (t) => {
  const importer = createImporter();
  const el = substance.DefaultDOMElement.parseSnippet('<p>TEST</p>', 'html');
  const node = importer.convertElement(el);
  t.equal(node.type, 'paragraph', 'should have converted element to node');
  t.end();
});

// Note: it is not common to convert inline nodes outside of their context -- that's why they are called inline
test$12("convertElement() -- inline node", (t) => {
  const testConverter = {
    type: 'test-inline-node',
    matchElement(el) { return el.is('[data-type=test-inline-node]') },
    import(el, node) {
      node.content = el.textContent;
    }
  };
  const importer = createImporter([pConverter, testConverter]);
  const el = substance.DefaultDOMElement.parseSnippet('<span data-type="test-inline-node">TEST</span>', 'html');
  const node = importer.convertElement(el);
  t.equal(node.type, 'test-inline-node', 'should have converted element to node');
  t.end();
});

// Note: it is not common to convert annotations outside of their context
test$12("convertElement() -- annotation element", (t) => {
  const importer = createImporter();
  const el = substance.DefaultDOMElement.parseSnippet('<b>TEST</b>', 'html');
  const node = importer.convertElement(el);
  t.equal(node.type, 'strong', 'should have converted element to node');
  t.end();
});

test$12("convertElement() should throw if no converter found", (t) => {
  const importer = createImporter([pConverter]);
  const el = substance.DefaultDOMElement.parseSnippet('<h1>TEST</h1>', 'html');
  t.throws(() => {
    importer.convertElement(el);
  }, 'should throw if no converter found');
  t.end();
});

test$12("converting paragraph with inline node", (t) => {
  const testConverter = {
    type: 'test-inline-node',
    matchElement(el) { return el.is('[data-type=test-inline-node]') }
  };
  const importer = createImporter([pConverter, testConverter]);
  const el = substance.DefaultDOMElement.parseSnippet('<p>abc <span data-type="test-inline-node">TEST</span> def</p>', 'html');
  const node = importer.convertElement(el);
  t.equal(node.content, `abc ${substance.DOMImporter.INVISIBLE_CHARACTER} def`, 'should have inserted an invisible character');
  t.end();
});

test$12("converting an annotated paragraph", (t) => {
  const importer = createImporter();
  const el = substance.DefaultDOMElement.parseSnippet('<p>abc <b>TEST</b> def</p>', 'html');
  const node = importer.convertElement(el);
  const doc = node.getDocument();
  const annos = doc.getAnnotations([node.id]);
  t.equal(annos.length, 1, 'there should be one annotation');
  t.end();
});


test$12("plainText()", (t) => {
  const testConverter = Object.assign({}, pConverter, {
    import(el, node, converter) {
      node.content = converter.plainText(el);
    }
  });
  const importer = createImporter([testConverter]);
  const el = substance.DefaultDOMElement.parseSnippet('<p>abc <b>TEST</b> def</p>', 'html');
  const node = importer.convertElement(el);
  t.equal(node.content, `abc TEST def`, 'should have converted plain text');
  t.end();
});

test$12("convertContainer() -- trailing text", (t) => {
  const importer = createImporter();
  const els = substance.DefaultDOMElement.parseSnippet('<p>A paragraph</p> and some trailing text', 'html');
  let container = importer.convertContainer(els, 'body');
  t.equal(container.nodes.length, 2, 'should have two nodes');
  t.deepEqual(container.getNodes().map(n=>n.type), ['paragraph', 'paragraph'], 'both should be paragraphs');
  t.end();
});

test$12("resolving id collisions", (t) => {
  // converter with 'tagName' gets a default matcher
  const importer = createImporter();
  const els = substance.DefaultDOMElement.parseSnippet('<p id="foo">A paragraph</p><p id="foo">and another one with the same id</p>', 'html');
  let container = importer.convertContainer(els, 'body');
  t.equal(container.nodes.length, 2, 'should have two nodes');
  t.equal(container.nodes[0], 'foo', 'first should have the original id');
  t.notEqual(container.nodes[1], 'foo', 'second should have a new one');
  t.end();
});

/*
TODO:
  - import with options
*/

function createImporter(converters) {
  const config = getTestConfig();
  const schema = config.getSchema();
  if (!converters) {
    converters = config.getConverterRegistry().get('html').values();
  }
  return new substance.DOMImporter({
    schema,
    converters: converters
  })
}

function checkValues(t, actual, expected) {
  if (!substance.isObject(actual)) {
    t.fail('Provided object is invalid: actual=' + String(actual));
    return
  }
  substance.forEach(expected, (expectedVal, key) => {
    if (substance.isPlainObject(expectedVal) || substance.isArray(expectedVal)) {
      t.deepEqual(actual[key], expectedVal, key + ' should be correct');
    } else {
      t.equal(actual[key], expectedVal, key + ' should be correct');
    }
  });
}

const test$13 = substanceTest.module('DOMSelection');

test$13.UI("Mapping a cursor inside a TextProperty from DOM to model", function(t) {
  let { editor, surface } = setupEditor(t, _p1);
  let domSelection = surface.context.domSelection;
  let node = editor.el.find('[data-path="p1.content"]').getFirstChild();
  substance.setDOMSelection(node, 3);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

test$13.UI("Mapping a cursor in an empty paragraph from DOM to model", function(t) {
  let { editor, surface } = setupEditor(t, _empty);
  let domSelection = surface.context.domSelection;
  let node = editor.el.find('[data-path="empty.content"]').getFirstChild();
  substance.setDOMSelection(node, 0);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

test$13.UI("Mapping a ContainerSelection from DOM to model", function(t) {
  let { editor, surface } = setupEditor(t, surfaceWithParagraphs);
  let domSelection = surface.context.domSelection;
  let p1Text = editor.el.find('[data-path="p1.content"]').getFirstChild();
  let p2Text = editor.el.find('[data-path="p2.content"]').getFirstChild();
  substance.setDOMSelection(p1Text, 1, p2Text, 2);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 2,
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

test$13.FF("Issue #354: Wrong selection in FF when double clicking between lines", function(t) {
  let { editor, surface } = setupEditor(t, surfaceWithParagraphs);
  let domSelection = surface.context.domSelection;
  let surfaceEl = editor.el.find('[data-id="body"]');
  substance.setDOMSelection(surfaceEl, 0, surfaceEl, 1);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'node',
    nodeId: 'p1',
    mode: 'full',
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

// happens when using the same selection as in #354 in Chrome
test$13.UI("DOM selection that starts in a TextNode and ends in a paragraph on element level", function(t) {
  let { editor, surface } = setupEditor(t, surfaceWithParagraphs);
  let domSelection = surface.context.domSelection;
  let p1Text = editor.el.find('[data-path="p1.content"]').getFirstChild();
  let p2El = editor.el.find('[data-id="p2"]');
  substance.setDOMSelection(p1Text, 0, p2El, 0);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 0,
    endPath: ['p2', 'content'],
    endOffset: 0,
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

test$13.UI("Issue #376: Wrong selection mapping at end of paragraph", function(t) {
  let { editor, surface } = setupEditor(t, surfaceWithParagraphs);
  let domSelection = surface.context.domSelection;
  let p1span = editor.el.find('[data-id="p1"] span');
  let p2El = editor.el.find('[data-id="p2"]');
  substance.setDOMSelection(p1span, 1, p2El, 0);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 2,
    endPath: ['p2', 'content'],
    endOffset: 0,
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

test$13.UI("Rendering a ContainerSelection", function(t) {
  let { editor, doc, surface } = setupEditor(t, surfaceWithParagraphs);
  let domSelection = surface.context.domSelection;
  let sel = doc.createSelection({
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 1,
    containerId: 'body',
    surfaceId: 'body'
  });
  domSelection.setSelection(sel);

  let p1Text = editor.el.find('[data-path="p1.content"]').getFirstChild();
  let p2Text = editor.el.find('[data-path="p2.content"]').getFirstChild();

  let wSel = window.getSelection();
  t.equal(wSel.anchorNode, p1Text.getNativeElement(), 'anchorNode should be in first paragraph.');
  t.equal(wSel.anchorOffset, 1, 'anchorOffset should be correct.');
  t.equal(wSel.focusNode, p2Text.getNativeElement(), 'focusNode should be in second paragraph.');
  t.equal(wSel.focusOffset, 1, 'focusOffset should be correct.');
  t.end();
});

test$13.UI("Rendering a cursor after inline node", function(t) {
  let { editor, doc, surface } = setupEditor(t, paragraphWithInlineNodes);
  let domSelection = surface.context.domSelection;
  let sel = doc.createSelection({
    type: 'property',
    path: ['p', 'content'],
    startOffset: 3,
    containerId: 'body',
    surfaceId: 'body'
  });
  let {start, end} = domSelection.mapModelToDOMCoordinates(sel);
  let pSpan = editor.el.find('[data-path="p.content"]');
  let third = pSpan.getChildAt(2);
  t.ok(start.container === third.getNativeElement(), 'anchorNode should be correct.');
  t.equal(start.offset, 0, 'anchorOffset should be correct.');
  t.ok(end.container === start.container, 'focusNode should be correct.');
  t.equal(end.offset, 0, 'focusOffset should be correct.');
  t.end();
});

test$13.UI("Rendering a cursor after inline node at the end of a property", function(t) {
  let { editor, doc, surface } = setupEditor(t, paragraphWithInlineNodes);
  let domSelection = surface.context.domSelection;
  let sel = doc.createSelection({
    type: 'property',
    path: ['p', 'content'],
    startOffset: 13,
    containerId: 'body',
    surfaceId: 'body'
  });
  let {start, end} = domSelection.mapModelToDOMCoordinates(sel);
  let pSpan = editor.el.find('[data-path="p.content"]');
  t.ok(start.container === pSpan.getNativeElement(), 'anchorNode should be correct.');
  t.equal(start.offset, 6, 'anchorOffset should be correct.');
  t.ok(end.container === start.container, 'focusNode should be correct.');
  t.equal(end.offset, 6, 'focusOffset should be correct.');
  t.end();
});

// test.UI("Rendering a cursor before isolated node", function(t) {
//   let { doc, surface } = setupEditor(t, _t1)
//   let domSelection = surface.context.domSelection
//   let sel = doc.createSelection({
//     type: 'node',
//     nodeId: 't1',
//     mode: 'before',
//     containerId: 'body',
//     surfaceId: 'body'
//   })
//   domSelection.setSelection(sel)
//   // let {start, end} = domSelection.mapModelToDOMCoordinates(sel)
//   // let pSpan = editor.el.find('[data-path="p.content"]')
//   // t.ok(start.container === pSpan.getNativeElement(), 'anchorNode should be correct.')
//   // t.equal(start.offset, 6, 'anchorOffset should be correct.')
//   // t.ok(end.container === start.container, 'focusNode should be correct.')
//   // t.equal(end.offset, 6, 'focusOffset should be correct.')
//   t.end()
// })

// test.UI("Rendering a cursor after isolated node", function(t) {
//   let { doc, surface } = setupEditor(t, _t1)
//   let domSelection = surface.context.domSelection
//   let sel = doc.createSelection({
//     type: 'node',
//     nodeId: 't1',
//     mode: 'after',
//     containerId: 'body',
//     surfaceId: 'body'
//   })
//   domSelection.setSelection(sel)
//   // let {start, end} = domSelection.mapModelToDOMCoordinates(sel)
//   // let pSpan = editor.el.find('[data-path="p.content"]')
//   // t.ok(start.container === pSpan.getNativeElement(), 'anchorNode should be correct.')
//   // t.equal(start.offset, 6, 'anchorOffset should be correct.')
//   // t.ok(end.container === start.container, 'focusNode should be correct.')
//   // t.equal(end.offset, 6, 'focusOffset should be correct.')
//   t.end()
// })

function surfaceWithParagraphs(doc, body) {
  let tx = new substance.EditingInterface(doc);
  body.show(tx.create({
    type: 'paragraph',
    id: 'p1',
    content: 'AA'
  }));
  body.show(tx.create({
    type: 'paragraph',
    id: 'p2',
    content: 'BBB'
  }));
  body.show(tx.create({
    type: 'paragraph',
    id: 'p3',
    content: 'CCCC'
  }));
}

function paragraphWithInlineNodes(doc, body) {
  let tx = new substance.EditingInterface(doc);
  body.show(tx.create({
    type: 'paragraph',
    id: 'p',
    content: '0123456789'
  }));
  // -> 01X234X56789X
  tx.setSelection({type: 'property', path: ['p', 'content'], startOffset: 2});
  tx.insertInlineNode({ type: 'test-inline-node', id: 'in1', content: '[1]'});
  tx.setSelection({type: 'property', path: ['p', 'content'], startOffset: 6});
  tx.insertInlineNode({ type: 'test-inline-node', id: 'in2', content: '[2]'});
  tx.setSelection({type: 'property', path: ['p', 'content'], startOffset: 12});
  tx.insertInlineNode({ type: 'test-inline-node', id: 'in3', content: '[3]'});
}

function headersAndParagraphs(tx) {
  var body = tx.get('body');
  tx.create({
    type: 'heading',
    id: 'h1',
    content: 'Section 1',
    level: 1
  });
  body.show('h1');
  tx.create({
    type: 'paragraph',
    id: 'p1',
    content: 'Paragraph 1'
  });
  body.show('p1');
  tx.create({
    type: 'heading',
    id: 'h2',
    content: 'Section 2',
    level: 1
  });
  body.show('h2');
  tx.create({
    type: 'paragraph',
    id: 'p2',
    content: 'Paragraph with annotation'
  });
  body.show('p2');
  tx.create({
    type: 'emphasis',
    id: 'em1',
    start: {
      path: ['p2', 'content'],
      offset: 15,
    },
    end: {
      offset: 25
    }
  });
  tx.create({
    type: 'heading',
    id: 'h3',
    content: 'Section 2.2',
    level: 2
  });
  body.show('h3');
  tx.create({
    type: 'paragraph',
    id: 'p3',
    content: 'Paragraph 3'
  });
  body.show('p3');
  // tx.create({
  //   type: "test-node",
  //   id: "test",
  //   boolVal: true,
  //   stringVal: "Test",
  //   arrayVal: [1, 2, 3, 4],
  //   objectVal: { "a": 1, "b": 2 }
  // })
}

/* eslint-disable no-use-before-define */
const test$14 = substanceTest.module('Editing');

// TODO: consolidate specification and 'category labels' of tests

// TODO: we should enable t.sandbox for all tests so that this implementation is
// tested on all platforms

test$14("IT1: Inserting text with cursor in the middle of a TextProperty", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'xxx'+P1_TEXT.slice(3), 'Text should have been inserted correctly.');
  t.equal(sel.start.offset, 6, 'Cursor should be after inserted text');
  t.end();
});

test$14("IT2: Inserting text with cursor within TextProperty inside annotation", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0,4)+'xxx'+P1_TEXT.slice(4), 'Text should have been inserted correctly.');
  t.equal(s1.end.offset, 8, 'Annotation should have been expanded.');
  t.equal(sel.start.offset, 7, 'Cursor should be after inserted text');
  t.end();
});

test$14("IT3: Inserting text with cursor within TextProperty at the start of an annotation", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'xxx'+P1_TEXT.slice(3), 'Text should have been inserted correctly.');
  t.equal(s1.start.offset, 6, 'Annotation should have been moved.');
  t.equal(s1.end.offset, 8, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 6, 'Cursor should be after inserted text');
  t.end();
});

test$14("IT4: Inserting text with cursor within TextProperty at the end of an annotation", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0,5)+'xxx'+P1_TEXT.slice(5), 'Text should have been inserted correctly.');
  t.equal(s1.start.offset, 3, 'Annotation should have been moved.');
  t.equal(s1.end.offset, 8, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 8, 'Cursor should be after inserted text');
  t.end();
});

test$14("IT5: Inserting text with range within TextProperty", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 2,
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.equal(p1.getText(), P1_TEXT.slice(0, 2)+'xxx'+P1_TEXT.slice(5), 'Text should have been inserted correctly.');
  t.equal(sel.start.offset, 5, 'Cursor should be after inserted text');
  t.ok(sel.isCollapsed(), '... collapsed');
  t.end();
});

test$14("IT6: Inserting text with range within TextProperty overlapping an annotion aligned at the left side", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  doc.set(['s1', 'end', 'offset'], 6);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0, 3)+'xxx'+P1_TEXT.slice(5), 'Text should have been inserted correctly.');
  t.equal(s1.start.offset, 3, 'Annotation should have been moved.');
  t.equal(s1.end.offset, 7, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 6, 'Cursor should be after inserted text');
  t.ok(sel.isCollapsed(), '... collapsed');
  t.end();
});

test$14("IT7: Inserting text with range within TextProperty inside an annotion", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0, 3)+'xxx'+P1_TEXT.slice(5), 'Text should have been inserted correctly.');
  t.equal(s1.start.offset, 3, 'Annotation should have been moved.');
  t.equal(s1.end.offset, 6, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 6, 'Cursor should be after inserted text');
  t.ok(sel.isCollapsed(), '... collapsed');
  t.end();
});

test$14("IT8: Inserting text with range within TextProperty starting inside an annotion", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  doc.set(['s1', 'end', 'offset'], 6);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    endOffset: 6,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0, 4)+'xxx'+P1_TEXT.slice(6), 'Text should have been inserted correctly.');
  t.equal(s1.start.offset, 3, 'Annotation should have been moved.');
  t.equal(s1.end.offset, 7, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 7, 'Cursor should be after inserted text');
  t.ok(sel.isCollapsed(), '... collapsed');
  t.end();
});

test$14("IT9: Inserting text after an InlineNode node", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _il1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let il1 = doc.get('il1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'\uFEFF'+'Y'+P1_TEXT.slice(3), 'Text should have been inserted correctly.'); // eslint-disable-line no-useless-concat
  t.equal(sel.start.offset, 5, 'Cursor should be after inserted character');
  t.deepEqual([il1.start.offset, il1.end.offset], [3,4], 'InlineNode should have correct dimensions');
  t.end();
});

test$14("IT10: Typing over an InlineNode node", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _il1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    endOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });
  let p1 = doc.get('p1');
  let il1 = doc.get('il1');
  t.nil(il1, 'InlineNode should have been deleted.');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'Y'+P1_TEXT.slice(3), 'Text should have been inserted correctly.'); // eslint-disable-line no-useless-concat
  t.end();
});

test$14("IT11: Typing over a selected IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _in1);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });
  let body = doc.get('body');
  t.isNil(doc.get('in1'), 'IsolatedNode should have been deleted.');
  t.equal(body.getLength(), 1, 'There should be one node.');
  t.equal(body.getChildAt(0).getText(), 'Y', '.. containing the typed text.');
  t.end();
});

test$14("IT12: Typing before an IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _in1);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    mode: 'before',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });

  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should be two nodes.');
  t.equal(body.getChildAt(0).getText(), 'Y', '.. first one containing the typed text.');
  t.end();
});

test$14("IT13: Typing after an IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _in1);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    mode: 'after',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });

  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should be two nodes.');
  t.equal(body.getChildAt(1).getText(), 'Y', '.. second one containing the typed text.');
  t.end();
});

test$14("IT14: Typing over a ContainerSelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getPath(),
    startOffset: 3,
    endPath: p2.getPath(),
    endOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });

  let body = doc.get('body');
  t.deepEqual(body.nodes, ['p1'], 'Only the first paragraph should be left.');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'Y'+P2_TEXT.slice(4), 'The content should have been merged correctly');
  t.end();
});

test$14("IT15: Inserting text into a ListItem", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let li = doc.get('l1-1');
  t.equal(li.getText(), LI1_TEXT.slice(0,3)+'xxx'+LI1_TEXT.slice(3), 'Text should have been inserted correctly.');
  t.equal(sel.start.offset, 6, 'Cursor should be after inserted text');
  t.end();
});

test$14("IT16: Inserting text in an IsolatedNode", (t) => {
  let { doc, editorSession } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['in1', 'title'],
    startOffset: 3
  });

  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  });
  let in1 = doc.get('in1');
  t.equal(in1.title, IN1_TITLE.slice(0,3)+'xxx'+IN1_TITLE.slice(3), 'Text should be inserted into field');
  t.end();
});

test$14("II1: Inserting InlineNode node into a TextProperty", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertInlineNode({
      type: 'test-inline-node',
      id: 'il1',
      content: 'X'
    });
  });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let il1 = doc.get('il1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'\uFEFF'+P1_TEXT.slice(3), 'Text should have been inserted correctly.');
  t.equal(il1.start.offset, 3, 'Annotation should have been moved.');
  t.equal(il1.end.offset, 4, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 4, 'Cursor should be after inserted inline node');
  t.end();
});

test$14("IB2: Inserting BlockNode using cursor at start of a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  let p1 = doc.get('p1');
  t.equal(body.nodes[0], 'ib1', 'First node should be inserted block node.');
  t.equal(body.nodes[1], 'p1', 'Second node should be inserted block node.');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, p1.getTextPath(), '... on paragraph');
  t.equal(sel.start.offset, 0, '... first position');
  t.end();
});

test$14("IB3: Inserting an existing BlockNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.create({
      type: 'test-block',
      id: 'ib1'
    });
  });
  const ib1 = doc.get('ib1');
  editorSession.transaction((tx) => {
    tx.insertBlockNode(ib1);
  });
  let body = doc.get('body');
  t.ok(body.getChildAt(0) === ib1, 'First node should be the very block node instance created before.');
  t.end();
});

test$14("IB4: Inserting a BlockNode when an IsolatedNode is selected", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.equals(body.length, 3, 'There should be three nodes.');
  t.equals(body.nodes[1], 'ib1', 'The second one should be the inserted block node.');
  t.isNil(doc.get('in1'), 'The IsolatedNode should have been deleted.');
  t.end();
});

test$14("IB5: Inserting a BlockNode before an IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    mode: 'before',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.deepEqual(body.nodes, ['p1', 'ib1', 'in1', 'p2'], 'The body nodes should be in proper order.');
  t.end();
});

test$14("IB6: Inserting a BlockNode after an IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    mode: 'after',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.deepEqual(body.nodes, ['p1', 'in1', 'ib1', 'p2'], 'The body nodes should be in proper order.');
  t.end();
});

test$14("IB7: Inserting a BlockNode with an expanded PropertySelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'property',
    path: p1.getPath(),
    startOffset: 3,
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 3, 'There should be 3 nodes.');
  t.equal(body.nodes[1], 'ib1', 'The second should be the inserted block node.');
  t.equal(p1.getText(), P1_TEXT.slice(0, 3), 'The paragraph should have been truncated.');
  t.equal(body.getChildAt(2).getText(), P1_TEXT.slice(5), '.. and the tail stored in a new paragraph.');
  t.end();
});

test$14("IB8: Inserting a BlockNode into an empty paragraph", (t) => {
  let { editorSession, doc } = setupEditor(t, _empty);
  let empty = doc.get('empty');
  editorSession.setSelection({
    type: 'property',
    path: empty.getPath(),
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.deepEqual(body.nodes, ['ib1'], 'There should only be the inserted block node.');
  t.nil(doc.get('empty'), 'The empty paragraph should have been deleted.');
  t.end();
});

test$14("IB9: Inserting a BlockNode after a text node", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'property',
    path: p1.getPath(),
    startOffset: p1.getLength(),
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.deepEqual(body.nodes, ['p1', 'ib1'], 'The block node should have been inserted after the paragraph.');
  t.end();
});

test$14("IB10: Inserting a BlockNode with a collapsed ContainerSelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getPath(),
    startOffset: 3,
    endPath: p1.getPath(),
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 3, 'There should be three nodes.');
  t.equal(body.nodes[1], 'ib1', '.. the second one being the inserted block node.');
  t.end();
});

test$14("IB11: Inserting a BlockNode with a ContainerSelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getPath(),
    startOffset: 3,
    endPath: p2.getPath(),
    endOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.deepEqual(body.nodes, ['p1', 'ib1', 'p2'], 'There should be three nodes.');
  t.equal(p1.getText(), P1_TEXT.slice(0, 3), 'The first paragraph should have been truncated correctly');
  t.equal(p2.getText(), P2_TEXT.slice(4), 'The second paragraph should have been truncated correctly');
  t.end();
});

test$14("DEL0: Deleting without selection", (t) => {
  let { editorSession } = setupEditor(t, _p1, _p2);
  editorSession.setSelection(null);
  t.doesNotThrow(() => {
    editorSession.transaction((tx) => {
      tx.deleteCharacter('right');
    });
  }, 'Should not throw');
  t.end();
});

test$14("DEL1: Deleting a property selection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'property',
    path: p1.getPath(),
    startOffset: 0,
    endOffset: 3
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  t.equal(p1.getText(), P1_TEXT.slice(3), 'Parargaph should be truncated');
  t.end();
});

test$14("DEL2: Deleting using DELETE with cursor at the end of a TextNode at the end of a Container", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['p2', 'content'],
    startOffset: P2_TEXT.length,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  // nothing should have happened
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should still be 2 nodes in body');
  let p2 = doc.get('p2');
  t.equal(p2.getText(), P2_TEXT, 'p2 should still have same content.');
  t.equal(sel.start.offset, P2_TEXT.length, 'Cursor should still be at the same position');
  t.end();
});

test$14("DEL3: Deleting using DELETE with cursor in the middle of a TextProperty", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.transaction((tx) => {
    tx.setSelection({
      type: 'property',
      path: ['p1', 'content'],
      startOffset: 3
    });
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+P1_TEXT.slice(4), 'One character should have been deleted');
  t.equal(sel.start.offset, 3, 'Cursor should be at the same position');
  t.end();
});

test$14("DEL4: Deleting using DELETE with cursor inside an empty TextNode and TextNode as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _empty, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  let p2 = doc.get('p2');
  t.equal(body.nodes.length, 2, 'There should be only 2 nodes left.');
  t.equal(p2.getText(), P2_TEXT, 'p2 should not be affected');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, p2.getTextPath(), '... on p2');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("DEL5: Deleting using DELETE with cursor inside an empty TextNode and IsolatedNode as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _empty, _block1);
  editorSession.setSelection({
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should be only 2 nodes left.');
  t.ok(sel.isNodeSelection(), 'Selection should be a NodeSelection');
  t.equal(sel.getNodeId(), 'block1', '... block1');
  t.end();
});

test$14("DEL6: Deleting using DELETE with cursor inside an empty TextNode and List as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _empty, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should be only 2 nodes left.');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, ['l1-1', 'content'], '... on first list item');
  t.equal(sel.start.offset, 0,'... at first position');
  t.end();
});

test$14("DEL7: Deleting using DELETE with cursor at the end of a non-empty TextNode and TextNode as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: P1_TEXT.length,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  let p1 = doc.get('p1');
  t.equal(body.nodes.length, 1, 'There should be only one node left.');
  t.ok(sel.isCollapsed(), 'Selection should be a collapsed');
  t.deepEqual(sel.start.path, p1.getTextPath(), '... on p1');
  t.equal(sel.start.offset, P1_TEXT.length, '... at the same position as before');
  t.end();
});

test$14("DEL8: Deleting using DELETE with cursor at the end of a non-empty TextNode and IsolatedNode as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: P1_TEXT.length,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  // NOTE: if there is no merge possible, nothing should happen
  // only the selection should be updated
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should still be 2 nodes left.');
  t.ok(sel.isNodeSelection(), 'Selection should be a NodeSelection');
  t.equal(sel.getNodeId(), 'block1', '... in block1');
  t.ok(sel.isFull(), '... selecting the whole node');
  t.end();
});

test$14("DEL9: Deleting using DELETE with cursor after an IsolatedNode and a TextNode as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _block1, _p1);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'block1',
    mode: 'after',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let p1 = doc.get('p1');
  t.equal(p1.getText(), P1_TEXT.slice(1), 'First character of paragraph should have been deleted.');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: p1.getPath(),
    startOffset: 0
  });
  t.end();
});

test$14("DEL10: Deleting using DELETE with cursor after an IsolatedNode and an IsolatedNode as successor", (t) => {
  let { editorSession } = setupEditor(t, _block1, _block2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'block1',
    mode: 'after',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  _checkSelection(t, editorSession.getSelection(), {
    type: 'node',
    nodeId: 'block2',
    mode: 'full'
  });
  t.end();
});

test$14("DEL11: Deleting using BACKSPACE with cursor in the middle of a TextProperty", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+P1_TEXT.slice(4), 'one character should have been deleted');
  t.equal(sel.start.offset, 3, 'Cursor should have shifted by one character');
  t.end();
});

test$14("DEL12: Deleting using BACKSPACE with cursor inside an empty TextNode and TextNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _empty, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.isNil(doc.get('empty'), 'empty node should have been deleted');
  t.equal(p1.getText(), P1_TEXT, 'p1 should be untouched');
  t.deepEqual(sel.start.path, p1.getTextPath(), 'Cursor should be in p1');
  t.equal(sel.start.offset, P1_TEXT.length, '... at last position');
  t.end();
});

test$14("DEL13: Deleting using BACKSPACE with cursor inside an empty TextNode and IsolatedNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _block1, _empty, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  t.isNil(doc.get('empty'), 'empty node should have been deleted');
  t.ok(sel.isNodeSelection(), 'Selection should be a node selection');
  t.equal(sel.getNodeId(), 'block1', '... on block1');
  t.end();
});

test$14("DEL14: Deleting using BACKSPACE with cursor at the start of a non-empty TextNode and TextNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['p2', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.isNil(doc.get('p2'), 'p2 should have been deleted');
  t.equal(p1.getText(), P1_TEXT+P2_TEXT, 'Text should have been merged');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, p1.getTextPath(), '... on p1');
  t.ok(sel.start.offset, P1_TEXT.length, '... cursor should after the original text of p1');
  t.end();
});

test$14("DEL15: Deleting using BACKSPACE with cursor at the start of a non-empty TextNode and IsolatedNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _block1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['p2', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should still be two nodes');
  t.ok(sel.isNodeSelection(), 'Selection should be a NodeSelection');
  t.equal(sel.getNodeId(), 'block1', '... on block1');
  t.ok(sel.isFull(), '... selecting the full node');
  t.end();
});

test$14("DEL16: Deleting using BACKSPACE with cursor after IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _block1, _p2);
  editorSession.setSelection({
    type: 'node',
    mode: 'after',
    nodeId: 'block1',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.isNil(doc.get('block1'), 'IsolatedNode should have been deleted');
  t.equal(body.getLength(), 2, 'There should still be two nodes');
  let pnew = body.getChildAt(0);
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, pnew.getTextPath(), '... on new paragraph');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("DEL17: Deleting using BACKSPACE with cursor before IsolatedNode with TextNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block2);
  editorSession.setSelection({
    type: 'node',
    mode: 'before',
    nodeId: 'block2',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  let p1 = doc.get('p1');
  t.equal(body.getLength(), 2, 'There should still be two nodes');
  t.equal(p1.getText(), P1_TEXT.slice(0, -1), 'Last character of p1 should have been deleted');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, p1.getTextPath(), '... on p1');
  t.equal(sel.start.offset, P1_TEXT.length-1, '... at last position');
  t.end();
});

test$14("DEL18: Deleting using BACKSPACE with cursor before IsolatedNode and IsolatedNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _block1, _block2);
  editorSession.setSelection({
    type: 'node',
    mode: 'before',
    nodeId: 'block2',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should still be two nodes');
  t.ok(sel.isNodeSelection(), 'Selection should be a NodeSelection');
  t.equal(sel.getNodeId(), 'block1', '... on block1');
  t.ok(sel.isFull(), '... selection the entire node');
  t.end();
});

test$14("DEL19: Deleting an entirely selected IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'block1',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.isNil(doc.get('block1'), 'IsolatedNode should have been deleted');
  let pnew = body.getChildAt(1);
  t.equal(body.getLength(), 3, 'There should be 3 nodes');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, pnew.getTextPath(), '... on new paragraph');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("DEL20: Deleting a range starting before a TextNode and ending after a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _block2, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getTextPath(),
    startOffset: 0,
    endPath: p2.getTextPath(),
    endOffset: p2.getLength(),
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be only one node left');
  let first = body.getChildAt(0);
  t.ok(first.isText(), '... which is a TextNode');
  t.ok(first.isEmpty(), '... which is empty');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, first.getTextPath(), '... on that TextNode');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("DEL21: Deleting a range starting in the middle of a TextNode and ending after a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _block2, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getTextPath(),
    startOffset: 3,
    endPath: p2.getTextPath(),
    endOffset: p2.getLength(),
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be only one node left');
  let first = body.getChildAt(0);
  t.ok(first.isText(), '... which is a TextNode');
  t.equal(first.getText(), P1_TEXT.slice(0, 3), '... with truncated content');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, first.getTextPath(), '... on that TextNode');
  t.equal(sel.start.offset, 3, '... at last position');
  t.end();
});

test$14("DEL22: Deleting a range starting before a TextNode and ending in the middle of a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _block2, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getTextPath(),
    startOffset: 0,
    endPath: p2.getTextPath(),
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be only one node left');
  let first = body.getChildAt(0);
  t.ok(first.isText(), '... which is a TextNode');
  t.equal(first.getText(), P2_TEXT.slice(3), '... with sliced content');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, first.getTextPath(), '... on that TextNode');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("DEL23: Deleting a range starting in the middle of a TextNode and ending in the middle of a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _block2, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getTextPath(),
    startOffset: 3,
    endPath: p2.getTextPath(),
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be only one node left');
  let first = body.getChildAt(0);
  t.ok(first.isText(), '... which is a TextNode');
  t.equal(first.getText(), P1_TEXT.slice(0,3)+P2_TEXT.slice(3), '... with merged content');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, first.getTextPath(), '... on that TextNode');
  t.equal(sel.start.offset, 3, '... at correct position');
  t.end();
});

test$14("DEL24: Deleting a range starting in the middle of a TextNode and ending in the middle of a ListItem", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _block2, _l1, _l11, _l12);
  let p1 = doc.get('p1');
  let l1 = doc.get('l1');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getTextPath(),
    startOffset: 3,
    endPath: ['l1-1', 'content'],
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should be 2 nodes left');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+LI1_TEXT.slice(3), '... with merged content');
  t.equal(l1.items.length, 1, 'The list should have only 1 item left');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, p1.getTextPath(), '... on p1');
  t.equal(sel.start.offset, 3, '... at correct position');
  t.end();
});

test$14("DEL25: Deleting a range starting in the middle of a ListItem and ending in the middle of a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12, _block1, _block2, _p1);
  let p1 = doc.get('p1');
  let l1 = doc.get('l1');
  editorSession.setSelection({
    type: 'container',
    startPath: ['l1-2', 'content'],
    startOffset: 3,
    endPath: p1.getTextPath(),
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be 1 node left');
  t.equal(l1.items.length, 2, 'The list should have 2 items');
  let li2 = l1.getItemAt(1);
  t.equal(li2.getText(), LI2_TEXT.slice(0,3)+P1_TEXT.slice(3), 'The second item should container merged content');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, li2.getTextPath(), '... on second list item');
  t.equal(sel.start.offset, 3, '... at correct position');
  t.end();
});

test$14("DEL26: Deleting a range within a ListItem", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  let l1 = doc.get('l1');
  editorSession.setSelection({
    type: 'container',
    startPath: ['l1-2', 'content'],
    startOffset: 3,
    endPath: ['l1-2', 'content'],
    endOffset: 6,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  t.equal(l1.items.length, 2, 'The list should have 2 items');
  let li2 = l1.getItemAt(1);
  t.equal(li2.getText(), LI2_TEXT.slice(0,3)+LI2_TEXT.slice(6), 'The second item should be changed');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, li2.getTextPath(), '... on second list item');
  t.equal(sel.start.offset, 3, '... at correct position');
  t.end();
});

test$14("DEL27: Deleting a range across two ListItems within the same List", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  let l1 = doc.get('l1');
  editorSession.setSelection({
    type: 'container',
    startPath: ['l1-1', 'content'],
    startOffset: 3,
    endPath: ['l1-2', 'content'],
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  t.equal(l1.items.length, 1, 'The list should have 1 item');
  let li1 = l1.getItemAt(0);
  t.equal(li1.getText(), LI1_TEXT.slice(0,3)+LI2_TEXT.slice(3), 'The items should be merged');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, li1.getTextPath(), '... on the list item');
  t.equal(sel.start.offset, 3, '... at correct position');
  t.end();
});

test$14("DEL28: Deleting a ContainerSelection within a single TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getPath(),
    startOffset: 3,
    endPath: p1.getPath(),
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+P1_TEXT.slice(5), 'The text should have been deleted');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: p1.getPath(),
    startOffset: 3
  });
  t.end();
});

test$14("DEL29: Deleting a character inside an IsolatedNode", (t) => {
  let { doc, editorSession } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['in1', 'title'],
    startOffset: 3
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let in1 = doc.get('in1');
  t.equal(in1.title, IN1_TITLE.slice(0,3)+IN1_TITLE.slice(4), 'Text should be inserted into field');
  t.end();
});

test$14("DEL30: Merging two ListItems using DELETE", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: LI1_TEXT.length,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let l = doc.get('l1');
  t.equal(l.items.length, 1, 'Only one list item should be left');
  let li = l.getItemAt(0);
  t.equal(li.getText(), LI1_TEXT+LI2_TEXT, 'The list item should have the merged text');
  t.ok(sel.isCollapsed(), 'The selection should be collapsed');
  t.deepEqual(sel.start.path, li.getTextPath(), '... on the list item');
  t.equal(sel.start.offset, LI1_TEXT.length, '... at the end of the original content');
  t.end();
});

test$14("DEL31: Merging a List into previous List using DELETE", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12, _l2, _l21, _l22);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-2', 'content'],
    startOffset: LI2_TEXT.length,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let l1 = doc.get('l1');
  t.equal(l1.items.length, 4, 'First list should have 4 items');
  t.ok(sel.isCollapsed(), 'The selection should be collapsed');
  t.deepEqual(sel.start.path, ['l1-2', 'content'], '... on the same item');
  t.equal(sel.start.offset, LI2_TEXT.length, '... at the same position');
  t.end();
});

test$14("DEL32: Merging an empty List into previous TextNode using DELETE", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _l1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'property',
    path: p1.getPath(),
    startOffset: p1.getLength(),
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  t.isNil(doc.get('l2'), 'Second list should have been removed.');
  t.end();
});

test$14("BR1: Breaking without selection", (t) => {
  let { editorSession } = setupEditor(t, _p1);
  editorSession.setSelection(null);
  t.doesNotThrow(() => {
    editorSession.transaction((tx) => {
      tx.break();
    });
  });
  t.end();
});

test$14("BR2: Breaking a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  t.equals(body.length, 2, 'There should be 2 nodes');
  let first = body.getChildAt(0);
  let second = body.getChildAt(1);
  t.equals(second.type, 'paragraph', 'Second should be a paragraph');
  t.equals(first.getText(), P1_TEXT.slice(0,3), 'First should be truncated');
  t.equals(second.getText(), P1_TEXT.slice(3), '.. and second should contain the tail');
  t.end();
});

test$14("BR3: Breaking annotated text with cursor before the annotation", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  let second = body.getChildAt(1);
  let anno = doc.get('s1');
  t.deepEqual(anno.start.path, second.getPath(), 'Annotation should now be on the second paragraph');
  t.equal(anno.start.offset, 2, '.. starting at character 2');
  t.equal(anno.end.offset, 4, '.. ending at character 4');
  t.end();
});

test$14("BR4: Breaking annotated text with cursor inside the annotation", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  let secondPara = body.getChildAt(1);
  let anno = doc.get('s1');
  t.equal(anno.start.offset, 3, '.. starting at character 3');
  t.equal(anno.end.offset, 4, '.. ending at character 4');
  let secondAnno = doc.getIndex('annotations').get([secondPara.id, 'content'])[0];
  t.equal(secondAnno.start.offset, 0, '.. starting at character 0');
  t.equal(secondAnno.end.offset, 1, '.. ending at character 1');
  t.end();
});

test$14("BR5: Breaking a TextNode at the first position", (t) => {
  let { editorSession, doc } = setupEditor(t, _h1);
  let h1 = doc.get('h1');
  editorSession.setSelection({
    type: 'property',
    path: h1.getPath(),
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  t.equals(body.length, 2, 'There should be 2 nodes');
  let first = body.getChildAt(0);
  let second = body.getChildAt(1);
  t.equals(second, h1, 'Second should be the original node');
  t.equals(first.type, h1.type, 'First should be of the same type');
  t.equals(first.getText(), '', '.. but empty');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: h1.getPath(),
    startOffset: 0
  });
  t.end();
});

test$14("BR5: Breaking a TextNode at the last position", (t) => {
  let { editorSession, doc } = setupEditor(t, _h1);
  let h1 = doc.get('h1');
  editorSession.setSelection({
    type: 'property',
    path: h1.getPath(),
    startOffset: h1.getLength(),
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  t.equals(body.length, 2, 'There should be 2 nodes');
  let first = body.getChildAt(0);
  let second = body.getChildAt(1);
  t.equals(first, h1, 'First should be the original node');
  t.equals(second.type, 'paragraph', 'Second should be a paragraph.');
  t.equals(second.getText(), '', '.. without content');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: second.getPath(),
    startOffset: 0
  });
  t.end();
});

test$14("BR6: Breaking a ContainerSelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  editorSession.setSelection({
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 3,
    endPath: ['p2', 'content'],
    endOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.break();
  });
  let p2 = doc.get('p2');
  t.equal(p2.getText(), P2_TEXT.slice(4));
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: ['p2', 'content'],
    startOffset: 0,
    endOffset: 0
  });
  t.end();
});

test$14("BR7: Breaking a NodeSelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.break();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 4, 'There should be one more node');
  let newP = body.getChildAt(2);
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: newP.getPath(),
    startOffset: 0,
    endOffset: 0
  });
  t.end();
});

test$14("BR8: Breaking a NodeSelection (before)", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    mode: 'before',
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.break();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 4, 'There should be one more node');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'node',
    nodeId: 'in1',
    mode: 'before',
    containerId: 'body'
  });
  t.end();
});

test$14("BR9: Breaking with an expanded PropertySelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['p1','content'],
    startOffset: 3,
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.break();
  });
  let body = doc.get('body');
  let first = body.getChildAt(0);
  let second = body.getChildAt(1);
  t.equal(first.getText(), P1_TEXT.slice(0, 3), 'First paragraph should be truncated');
  t.equal(second.getText(), P1_TEXT.slice(5), '.. and tail should be moved to second paragraph');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: second.getPath(),
    startOffset: 0
  });
  t.end();
});

test$14("BR10: Breaking a ListItem with cursor in the middle of text", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let sel = editorSession.getSelection();
  let l = doc.get('l1');
  t.equal(l.items.length, 3, 'List should have 3 items');
  let li1 = doc.get(l.items[0]);
  let li2 = doc.get(l.items[1]);
  t.equal(li1.getText(), LI1_TEXT.slice(0,3), 'First item should have been truncated.');
  t.equal(li2.getText(), LI1_TEXT.slice(3), 'remaining line should have been inserted into new list item.');
  t.deepEqual(sel.start.path, li2.getTextPath(), 'Cursor should in second item');
  t.equal(sel.start.offset, 0, 'Cursor should be at begin of item.');
  t.end();
});

test$14("BR11: Breaking a ListItem with cursor at begin of text", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let sel = editorSession.getSelection();
  let l = doc.get('l1');
  t.equal(l.items.length, 3, 'List should have 3 items');
  let li1 = doc.get(l.items[0]);
  let li2 = doc.get(l.items[1]);
  t.equal(li1.getText(), '', 'First item should be empty.');
  t.equal(li2.getText(), LI1_TEXT, 'Text should have moved to next item.');
  t.deepEqual(sel.start.path, li2.getTextPath(), 'Cursor should be in second item');
  t.equal(sel.start.offset, 0, '... at begin of item.');
  t.end();
});

test$14("BR12: Splitting a List by breaking an empty ListItem", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l1_empty, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 3, 'There should be three nodes');
  let l1 = body.getChildAt(0);
  let p = body.getChildAt(1);
  let l2 = body.getChildAt(2);
  t.ok(l1.isList() && p.isText() && l2.isList(), '... list, paragraph, and list');
  t.equal(l1.items.length, 1, 'The first list should now have only one item');
  t.equal(l1.items[0], 'l1-1', '... with id "li-1"');
  t.equal(p.getText(), '', 'The paragraph should be empty');
  t.equal(l2.items.length, 1, 'The second list should now have only one item');
  t.equal(l2.items[0], 'l1-2', '... with id "li-2"');
  t.ok(sel.isCollapsed(), 'The selection should be collapsed');
  t.deepEqual(sel.start.path, p.getTextPath(), '... on the new paragraph');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("BR13: Breaking the last empty list item", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12, _l1_empty);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should be two nodes');
  let l1 = body.getChildAt(0);
  let p = body.getChildAt(1);
  t.ok(l1.isList() && p.isText(), '... list, and paragraph');
  t.equal(l1.items.length, 2, 'The list should now have only two items');
  t.equal(p.getText(), '', 'The paragraph should be empty');
  t.ok(sel.isCollapsed(), 'The selection should be collapsed');
  t.deepEqual(sel.start.path, p.getTextPath(), '... on the new paragraph');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("BR14: Breaking a list with only one empty item", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l1_empty);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be only one node left');
  let p = body.getChildAt(0);
  t.equal(p.type, 'paragraph', '... of type paragraph');
  t.isNil(doc.get('l1'), 'The list should have been removed');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: p.getPath(),
    startOffset: 0
  });
  t.end();
});

test$14("BR15: Breaking a list with a first empty item", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l1_empty, _l11);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should be two nodes now');
  let p = body.getChildAt(0);
  let l1 = body.getChildAt(1);
  t.equal(p.type, 'paragraph', 'First should be a paragraph');
  t.equal(p.getText(), '', '.. without content');
  t.equal(l1.type, 'list', 'Second should be a list');
  t.equal(l1.getLength(), 1, '.. with only one item left.');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: p.getPath(),
    startOffset: 0
  });
  t.end();
});


test$14("L5-1: Toggling a ListItem using BACKSPACE", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-2', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should be two nodes now');
  let l = doc.get('l1');
  let p = body.getChildAt(1);
  t.equal(l.items.length, 1, 'Only one list item should be left');
  t.equal(p.type, 'paragraph', 'The second one should be a paragraph');
  t.equal(p.getText(), LI2_TEXT, '.. with the text of the second item');
  t.ok(sel.isCollapsed(), 'The selection should be collapsed');
  t.deepEqual(sel.start.path, p.getTextPath(), '... on the new paragraph');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("L8-1: Toggling a paragraph into a List", (t) => {
  let { doc, editorSession} = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.toggleList({ ordered: false});
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 1, 'There should be 1 node');
  let l = body.getNodeAt(0);
  t.equal(l.type, 'list', '.. a list');
  t.equal(l.ordered, false, '.. unordered');
  t.equal(l.getLength(), 1, '.. with one item');
  let li = l.getItemAt(0);
  t.equal(li.getText(), P1_TEXT, '.. with text of p1');
  t.end();
});

test$14("L8-2: Toggling first list-item into a paragraph", (t) => {
  let { doc, editorSession} = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.toggleList();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should be 2 nodes');
  let p = body.getNodeAt(0);
  let l = body.getNodeAt(1);
  t.equal(p.type, 'paragraph', '.. first should be paragraph');
  t.equal(p.getText(), LI1_TEXT, '.. with text of first list item');
  t.equal(l.type, 'list', '.. second should be alist');
  t.equal(l.getLength(), 1, '.. with one item');
  t.end();
});

test$14("L8-3: Toggling last list-item into a paragraph", (t) => {
  let { doc, editorSession} = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-2', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.toggleList();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should be 2 nodes');
  let l = body.getNodeAt(0);
  let p = body.getNodeAt(1);
  t.equal(l.type, 'list', '.. first should be alist');
  t.equal(l.getLength(), 1, '.. with one item');
  t.equal(p.type, 'paragraph', '.. second should be paragraph');
  t.equal(p.getText(), LI2_TEXT, '.. with text of second list item');
  t.end();
});

test$14("L8-4: Toggling a middle list-item into a paragraph splitting the list", (t) => {
  let { doc, editorSession} = setupEditor(t, _l1, _l11, _l13, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-3', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.toggleList();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 3, 'There should be 3 nodes');
  let l = body.getNodeAt(0);
  let p = body.getNodeAt(1);
  let l2 = body.getNodeAt(2);
  t.equal(l.type, 'list', '.. first should be alist');
  t.equal(l.getLength(), 1, '.. with one item');
  t.equal(p.type, 'paragraph', '.. second should be paragraph');
  t.equal(p.getText(), LI3_TEXT, '.. with text of second list item');
  t.equal(l2.type, 'list', '.. last should be a list again');
  t.equal(l2.getLength(), 1, '.. with one item');
  t.end();
});

test$14("L8-5: Toggling the only list-item of a list into a paragraph", (t) => {
  let { doc, editorSession} = setupEditor(t, _l1, _l11);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.toggleList();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 1, 'There should be one node left');
  let p = body.getNodeAt(0);
  t.equal(p.type, 'paragraph', '.. which should be paragraph');
  t.equal(p.getText(), LI1_TEXT, '.. with text of first list item');
  t.end();
});

test$14("IND1: Indenting a ListItem", (t) => {
  let { doc, editorSession } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.indent();
  });
  let li1 = doc.get('l1-1');
  t.equal(li1.level, 2, 'Indentation level should have increased');
  t.end();
});

test$14("IND2: Dedenting a ListItem", (t) => {
  let { doc, editorSession } = setupEditor(t, _l1, _l11, _l12, _li1plus);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.dedent();
  });
  let li1 = doc.get('l1-1');
  t.equal(li1.level, 1, 'Indentation level should have decreased');
  t.end();
});

test$14("CP1: Copying a property selection", (t) => {
  let { doc } = setupEditor(t, headersAndParagraphs);
  let editor = new substance.EditingInterface(doc);
  editor.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    endOffset: 9,
    containerId: 'body'
  });
  let copy = editor.copySelection();
  let textNode = copy.get(substance.Document.TEXT_SNIPPET_ID);
  t.notNil(textNode, 'There should be a text node for the property fragment.');
  t.equal(textNode.content, 'graph', 'Selected text should be copied.');
  t.end();
});

test$14("CP2: Copying a property selection with annotated text", (t) => {
  let { doc } = setupEditor(t, headersAndParagraphs);
  let editor = new substance.EditingInterface(doc);
  editor.setSelection({
    type: 'property',
    path: ['p2', 'content'],
    startOffset: 10,
    endOffset: 19,
    containerId: 'body'
  });
  let copy = editor.copySelection();
  t.equal(copy.get([substance.Document.TEXT_SNIPPET_ID, 'content']), 'with anno', 'Selected text should be copied.');
  let annos = copy.getIndex('annotations').get([substance.Document.TEXT_SNIPPET_ID, 'content']);
  t.equal(annos.length, 1, 'There should be one annotation on copied text.');
  let anno = annos[0];
  t.equal(anno.type, "emphasis", "The annotation should be 'emphasis'.");
  t.deepEqual([anno.start.offset, anno.end.offset], [5, 9], 'The annotation should be over the text "anno".');
  t.end();
});

test$14("CP3: Copying a container selection", (t) => {
  let { doc } = setupEditor(t, headersAndParagraphs);
  let editor = new substance.EditingInterface(doc);
  editor.setSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['h1', 'content'],
    startOffset: 4,
    endPath: ['p2', 'content'],
    endOffset: 9,
  });
  let copy = editor.copySelection();
  let content = copy.get(substance.Document.SNIPPET_ID);
  t.notNil(content, 'There should be a container node with id "content".');
  // 4 nodes? 'body', 'snippets', 'p1', 'p2'
  t.equal(content.nodes.length, 4, 'There should be 4 nodes in the copied document.');
  let first = copy.get(content.nodes[0]);
  t.equal(first.type, 'heading', "The first node should be a heading.");
  t.equal(first.content, 'ion 1', "Its content should be truncated to 'ion 1'.");
  let last = copy.get(content.nodes[3]);
  t.equal(last.type, 'paragraph', "The last node should be a paragraph.");
  t.equal(last.content, 'Paragraph', "Its content should be truncated to 'Paragraph'.");
  t.end();
});

test$14("CP4: Copying a paragraph", (t) => {
  let { doc } = setupEditor(t, _p1, _p2);
  let editor = new substance.EditingInterface(doc);
  editor.setSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2'],
    startOffset: 0,
    endPath: ['p2'],
    endOffset: 1
  });
  let copy = editor.copySelection();
  let p2 = copy.get('p2');
  t.equal(p2.content, doc.get('p2').content, 'The whole paragraph should be copied.');
  t.end();
});

test$14("CP5: Copying two ListItems", (t) => {
  let { doc } = setupEditor(t, _l1, _l11, _l12);
  let editor = new substance.EditingInterface(doc);
  editor.setSelection({
    type: 'container',
    startPath: ['l1-1', 'content'],
    startOffset: 3,
    endPath: ['l1-2', 'content'],
    endOffset: 3,
    containerId: 'body'
  });
  let copy = editor.copySelection();
  let content = copy.getContainer();
  t.equal(content.getLength(), 1, 'There should be one node');
  let l1 = copy.get('l1');
  t.notNil(l1, 'l1 should exist');
  t.equal(l1.getLength(), 2, '.. having 2 items');
  let li1 = l1.getItemAt(0);
  let li2 = l1.getItemAt(1);
  t.equal(li1.getText(), LI1_TEXT.slice(3), 'The first item should have correct content');
  t.equal(li2.getText(), LI2_TEXT.slice(0, 3), 'The second item should have correct content');
  t.end();
});

test$14("CP6: Copying a paragraph and a ListItem", (t) => {
  let { doc } = setupEditor(t, _p1, _l1, _l11, _l12);
  let editor = new substance.EditingInterface(doc);
  editor.setSelection({
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 3,
    endPath: ['l1-1', 'content'],
    endOffset: 3,
    containerId: 'body'
  });
  let copy = editor.copySelection();
  let content = copy.getContainer();
  t.equal(content.getLength(), 2, 'There should be 2 nodes');
  let l1 = copy.get('l1');
  t.notNil(l1, 'l1 should exist');
  t.equal(l1.getLength(), 1, '.. having 1 item');
  let li1 = l1.getItemAt(0);
  t.equal(li1.getText(), LI1_TEXT.slice(0, 3), 'The first item should have correct content');
  t.end();
});

test$14("CP7 Copying a table", (t) => {
  let { doc } = setupEditor(t, _t1);
  let editor = new substance.EditingInterface(doc);
  editor.setSelection({
    type: 'node',
    nodeId: 't1',
    containerId: 'body',
  });
  let copy = editor.copySelection();
  let t1 = copy.get('t1');
  t.notNil(t1, 'Table node should exist');
  t.equals(t1.getRowCount(), 2, '.. with two rows');
  t.equals(t1.getColCount(), 2, '.. with two cols');
  for (let row = 0; row < 2; row++) {
    for (let col = 0; col < 2; col++) {
      let cell = t1.getCellAt(row,col);
      t.notNil(cell, `Cell (${row+1},${col+1}) should exist`);
      t.equals(cell.content, T_CONTENT[row][col], '.. and have correct content');
    }
  }
  t.end();
});

test$14("CP8 Copying a sparse table", (t) => {
  let { doc } = setupEditor(t, _t1_sparse);
  let editor = new substance.EditingInterface(doc);
  editor.setSelection({
    type: 'node',
    nodeId: 't1',
    containerId: 'body',
  });
  let copy = editor.copySelection();
  let t1 = copy.get('t1');
  t.notNil(t1, 'Table node should exist');
  t.equals(t1.getRowCount(), 2, '.. with two rows');
  t.equals(t1.getColCount(), 2, '.. with two cols');
  for (let i = 0; i < 2; i++) {
    let cell = t1.getCellAt(i,i);
    t.notNil(cell, `Cell (${i+1},${i+1}) should exist`);
    t.equals(cell.content, T_CONTENT[i][i], '.. and have correct content');
    t.nil(t1.getCellAt(1-i, i), `Cell (${1-i+1},${i+1}) should not exist`);
  }
  t.end();
});

// NOTE: with Copy'n'Paste within the same document,
// we must make sure to check that things are cloned correctly,
// e.g., not that inadvertantly the copy references one of the original children

test$14("CP9: Copy and Pasting a List", (t) => {
  let { doc, editorSession } = setupEditor(t, _p1, _empty, _l1, _l11, _l12);
  editorSession.transaction((tx) => {
    tx.setSelection({
      type: 'container',
      startPath: ['l1-1', 'content'],
      startOffset: 0,
      endPath: ['l1-2', 'content'],
      endOffset: LI2_TEXT.length,
      containerId: 'body'
    });
    let copy = tx.copySelection();
    tx.setSelection({
      type: 'property',
      path: ['empty', 'content'],
      startOffset: 0,
      containerId: 'body'
    });
    tx.paste(copy);
  });
  let body = doc.get('body');
  let nodes = body.getNodes();
  t.equal(body.length, 3, "There should be 3 nodes");
  t.deepEqual(['paragraph', 'list', 'list'], nodes.map(n => n.type), '.. a paragraph and 2 lists');
  let li1 = body.getChildAt(1);
  let li2 = body.getChildAt(2);
  t.notDeepEqual(li1.items, li2.items, 'List items should not be the same.');
  t.equal(li1.getItemAt(0).getText(), li2.getItemAt(0).getText(), 'First list item should have same content');
  t.equal(li1.getItemAt(1).getText(), li2.getItemAt(1).getText(), 'Second list item should have same content');
  t.end();
});

test$14("CP10: Copy and Pasting a List partially", (t) => {
  let { doc, editorSession } = setupEditor(t, _p1, _empty, _l1, _l11, _l12);
  editorSession.transaction((tx) => {
    tx.setSelection({
      type: 'container',
      startPath: ['l1-1', 'content'],
      startOffset: 3,
      endPath: ['l1-2', 'content'],
      endOffset: 3,
      containerId: 'body'
    });
    let copy = tx.copySelection();
    tx.setSelection({
      type: 'property',
      path: ['empty', 'content'],
      startOffset: 0,
      containerId: 'body'
    });
    tx.paste(copy);
  });
  let body = doc.get('body');
  let nodes = body.getNodes();
  t.equal(body.length, 3, "There should be 3 nodes");
  t.deepEqual(['paragraph', 'list', 'list'], nodes.map(n => n.type), '.. a paragraph and 2 lists');
  let li1 = body.getChildAt(1);
  let li2 = body.getChildAt(2);
  t.notDeepEqual(li1.items, li2.items, 'List items should not be the same.');
  t.equal(li1.getItemAt(0).getText(), LI1_TEXT.slice(3), 'First list item should be truncated');
  t.equal(li1.getItemAt(1).getText(), LI2_TEXT.slice(0,3), 'Second list item should be truncated');
  t.end();
});


test$14("AN1: Annotating without unknown annotation type", (t) => {
  let { editorSession } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    endOffset: 3
  });
  t.throws(() => {
    editorSession.transaction((tx)=>{
      tx.annotate({type: 'unknown-annotation'});
    });
  }, 'Should throw an exception');
  t.end();
});

test$14("AN2: Tying to annotate using a non-annotation type", (t) => {
  let { editorSession } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    endOffset: 3
  });
  t.throws(() => {
    editorSession.transaction((tx)=>{
      tx.annotate({type: 'paragraph'});
    });
  }, 'Should throw an exception');
  t.end();
});

test$14("ST1: Switching text type", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'property',
    path: p1.getPath(),
    startOffset: 0,
    endOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.switchTextType({type: 'heading', level: 1});
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 1, 'There should be one node.');
  let first = body.getChildAt(0);
  t.equal(first.type, 'heading', '.. of type heading');
  t.equal(first.getText(), P1_TEXT, '.. with correct content');
  let annos = doc.getAnnotations(first.getPath());
  t.equal(annos.length, 1, '.. and with one annotation');
  t.end();
});

function _checkSelection(t, actual, exp) {
  let _data = actual.toJSON();
  let data = {};
  substance.forEach(exp, (value, key) => {
    data[key] = _data[key];
  });
  t.deepEqual(data, exp, 'Selection should be correct');
}

const test$15 = substanceTest.module('EditorSession');

test$15("Keeping TransactionDocument up-to-date.", function(t) {
  let { editorSession, doc } = setupEditor(t, simple);
  let stageDoc = editorSession._transaction._stageDoc;
  stageDoc._apply = substanceTest.spy(stageDoc, '_apply');
  doc.create({ type: 'paragraph', id: 'foo', content: 'foo'});
  var p = stageDoc.get('foo');
  t.equal(stageDoc._apply.callCount, 1, "Stage should have been updated.");
  t.notNil(p, "Stage should contain new paragraph node.");
  t.equal(p.content, "foo");
  t.end();
});

test$15("Undoing and redoing a change.", function(t) {
  let { editorSession, doc } = setupEditor(t, simple);
  editorSession.transaction(function(tx) {
    tx.update(['p1', 'content'], { type: 'insert', start: 3, text: "XXX" });
  });
  t.equal(doc.get(['p1', 'content']), '012XXX3456789', 'Text should have been inserted.');
  t.equal(editorSession.canUndo(), true, 'Undo should be possible');
  editorSession.undo();
  t.equal(doc.get(['p1', 'content']), '0123456789', 'Original text should have been recovered.');
  t.equal(editorSession.canUndo(), false, 'Undo should be disabled');
  t.equal(editorSession.canRedo(), true, 'Redo should be possible');
  editorSession.redo();
  t.equal(doc.get(['p1', 'content']), '012XXX3456789', 'Text should have been changed again.');
  t.end();
});

test$15("Selections after undo/redo.", function(t) {
  let { editorSession, doc } = setupEditor(t, simple);
  var path = ['p1', 'content'];
  editorSession.setSelection({
    type: 'property',
    path: path,
    startOffset: 3
  });
  editorSession.transaction(function(tx) {
    tx.update(path, { type: 'insert', start: 3, text: "XXX" });
    tx.setSelection({
      type: 'property',
      path: path,
      startOffset: 6
    });
  });
  editorSession.undo();
  var sel = editorSession.getSelection();
  t.ok(sel.equals(doc.createSelection({
    type: 'property',
    path: path,
    startOffset: 3
  })), 'Selection should be set correctly after undo.');
  editorSession.redo();
  sel = editorSession.getSelection();
  t.ok(sel.equals(doc.createSelection({
    type: 'property',
    path: path,
    startOffset: 6
  })), 'Selection should be set correctly after redo.');
  t.end();
});

const test$16 = substanceTest.module('Fragmenter');

var TEXT = 'ABCDEFGHI';

test$16("No annos.", function(t) {
  var annos = [];
  var html = _render(TEXT, annos);
  t.equal(html, TEXT);
  t.end();
});

test$16("With one anno.", function(t) {
  var annos = [new Anno('b', 'b1', 3, 6)];
  var html = _render(TEXT, annos);
  t.equal(html, 'ABC<b>DEF</b>GHI');
  t.end();
});

test$16("With one anchor.", function(t) {
  var annos = [new Anno('a', 'a1', 3, 3, {
    isAnchor: true
  })];
  var html = _render(TEXT, annos);
  t.equal(html, 'ABC<a></a>DEFGHI');
  t.end();
});

test$16("With one inline.", function(t) {
  var annos = [new Anno('i', 'i1', 3, 4)];
  var html = _render(TEXT, annos);
  t.equal(html, 'ABC<i>D</i>EFGHI');
  t.end();
});

test$16("One nested anno.", function(t) {
  var annos = [new Anno('b', 'b1', 3, 6), new Anno('i', 'i1', 4, 5)];
  var html = _render(TEXT, annos);
  t.equal(html, 'ABC<b>D<i>E</i>F</b>GHI');
  t.end();
});

test$16("Overlapping annos.", function(t) {
  var annos = [new Anno('b', 'b1', 3, 6), new Anno('i', 'i1', 4, 8)];
  var html = _render(TEXT, annos);
  t.equal(html, 'ABC<b>D<i>EF</i></b><i>GH</i>I');
  t.end();
});

test$16("Equal annos.", function(t) {
  var annos = [new Anno('b', 'b1', 3, 6), new Anno('i', 'i1', 3, 6)];
  var html = _render(TEXT, annos);
  t.equal(html, 'ABC<b><i>DEF</i></b>GHI');
  t.end();
});

test$16("Overlapping with fragmentation hint.", function(t) {
  var annos = [
    new Anno('b', 'b1', 3, 6),
    new Anno('a', 'link1', 4, 8, {
      fragmentation: substance.Fragmenter.SHOULD_NOT_SPLIT
    })
  ];
  var html = _render(TEXT, annos);
  t.equal(html, 'ABC<b>D</b><a><b>EF</b>GH</a>I');
  t.end();
});

test$16("Anchors should rendered as early as possible.", function(t) {
  var annos = [
    new Anno('b', 'b1', 3, 6),
    new Anno('a', 'a1', 3, 3, {
      isAnchor: true
    })
  ];
  var html = _render(TEXT, annos);
  t.equal(html, 'ABC<a></a><b>DEF</b>GHI');
  t.end();
});


test$16("Two subsequent inline nodes.", function(t) {
  var annos = [
    new Anno('a', 'inline1', 3, 4, {
      isInline: true
    }),
    new Anno('b', 'inline2', 4, 5, {
      isInline: true
    })
  ];
  var html = _render(TEXT, annos);
  t.equal(html, 'ABC<a>D</a><b>E</b>FGHI');
  t.end();
});

test$16("Collapsed annotation.", function(t) {
  var annos = [
    new Anno('a', 'a1', 0, 0, {
    })
  ];
  var html = _render(TEXT, annos);
  t.equal(html, '<a></a>ABCDEFGHI');
  t.end();
});

test$16("Two collapsed annotations.", function(t) {
  var annos = [
    new Anno('a', 'a1', 0, 0, {
    }),
    new Anno('b', 'b2', 0, 0, {
    })
  ];
  var html = _render(TEXT, annos);
  t.equal(html, '<a></a><b></b>ABCDEFGHI');
  t.end();
});

test$16("Anchors should not fragment other annotations.", function(t) {
  var annos = [
    new Anno('a', 'a1', 3, 6),
    new Anno('b', 'b1', 4, 4, {
      isAnchor: true
    })
  ];
  var html = _render(TEXT, annos);
  t.equal(html, 'ABC<a>D<b></b>EF</a>GHI');
  t.end();
});

class Anno extends substance.Annotation {

  constructor(tagName, id, startOffset, endOffset, opts) {
    super(null, {
      id: id,
      start: { path: [id, 'content'], offset: startOffset},
      end: { path: [id, 'content'], offset: endOffset }
    });

    opts = opts || {};
    this.tagName = tagName;

    this._isAnchor = false;
    this._isInline = false;

    if (opts.hasOwnProperty('fragmentation')) {
      this.fragmentationHint = opts.fragmentation;
    }

    if (opts.hasOwnProperty('isAnchor')) {
      this._isAnchor = opts.isAnchor;
      this.zeroWidth = true;
      this.offset = startOffset;
    }

    if (opts.hasOwnProperty('isInline')) {
      this._isInline = opts.isInline;
    }
  }

  // anchors are special annotations that have zero width
  isAnchor() {
    return this._isAnchor
  }

  // inline nodes are implementated as annotations bound to a single character
  // I.e. the always have a length of 1
  isInline() {
    return this._isInline
  }

}

function _render(text, annotations, opts) {
  opts = opts || {};
  var output = [];
  var fragmenter = new substance.Fragmenter();
  fragmenter.onText = function(context, text) {
    output.push(text);
  };
  fragmenter.onEnter = function(fragment) {
    var node = fragment.node;
    if (opts.withId) {
      output.push('<' + node.tagName + ' id="' + node.id +'">');
    } else {
      output.push('<' + node.tagName + '>');
    }
  };
  fragmenter.onExit = function(fragment) {
    var node = fragment.node;
    output.push('</' + node.tagName + '>');
  };
  fragmenter.start(output, text, annotations);
  return output.join('')
}

const CONTENT = '0123456789';

htmlExporterTests();

if (substance.platform.inBrowser) {
  htmlExporterTests('memory');
}

function htmlExporterTests(memory) {

  const test = substanceTest.module('HTMLExporter' + ( memory ? ' [memory]' : ''), {
    before: function(t) {
      t.elementFactory = memory ? substance.MemoryDOMElement.createDocument('html') : substance.DefaultDOMElement.createDocument('html');
    }
  });

  test("Exporting paragraph", function(t) {
    let { doc, exporter } = setup(t);
    let p1 = doc.create({ type: 'paragraph', id: 'p1', content: CONTENT });
    let el = exporter.convertNode(p1);
    let actual = el.outerHTML;
    let expected = '<p data-id="p1">' + CONTENT + '</p>';
    t.equal(actual, expected, 'Exported HTML should be correct');
    t.end();
  });

  test("Exporting paragraph with strong", function(t) {
    let { doc, exporter } = setup(t);
    let p1 = doc.create({ type: 'paragraph', id: 'p1', content: CONTENT });
    doc.create({
      type: 'strong',
      id: 's1',
      start: {
        path: ['p1', 'content'],
        offset: 4
      },
      end: {
        offset: 7
      }
    });
    let el = exporter.convertNode(p1);
    let actual = el.outerHTML;
    let expected = '<p data-id="p1">0123<strong data-id="s1">456</strong>789</p>';
    t.equal(actual, expected, 'Exported HTML should be correct');
    t.end();
  });

  test("Exporting h1", function(t) {
    let { doc, exporter } = setup(t);
    let h1 = doc.create({ type: 'heading', id: 'h1', level: 1, content: CONTENT });
    let el = exporter.convertNode(h1);
    let actual = el.outerHTML;
    let expected = '<h1 data-id="h1">' + CONTENT + '</h1>';
    t.equal(actual, expected, 'Exported HTML should be correct');
    t.end();
  });

  test("Exporting h2", function(t) {
    let { doc, exporter } = setup(t);
    let h2 = doc.create({ type: 'heading', id: 'h2', level: 2, content: CONTENT });
    let el = exporter.convertNode(h2);
    let actual = el.outerHTML;
    let expected = '<h2 data-id="h2">' + CONTENT + '</h2>';
    t.equal(actual, expected, 'Exported HTML should be correct');
    t.end();
  });

  test("Exporting simple document", function(t) {
    let { doc, exporter } = setup(t, simple);
    let el = exporter.exportDocument(doc);
    let actual = el.html();
    let expected = [
      '<p data-id="p1">' + CONTENT + '</p>',
      '<p data-id="p2">' + CONTENT + '</p>',
      '<p data-id="p3">' + CONTENT + '</p>',
      '<p data-id="p4">' + CONTENT + '</p>'
    ].join('');
    t.equal(actual, expected, 'Exported HTML should be correct');
    t.end();
  });

  test("Exporting a link", function(t) {
    let { doc, exporter } = setup(t);
    let p1 = doc.create({ type: 'paragraph', id: 'p1', content: CONTENT });
    doc.create({
      type: 'link',
      id: 'l1',
      start: {
        path: ['p1', 'content'],
        offset: 4,
      },
      end: {
        offset: 7
      },
      url: 'foo',
      title: 'bar'
    });
    let el = exporter.convertNode(p1);
    let childNodes = el.getChildNodes();
    t.equal(childNodes.length, 3, 'Exported paragraph should have 3 child nodes');
    t.equal(childNodes[0].textContent, "0123", '.. 1. should have correct text');
    t.equal(childNodes[1].textContent, "456", '.. 2. should have correct text');
    t.equal(childNodes[2].textContent, "789", '.. 3. should have correct text');
    let a = childNodes[1];
    t.equal(a.attr('data-id'), 'l1', '.. <a> should have data-id set');
    t.equal(a.attr('href'), 'foo', '.. and correct href attribute');
    t.equal(a.attr('title'), 'bar', '.. and correct title attribute');
    t.end();
  });

  test("Exporting an unordered list", function(t) {
    let { doc, exporter } = setup(t);
    let l1 = _l1$1(doc);
    let el = exporter.convertNode(l1);
    let childNodes = el.getChildNodes();
    t.equal(el.tagName, 'ul', 'Exported element should be a <ul>');
    t.equal(el.attr('data-id'), 'l1', '.. with correct id');
    t.equal(childNodes.length, 2, '.. and two child nodes');
    t.equal(childNodes[0].tagName, "li", '.. a <li>');
    t.equal(childNodes[0].textContent, 'Foo', ".. with content 'Foo'");
    t.equal(childNodes[1].tagName, "li", '.. and a <li>');
    t.equal(childNodes[1].textContent, 'Bar', ".. with content 'Bar'");
    t.end();
  });

  test("Exporting an ordered list", function(t) {
    let { doc, exporter } = setup(t);
    let ol = doc.create({
      type: 'list',
      id: 'ol1',
      ordered: true
    });
    let el = exporter.convertNode(ol);
    t.equal(el.tagName, 'ol', 'Exported element should be a <ol>');
    t.end();
  });

  test("Exporting a nested list", function(t) {
    let { doc, exporter } = setup(t);
    let l = _l2$1(doc);
    let el = exporter.convertNode(l);
    let items = el.findAll('li');
    let nestedList = el.find('ul');
    t.equal(items.length, 4, 'Exported should contain 4 list items');
    t.notNil(nestedList, '.. and a nested list');
    t.equal(nestedList.childNodes.length, 2, '.. which has 2 child nodes');
    t.end();
  });

  function setup(t, fixture) {
    let config = getTestConfig();
    let exporter = config.createExporter('html', {}, { elementFactory: t.elementFactory });
    let doc = createTestArticle(fixture);
    return { exporter, doc }
  }

}

function _li1(doc) {
  doc.create({
    type: 'list-item',
    id: 'li1',
    content: 'Foo',
    level: 1
  });
}

function _li2(doc) {
  doc.create({
    type: 'list-item',
    id: 'li2',
    content: 'Bar',
    level: 1
  });
}

function _l1$1(doc) {
  _li1(doc);
  _li2(doc);
  return doc.create({
    type: 'list',
    id: 'l1',
    items: ['li1', 'li2'],
    ordered: false
  })
}

function _li3(doc) {
  doc.create({
    type: 'list-item',
    id: 'li3',
    content: 'Bla',
    level: 2
  });
}

function _li4(doc) {
  doc.create({
    type: 'list-item',
    id: 'li4',
    content: 'Blupp',
    level: 2
  });
}

function _l2$1(doc) {
  _li1(doc);
  _li2(doc);
  _li3(doc);
  _li4(doc);
  return doc.create({
    type: 'list',
    id: 'l1',
    items: ['li1', 'li3', 'li4', 'li2'],
    ordered: false
  })
}

const CONTENT$1 = '0123456789';


HTMLImporterTests();

if (substance.platform.inBrowser) {
  HTMLImporterTests('memory');
}

function HTMLImporterTests(memory) {

  const test = substanceTest.module('HTMLImporter' + ( memory ? ' [memory]' : ''), {
    before: () => {
      if (memory) substance.platform.inBrowser = false;
    },
    after: () => {
      substance.platform._reset();
    }
  });

  test("Importing paragraph", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<p data-id="p1">' + CONTENT$1 + '</p>';
    let el = substance.DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    checkValues(t, node, {
      id: "p1",
      type: "paragraph",
      content: CONTENT$1
    }, 'paragraph should have been imported correctly');
    t.end();
  });

  test("Importing paragraph with strong", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<p data-id="p1">0123<strong data-id="s1">456</strong>789</p>';
    let el = substance.DefaultDOMElement.parseSnippet(html, 'html');
    let p1 = importer.convertElement(el);
    let doc = p1.getDocument();
    let s1 = doc.get('s1');
    t.equal(p1.content, CONTENT$1, 'paragraph should have correct content');
    t.equal(s1.getText(), '456', 'annotation should provide correct text');
    t.end();
  });

  test("Importing h1", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<h1 data-id="h1">' + CONTENT$1 + '</h1>';
    let el = substance.DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    checkValues(t, node, {
      id: "h1",
      type: "heading",
      level: 1,
      content: CONTENT$1
    }, 'heading should have been imported correctly');
    t.end();
  });

  test("Importing h2", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<h2 data-id="h2">' + CONTENT$1 + '</h2>';
    let el = substance.DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    checkValues(t, node, {
      id: "h2",
      type: "heading",
      level: 2,
      content: CONTENT$1
    }, 'heading should have been imported correctly');
    t.end();
  });

  test("Importing an unordered list", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<ul data-id="l1"><li>Foo</li><li>Bar</li></ul>';
    let el = substance.DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    t.equal(node.type, 'list', 'Imported node should be a list');
    t.equal(node.id, 'l1', 'id should be correct');
    t.equal(node.ordered, false, 'node should unordered');
    t.equal(node.items.length, 2, 'it should have 2 items');
    let li1 = node.getItemAt(0);
    let li2 = node.getItemAt(1);
    t.equal(li1.getText(), 'Foo', 'First item should have correct text');
    t.equal(li2.getText(), 'Bar', 'Second item should have correct text');
    t.end();
  });

  test("Importing an ordered list", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<ol data-id="l1"></ol>';
    let el = substance.DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    t.equal(node.type, 'list', 'Imported node should be a list');
    t.equal(node.id, 'l1', 'id should be correct');
    t.equal(node.ordered, true, 'node should ordered');
    t.end();
  });

  test("Importing a nested list", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<ul data-id="l1"><li>Foo</li><ul><li>Bla</li><li>Blupp</li></ul><li>Bar</li></ul>';
    let el = substance.DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    t.equal(node.items.length, 4, 'Imported node should have 4 items');
    let levels = node.getItems().map(item=>item.level);
    let content = node.getItems().map(item=>item.content);
    t.equal(String(levels), String([1,2,2,1]), 'Node levels should be correct');
    t.equal(String(content), String(['Foo','Bla','Blupp','Bar']), 'Items should have correct content');
    t.end();
  });

  test("Importing a nested list (bad style)", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<ul data-id="l1"><li>Foo<ul><li>Bla</li><li>Blupp</li></ul></li><li>Bar</li></ul>';
    let el = substance.DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    t.equal(node.items.length, 4, 'Imported node should have 4 items');
    let levels = node.getItems().map(item=>item.level);
    let content = node.getItems().map(item=>item.content);
    t.equal(String(levels), String([1,2,2,1]), 'Node levels should be correct');
    t.equal(String(content), String(['Foo','Bla','Blupp','Bar']), 'Items should have correct content');
    t.end();
  });

}

function _setupImporter(options = {}) {
  let config = getTestConfig();
  let importer = config.createImporter('html', {}, options);
  return importer
}

const test$17 = substanceTest.module('InlineNode');

// NOTE: surface ids are a bit ids of Surfaces and IsolatedNodes are not very intuitive
// body/in1 means parent surface of in1 is body -- while in1 is actually on p1.content, which is not a surface on its own

test$17("InlineNodes should be not selected when selection is null", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection(null);
  nodes.forEach(function(node){
    t.ok(node.isNotSelected(), "node '"+node.getId()+"' should not be selected.");
  });
  t.end();
});

test$17("InlineNodes should be not selected when selection is somewhere else", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 5,
    surfaceId: 'body'
  });
  nodes.forEach(function(node){
    t.ok(node.isNotSelected(), "node '"+node.getId()+"' should not be selected.");
  });
  t.end();
});

test$17("InlineNode should be 'selected' with when the inline node is selected", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 2,
    endOffset: 3,
    surfaceId: 'body'
  });
  var expected = {
    'body/in1': 'selected',
    'body/in2': undefined,
  };
  nodes.forEach(function(node){
    var id = node.getId();
    t.equal(node.getMode(), expected[id], "node '" + id + "' should be " + (expected[id] || 'not selected') );
  });
  t.end();
});

test$17("InlineNode should be 'co-selected' when selection is spanning an inline node", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 4,
    surfaceId: 'body'
  });
  var expected = {
    'body/in1': 'co-selected',
    'body/in2': undefined,
  };
  nodes.forEach(function(node){
    var id = node.getId();
    t.equal(node.getMode(), expected[id], "node '" + id + "' should be " + (expected[id] || 'not selected') );
  });
  t.end();
});

test$17("InlineNode should be 'focused' when having the selection", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 4,
    surfaceId: 'body/in1/sn1.title'
  });
  var expected = {
    'body/in1': 'focused',
    'body/in2': undefined,
  };
  nodes.forEach(function(node){
    var id = node.getId();
    t.equal(node.getMode(), expected[id], "node '" + id + "' should be " + (expected[id] || 'not selected') );
  });
  t.end();
});

// Similar to the previous but with another inline node being focused
test$17("InlineNode should be 'focused' when having the selection (II)", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['c_p', 'content'],
    startOffset: 1,
    endOffset: 4,
    surfaceId: 'body/in2/c'
  });
  var expected = {
    'body/in1': undefined,
    'body/in2': 'focused',
  };
  nodes.forEach(function(node){
    var id = node.getId();
    t.equal(node.getMode(), expected[id], "node '" + id + "' should be " + (expected[id] || 'not selected') );
  });
  t.end();
});

test$17("InlineNode should be 'co-focused' when a nested inline node has the selection", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedInlineNode);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['sn2', 'title'],
    startOffset: 2,
    surfaceId: 'body/in1/sn1.title/in2/sn2.title'
  });
  var expected = {
    'body/in1': 'co-focused',
    'body/in1/sn1.title/in2': 'focused',
  };
  nodes.forEach(function(node){
    var id = node.getId();
    t.equal(node.getMode(), expected[id], "node '" + id + "' should be " + (expected[id] || 'not selected') );
  });
  t.end();
});

// fixtures

function paragraphsWithInlineNodes(doc) {
  var tx = new substance.EditingInterface(doc);
  twoParagraphs(tx);
  var sn1 = tx.create({
    type: "structured-node",
    id: "sn1",
    title: "ABCDEFG"
  });
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 2
  });
  tx.insertInlineNode({
    type: 'inline-wrapper',
    id: 'in1',
    wrappedNode: sn1.id
  });
  var c = tx.create({
    type: "container",
    id: "c"
  });
  var c_p = tx.create({
    type: 'paragraph',
    id: "c_p",
    content: "ABCDEFG"
  });
  c.show(c_p);
  tx.setSelection({
    type: 'property',
    path: ['p2', 'content'],
    startOffset: 2
  });
  tx.insertInlineNode({
    type: 'inline-wrapper',
    id: 'in2',
    wrappedNode: c.id
  });
}

// co-focusing an inline node is only possible, if the inline node itself contains
// content with an inline node (or isolated node)
function nestedInlineNode(doc) {
  let tx = new substance.EditingInterface(doc);
  twoParagraphs(tx);
  let sn1 = tx.create({
    type: "structured-node",
    id: "sn1",
    title: "ABCDEFG"
  });
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 2
  });
  tx.insertInlineNode({
    type: 'inline-wrapper',
    id: 'in1',
    wrappedNode: sn1.id
  });
  let sn2 = doc.create({
    type: "structured-node",
    id: "sn2",
    title: "ABCDEFG"
  });
  tx.setSelection({
    type: 'property',
    path: ['sn1', 'title'],
    startOffset: 4
  });
  tx.insertInlineNode({
    type: 'inline-wrapper',
    id: 'in2',
    wrappedNode: sn2.id
  });
}

function nestedContainers(tx) {
  var body = tx.get('body');

  var p1 = tx.create({
    type: 'paragraph',
    id: 'p1',
    content: 'ABCDEF'
  });
  body.show(p1);

  var c1 = tx.create({
    type: 'container',
    id: 'c1',
    nodes: []
  });
  var c1p1 = tx.create({
    type: 'paragraph',
    id: 'c1_p1',
    content: 'Lorem ipsum dolor sit amet, consectetur adipiscing elit.'
  });
  c1.show(c1p1);

  var c2 = tx.create({
    type: 'container',
    id: 'c2',
    nodes: []
  });
  var c2p1 = tx.create({
    type: 'paragraph',
    id: 'c2_p1',
    content: 'Nunc turpis erat, sodales id aliquet eget, rutrum vel libero.'
  });
  c2.show(c2p1);
  c1.show(c2);

  var c1p2 = tx.create({
    type: 'paragraph',
    id: 'c1_p2',
    content: 'Donec dapibus vel leo sit amet auctor. Curabitur at diam urna.'
  });
  c1.show(c1p2);
  body.show(c1);

  var p2 = tx.create({
    type: 'paragraph',
    id: 'p2',
    content: 'ABCDEF'
  });
  body.show(p2);
}

const test$18 = substanceTest.module('IsolatedNode');

test$18("IsolatedNodes should be 'not-selected' when selection is null", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection(null);
  isolatedNodes.forEach(n => _notSelected(t, n));
  t.end();
});

test$18("IsolatedNodes should be 'not-selected' when selection is somewhere else", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    surfaceId: 'body'
  });
  isolatedNodes.forEach(function(isolated){
    t.ok(isolated.isNotSelected(), "isolated node '"+isolated.getId()+"' should not be selected.");
  });
  t.end();
});

test$18("IsolatedNode should be 'selected' with node selection", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'node',
    nodeId: 'c1',
    containerId: 'body',
    surfaceId: 'body'
  });
  let expected = {
    'body/c1': 'selected',
    'body/c1/c2': undefined,
  };
  isolatedNodes.forEach(n => _modeOk(t, n, expected));
  t.end();
});

test$18("IsolatedNode should be 'co-selected' with spanning container selection", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 2,
    surfaceId: 'body'
  });
  let expected = {
    'body/c1': 'co-selected',
    // Note: 'co-selection' does not propagate down
    // it is a state related to the parent container
    'body/c1/c2': undefined,
  };
  isolatedNodes.forEach(n => _modeOk(t, n, expected));
  t.end();
});

test$18("IsolatedNode should be 'focused' when having the selection", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'property',
    path: ['c1_p1', 'content'],
    startOffset: 0,
    surfaceId: 'body/c1'
  });
  let expected = {
    'body/c1': 'focused',
    'body/c1/c2': undefined,
  };
  isolatedNodes.forEach(n => _modeOk(t, n, expected));
  t.end();
});

test$18("IsolatedNode should be 'co-focused' when child is having the selection", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'property',
    path: ['c2_p1', 'content'],
    startOffset: 0,
    surfaceId: 'body/c1/c2'
  });
  let expected = {
    'body/c1': 'co-focused',
    'body/c1/c2': 'focused',
  };
  isolatedNodes.forEach(n => _modeOk(t, n, expected));
  t.end();
});

test$18("Issue #696: IsolatedNode should detect 'co-focused' robustly in presence of surface ids with same prefix", function(t) {
  // as experienced in #696 it happened that co-focused state was infered just
  // by using startsWith on the surface path. This was leading to wrong
  // co-focused states when e.g. two isolated nodes `body/entity` and `body/entity-1`
  // exist. I.e. one surfaceId was a prefix of another one.
  let { editorSession, editor } = setupEditor(t, _twoStructuredNodes);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'property',
    path: ['sn2', 'title'],
    startOffset: 0,
    surfaceId: 'body/sn2/sn2.title'
  });
  let expected = {
    'body/sn': null,
    'body/sn2': 'focused',
  };
  isolatedNodes.forEach(n => _modeOk(t, n, expected));
  t.end();
});

function _notSelected(t, isolated) {
  t.ok(isolated.isNotSelected(), "isolated node '"+isolated.getId()+"' should not be selected.");
}

function _modeOk(t, isolated, expected) {
  let id = isolated.getId();
  t.looseEqual(isolated.getMode(), expected[id], "mode of isolated node '" + id + "' should be correct");
}

function _twoStructuredNodes(doc) {
  let body = doc.get('body');
  body.show(doc.create({
    type: 'structured-node',
    id: 'sn',
    title: 'Foo'
  }));
  body.show(doc.create({
    type: 'structured-node',
    id: 'sn2',
    title: 'Bar'
  }));
}

const test$19 = substanceTest.module('ObjectOperation');

function checkObjectOperationTransform(test, a, b, input, expected) {
  let ops = substance.ObjectOperation.transform(a, b);
  let output = ops[1].apply(a.apply(substance.cloneDeep(input)));
  test.deepEqual(output, expected, `(b' o a)('${JSON.stringify(input)}') == '${JSON.stringify(expected)}' with a=${a.toString()}, b'=${ops[1].toString()}`);
  output = ops[0].apply(b.apply(substance.cloneDeep(input)));
  test.deepEqual(output, expected, `(a' o b)('${JSON.stringify(input)}') == '${JSON.stringify(expected)}' with b=${b.toString()}, a'=${ops[0].toString()}`);
}

test$19("Creating values.", (t) => {
  let path = ["a"];
  let val = { bla: "blupp" };
  let expected = {a: { bla: "blupp" } };
  let op = substance.ObjectOperation.Create(path, val);
  let obj = {};
  op.apply(obj);
  t.deepEqual(obj, expected, 'Should create value.');
  t.end();
});

test$19("Creating nested values.", (t) => {
  let path = ["a", "b"];
  let val = { bla: "blupp" };
  let expected = {a: { b: { bla: "blupp" } } };
  let op = substance.ObjectOperation.Create(path, val);
  let obj = {"a": {}};
  op.apply(obj);
  t.deepEqual(obj, expected, 'Should create nested value.');
  t.end();
});

test$19("Deleting values.", (t) => {
  let path = ["a"];
  let val = "bla";
  let op = substance.ObjectOperation.Delete(path, val);
  let expected = {};
  let obj = {"a": "bla"};
  op.apply(obj);
  t.deepEqual(obj, expected, 'Should delete value.');
  t.end();
});

test$19("Deleting nested values.", (t) => {
  let path = ["a", "b"];
  let val = "bla";
  let op = substance.ObjectOperation.Delete(path, val);
  let expected = { a: {} };
  let obj = { a: { b: "bla"} };
  op.apply(obj);
  t.deepEqual(obj, expected, 'Should delete nested value.');
  t.end();
});

test$19("Updating a text property.", (t) => {
  let obj = {a: "bla"};
  let path = ["a"];
  let op1 = substance.ObjectOperation.Update(path, substance.TextOperation.Delete(2, 'a'));
  let op2 = substance.ObjectOperation.Update(path, substance.TextOperation.Insert(2, 'upp'));
  let expected = {a: "blupp"};
  op1.apply(obj);
  op2.apply(obj);
  t.deepEqual(obj, expected);
  t.end();
});

test$19("Updating an array property.", (t) => {
  let obj = {a: [1,2,3,4,5]};
  let path = ["a"];
  let op1 = substance.ObjectOperation.Update(path, substance.ArrayOperation.Delete(2, 3));
  let op2 = substance.ObjectOperation.Update(path, substance.ArrayOperation.Insert(4, 6));
  let expected = {a: [1,2,4,5,6]};
  op1.apply(obj);
  op2.apply(obj);
  t.deepEqual(obj, expected);
  t.end();
});

test$19("Creating an update operation with invalid diff.", (t) => {
  t.throws(function() {
    substance.ObjectOperation.Update(['test'], "foo");
  }, "Should throw.");
  t.end();
});

test$19("Creating a top-level property using id.", (t) => {
  let obj = {};
  let id = "foo";
  let op = substance.ObjectOperation.Create(id, "bar");
  let expected = { "foo": "bar" };
  op.apply(obj);
  t.deepEqual(obj, expected);
  t.end();
});

test$19("Deleting a top-level property using id.", (t) => {
  let obj = { foo: "bar" };
  let id = "foo";
  let op = substance.ObjectOperation.Delete(id, "bar");
  let expected = {};
  op.apply(obj);
  t.deepEqual(obj, expected);
  t.end();
});

test$19("Apply operation on PathObject.", (t) => {
  let myObj = new substance.PathObject();
  let op = substance.ObjectOperation.Set(['foo', 'bar'], null, 'bla');
  op.apply(myObj);
  t.equal(myObj.get(['foo', 'bar']), 'bla');
  t.end();
});

test$19("Creating operation with invalid data.", (t) => {
  t.throws(function() {
    new substance.ObjectOperation();
  }, "Should throw when data is undefined.");
  t.throws(function() {
    new substance.ObjectOperation({});
  }, "Should throw when type is missing.");
  t.throws(function() {
    new substance.ObjectOperation({ type: 'foo', path: ['test'] });
  }, "Should throw when type is invalid.");
  t.throws(function() {
    new substance.ObjectOperation({
      type: substance.ObjectOperation.CREATE
    });
  }, "Should throw when path is missing.");
  t.throws(function() {
    new substance.ObjectOperation({
      type: substance.ObjectOperation.CREATE,
      path: ["test"],
    });
  }, "Should throw when created value is missing.");
  t.throws(function() {
    new substance.ObjectOperation({
      type: substance.ObjectOperation.UPDATE,
      path: ["test"]
    });
  }, "Should throw when update diff is missing.");
  t.throws(function() {
    new substance.ObjectOperation({
      type: substance.ObjectOperation.UPDATE,
      path: ["test"],
      diff: "foo"
    });
  }, "Should throw when update diff is invalid.");
  // we have relaxed that, so that it is possible to 'delete' a property by setting it to undefined
  // t.throws(function() {
  //   new ObjectOperation({
  //     type: ObjectOperation.SET,
  //     path: ["test"],
  //   })
  // }, "Should throw when value is missing.")
  // t.throws(function() {
  //   new ObjectOperation({
  //     type: ObjectOperation.SET,
  //     path: ["test"],
  //     val: 1
  //   })
  // }, "Should throw when old value is missing.")
  t.end();
});

test$19("Inverse of NOP is NOP", (t) => {
  let op = new substance.ObjectOperation({type: substance.ObjectOperation.NOP});
  let inverse = op.invert();
  t.ok(inverse.isNOP());
  t.end();
});

test$19("Inverse of Create is Delete and vice versa", (t) => {
  let op = substance.ObjectOperation.Create('test', 'foo');
  let inverse = op.invert();
  t.ok(inverse.isDelete());
  t.equal(inverse.getValue(), 'foo');
  inverse = inverse.invert();
  t.ok(inverse.isCreate());
  t.equal(inverse.getValue(), 'foo');
  t.end();
});

test$19("Inverse of Update is Update", (t) => {
  let op = substance.ObjectOperation.Update(['test', 'foo'], substance.ArrayOperation.Insert(1, 'a'));
  let inverse = op.invert();
  t.ok(inverse.isUpdate());
  t.deepEqual(inverse.getPath(), ['test', 'foo']);
  t.ok(inverse.diff instanceof substance.ArrayOperation);
  t.deepEqual(inverse.diff.toJSON(), op.diff.invert().toJSON());
  // same with a text operation as diff
  op = substance.ObjectOperation.Update(['test', 'foo'], substance.TextOperation.Insert(1, 'a'));
  inverse = op.invert();
  t.ok(inverse.diff instanceof substance.TextOperation);
  t.deepEqual(inverse.diff.toJSON(), op.diff.invert().toJSON());
  t.end();
});

test$19("Inverse of Set is Set", (t) => {
  let op = substance.ObjectOperation.Set(['test', 'foo'], 'foo', 'bar');
  let inverse = op.invert();
  t.ok(inverse.isSet());
  t.deepEqual(inverse.getPath(), ['test', 'foo']);
  t.equal(inverse.getValue(), 'foo');
  t.equal(inverse.original, 'bar');
  t.end();
});

test$19("Transformation: everything easy when not the same property", (t) => {
  let path1 = ["a"];
  let path2 = ["b"];
  let val1 = "bla";
  let val2 = "blupp";
  let a = substance.ObjectOperation.Create(path1, val1);
  let b = substance.ObjectOperation.Create(path2, val2);
  let ops = substance.ObjectOperation.transform(a, b);
  t.deepEqual(ops[0].toJSON(), a.toJSON());
  t.deepEqual(ops[1].toJSON(), b.toJSON());
  t.end();
});

test$19("Transformation: everything easy when NOP involved", (t) => {
  let path1 = ["a"];
  let val1 = "bla";
  let a = substance.ObjectOperation.Create(path1, val1);
  let b = new substance.ObjectOperation({type: substance.ObjectOperation.NOP});
  let ops = substance.ObjectOperation.transform(a, b);
  t.deepEqual(ops[0].toJSON(), a.toJSON());
  t.deepEqual(ops[1].toJSON(), b.toJSON());
  t.end();
});

test$19("Transformation: creating the same value (unresolvable conflict)", (t) => {
  let path = ["a"];
  let val1 = "bla";
  let val2 = "blupp";
  let a = substance.ObjectOperation.Create(path, val1);
  let b = substance.ObjectOperation.Create(path, val2);
  t.throws(function() {
    substance.ObjectOperation.transform(a, b);
  });
  t.end();
});

test$19("Transformation: creating and updating the same value (unresolvable conflict)", (t) => {
  let path = ["a"];
  let val1 = "bla";
  let a = substance.ObjectOperation.Create(path, val1);
  let b = substance.ObjectOperation.Update(path, substance.TextOperation.Insert(1, 'b'));
  t.throws(function() {
    substance.ObjectOperation.transform(a, b);
  });
  t.end();
});

test$19("Transformation: creating and setting the same value (unresolvable conflict)", (t) => {
  let path = ["a"];
  let val1 = "bla";
  let val2 = "blupp";
  let a = substance.ObjectOperation.Create(path, val1);
  let b = substance.ObjectOperation.Set(path, val1, val2);
  t.throws(function() {
    substance.ObjectOperation.transform(a, b);
  });
  t.end();
});

test$19("Transformation: creating and deleting the same value (unresolvable conflict)", (t) => {
  let path = ["a"];
  let val1 = "bla";
  let a = substance.ObjectOperation.Create(path, val1);
  let b = substance.ObjectOperation.Delete(path, val1);
  t.throws(function() {
    substance.ObjectOperation.transform(a, b);
  });
  t.end();
});

test$19("Transformation: deleting the same value", (t) => {
  let path = ["a"];
  let val = "bla";
  let input = {"a": val};
  let expected = {};
  let a = substance.ObjectOperation.Delete(path, val);
  let b = substance.ObjectOperation.Delete(path, val);
  checkObjectOperationTransform(t, a, b, input, expected);
  checkObjectOperationTransform(t, b, a, input, expected);
  t.end();
});

test$19("Transformation: deleting and updating the same value", (t) => {
  let path = ["a"];
  let a = substance.ObjectOperation.Delete(path, "bla");
  let b = substance.ObjectOperation.Update(path, substance.TextOperation.Insert(3, "pp"));
  let input = {a : "bla"};
  let expected1 = {a: "blapp"};
  let expected2 = {};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  // same with an array operation
  a = substance.ObjectOperation.Delete(path, [1,2,3]);
  b = substance.ObjectOperation.Update(path, substance.ArrayOperation.Insert(3, 4));
  input = {a : [1,2,3]};
  expected1 = {a: [1,2,3,4]};
  expected2 = {};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  t.end();
});

test$19("Transformation: deleting and setting the same value", (t) => {
  let path = ["a"];
  let a = substance.ObjectOperation.Delete(path, "bla");
  let b = substance.ObjectOperation.Set(path, "bla", "blupp");
  let input = {a : "bla"};
  let expected1 = {a: "blupp"};
  let expected2 = {};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  t.end();
});

test$19("Transformation: updating the same value", (t) => {
  let path = ["a"];
  let a = substance.ObjectOperation.Update(path, substance.TextOperation.Insert(3, "pp"));
  let b = substance.ObjectOperation.Update(path, substance.TextOperation.Insert(3, "ff"));
  let input = {a : "bla"};
  let expected1 = {a: "blappff"};
  let expected2 = {a: "blaffpp"};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  // same with array updates
  a = substance.ObjectOperation.Update(path, substance.ArrayOperation.Insert(2, 3));
  b = substance.ObjectOperation.Update(path, substance.ArrayOperation.Insert(2, 4));
  input = {a : [1,2,5]};
  expected1 = {a: [1,2,3,4,5]};
  expected2 = {a: [1,2,4,3,5]};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  t.end();
});

test$19("Transformation: updating and setting the same value (unresolvable conflict)", (t) => {
  let path = ["a"];
  let a = substance.ObjectOperation.Update(path, substance.TextOperation.Insert(3, "ff"));
  let b = substance.ObjectOperation.Set(path, "bla", "blupp");
  t.throws(function() {
    substance.ObjectOperation.transform(a, b);
  });
  t.end();
});

test$19("Transformation: setting the same value", (t) => {
  let path = ["a"];
  let a = substance.ObjectOperation.Set(path, "bla", "blapp");
  let b = substance.ObjectOperation.Set(path, "bla", "blupp");
  let input = {a : "bla"};
  let expected1 = {a: "blupp"};
  let expected2 = {a: "blapp"};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  t.end();
});

test$19("ObjectOperation with the same path are conflicts.", (t) => {
  let a = substance.ObjectOperation.Set(['bla', 'blupp'], null, 'foo');
  let b = substance.ObjectOperation.Set(['bla', 'blupp'], null, 'bar');
  t.ok(a.hasConflict(b) && b.hasConflict(a));
  t.end();
});

test$19("NOPs have never conflicts.", (t) => {
  let a = substance.ObjectOperation.Set(['bla', 'blupp'], null, 'foo');
  let b = new substance.ObjectOperation({type: substance.ObjectOperation.NOP});
  t.ok(!a.hasConflict(b) && !b.hasConflict(a));
  t.end();
});

test$19("With option 'no-conflict' conflicting operations can not be transformed.", (t) => {
  let a = substance.ObjectOperation.Create('bla', 'blupp');
  let b = substance.ObjectOperation.Create('bla', 'blupp');
  t.throws(function() {
    substance.ObjectOperation.transform(a, b, { "no-conflict": true });
  }, 'Transforming conflicting ops should throw when option "no-conflict" is enabled.');
  t.end();
});

test$19("JSON deserialisation", (t) => {
  let path = ["test", "foo"];

  let op = substance.ObjectOperation.fromJSON({
    type: substance.ObjectOperation.SET,
    path: path,
    val: "bla",
    original: "blupp"
  });
  t.equal(op.getType(), substance.ObjectOperation.SET);
  t.equal(op.getValue(), "bla");

  op = substance.ObjectOperation.fromJSON({
    type: substance.ObjectOperation.UPDATE,
    path: path,
    diff: substance.ArrayOperation.Insert(1, 'a'),
    propertyType: 'array'
  });
  t.ok(op.diff instanceof substance.ArrayOperation);
  t.deepEqual(op.getPath(), path);
  op = substance.ObjectOperation.fromJSON({
    type: substance.ObjectOperation.UPDATE,
    path: path,
    diff: substance.TextOperation.Insert(1, 'a'),
    propertyType: 'string'
  });
  t.ok(op.diff instanceof substance.TextOperation);
  t.deepEqual(op.getPath(), path);

  t.throws(function() {
    substance.ObjectOperation.fromJSON({
      type: substance.ObjectOperation.UPDATE,
      path: path,
      diff: "bla",
      propertyType: 'foo'
    });
  }, "Should throw for unknown update diff type.");
  t.end();
});

test$19("JSON serialisation", (t) => {
  let data = substance.ObjectOperation.Create('test', "bla").toJSON();
  t.equal(data.type, substance.ObjectOperation.CREATE);
  t.equal(data.val, "bla");
  data = substance.ObjectOperation.Delete('test', "bla").toJSON();
  t.equal(data.type, substance.ObjectOperation.DELETE);
  t.equal(data.val, "bla");
  data = substance.ObjectOperation.Update(['test', 'foo'], substance.ArrayOperation.Insert(1, 'a')).toJSON();
  t.equal(data.type, substance.ObjectOperation.UPDATE);
  t.equal(data.propertyType, 'array');
  data = substance.ObjectOperation.Update(['test', 'foo'], substance.TextOperation.Insert(1, 'a')).toJSON();
  t.equal(data.propertyType, 'string');
  data = substance.ObjectOperation.Set(['test', 'foo'], 'foo', 'bar').toJSON();
  t.equal(data.type, substance.ObjectOperation.SET);
  t.equal(data.val, 'bar');
  t.equal(data.original, 'foo');
  t.end();
});

const test$20 = substanceTest.module('OperationSerializer');

test$20('Serializing create', function(t) {
  var p = { id: 'p1', type: 'paragraph', content: 'foo' };
  var op = substance.ObjectOperation.Create(['p1'], p);
  var io = new substance.OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['c', p.id, p]);
  t.end();
});

test$20('Deserializing create', function(t) {
  var p = { id: 'p1', type: 'paragraph', content: 'foo' };
  var io = new substance.OperationSerializer();
  var data = ['c', 'p1', p];
  var op = io.deserialize(data);
  t.ok(op.type, 'create');
  t.deepEqual(op.getValue(), p);
  t.end();
});

test$20('Serializing delete', function(t) {
  var p = { id: 'p1', type: 'paragraph', content: 'foo' };
  var op = substance.ObjectOperation.Delete(['p1'], p);
  var io = new substance.OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['d', p.id, p]);
  t.end();
});

test$20('Deserializing delete', function(t) {
  var p = { id: 'p1', type: 'paragraph', content: 'foo' };
  var io = new substance.OperationSerializer();
  var data = ['d', 'p1', p];
  var op = io.deserialize(data);
  t.ok(op.type, 'delete');
  t.deepEqual(op.getValue(), p);
  t.end();
});

test$20('Serializing set', function(t) {
  var op = substance.ObjectOperation.Set(['p1', 'content'], 'foo', 'bar');
  var io = new substance.OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['s', 'p1.content', 'bar', 'foo']);
  t.end();
});

test$20('Serializing set with null', function(t) {
  var op = substance.ObjectOperation.Set(['p1', 'content'], null, null);
  var io = new substance.OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['s', 'p1.content', null, null]);
  t.end();
});

test$20('Serializing set with undefined', function(t) {
  // ATTENTION: undefined is not a valid JSON value, so it gets replaced
  // by null
  var op = substance.ObjectOperation.Set(['p1', 'content'], undefined, undefined);
  var io = new substance.OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['s', 'p1.content', undefined, undefined]);
  t.end();
});

test$20('Deserializing set', function(t) {
  var io = new substance.OperationSerializer();
  var data = ['s', 'p1.content', 'bar', 'foo'];
  var op = io.deserialize(data);
  t.equal(op.type, 'set');
  t.deepEqual(op.getPath(), ['p1', 'content']);
  t.deepEqual(op.getValue(), 'bar');
  t.deepEqual(op.getOldValue(), 'foo');
  t.end();
});


test$20('Serializing text insert', function(t) {
  var op = substance.ObjectOperation.Update(['p1', 'content'], substance.TextOperation.Insert(3, 'foo'));
  var io = new substance.OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['u', 'p1.content', 't+', 3, 'foo']);
  t.end();
});

test$20('Deserializing text insert', function(t) {
  var io = new substance.OperationSerializer();
  var data = ['u', 'p1.content', 't+', 3, 'foo'];
  var op = io.deserialize(data);
  t.ok(op.type, 'update');
  t.deepEqual(op.getPath(), ['p1', 'content']);
  var valueOp = op.getValueOp();
  t.ok(valueOp instanceof substance.TextOperation, 'Value operation should be a TextOperation');
  t.equal(valueOp.type, 'insert');
  t.equal(valueOp.pos, 3);
  t.equal(valueOp.str, 'foo');
  t.end();
});

test$20('Serializing text delete', function(t) {
  var op = substance.ObjectOperation.Update(['p1', 'content'], substance.TextOperation.Delete(3, 'foo'));
  var io = new substance.OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['u', 'p1.content', 't-', 3, 'foo']);
  t.end();
});

test$20('Deserializing text delete', function(t) {
  var io = new substance.OperationSerializer();
  var data = ['u', 'p1.content', 't-', 3, 'foo'];
  var op = io.deserialize(data);
  t.ok(op.type, 'update');
  t.deepEqual(op.getPath(), ['p1', 'content']);
  var valueOp = op.getValueOp();
  t.ok(valueOp instanceof substance.TextOperation, 'Value operation should be a TextOperation');
  t.equal(valueOp.type, 'delete');
  t.equal(valueOp.pos, 3);
  t.equal(valueOp.str, 'foo');
  t.end();
});

test$20('Serializing an array insert', function(t) {
  var op = substance.ObjectOperation.Update(['test', 'numbers'], substance.ArrayOperation.Insert(3, 1234));
  var io = new substance.OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['u', 'test.numbers', 'a+', 3, 1234]);
  t.end();
});

test$20('Deserializing an array insert', function(t) {
  var io = new substance.OperationSerializer();
  var data = ['u', 'test.numbers', 'a+', 3, 1234];
  var op = io.deserialize(data);
  t.ok(op.type, 'update');
  t.deepEqual(op.getPath(), ['test', 'numbers']);
  var valueOp = op.getValueOp();
  t.ok(valueOp instanceof substance.ArrayOperation, 'Value operation should be a ArrayOperation');
  t.equal(valueOp.type, 'insert');
  t.equal(valueOp.pos, 3);
  t.equal(valueOp.val, 1234);
  t.end();
});

test$20('Serializing an array delete', function(t) {
  var op = substance.ObjectOperation.Update(['test', 'numbers'], substance.ArrayOperation.Delete(3, 1234));
  var io = new substance.OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['u', 'test.numbers', 'a-', 3, 1234]);
  t.end();
});

test$20('Deserializing an array delete', function(t) {
  var io = new substance.OperationSerializer();
  var data = ['u', 'test.numbers', 'a-', 3, 1234];
  var op = io.deserialize(data);
  t.ok(op.type, 'update');
  t.deepEqual(op.getPath(), ['test', 'numbers']);
  var valueOp = op.getValueOp();
  t.ok(valueOp instanceof substance.ArrayOperation, 'Value operation should be a ArrayOperation');
  t.equal(valueOp.type, 'delete');
  t.equal(valueOp.pos, 3);
  t.equal(valueOp.val, 1234);
  t.end();
});

const test$21 = substanceTest.module('packages');

test$21('import all packages', (t) => {
  [
    substance.BasePackage,
    substance.BlockquotePackage,
    substance.BodyScrollPanePackage,
    substance.ButtonPackage,
    substance.CodePackage,
    substance.CodeblockPackage,
    substance.ContextMenuPackage,
    substance.DropzonesPackage,
    substance.EmphasisPackage,
    substance.FilePackage,
    substance.GridPackage,
    substance.GutterPackage,
    substance.HeadingPackage,
    substance.ImagePackage,
    substance.InlineWrapperPackage,
    substance.InputPackage,
    substance.LayoutPackage,
    substance.LinkPackage,
    substance.ListPackage,
    substance.ModalPackage,
    substance.OverlayPackage,
    substance.ParagraphPackage,
    substance.PersistencePackage,
    substance.ProseEditorPackage,
    substance.QuoteMarksPackage,
    substance.ScrollPanePackage,
    substance.ScrollbarPackage,
    substance.SpellCheckPackage,
    substance.SplitPanePackage,
    substance.StrongPackage,
    substance.SubscriptPackage,
    substance.SuperscriptPackage,
    substance.SwitchTextTypePackage,
    substance.TabbedPanePackage,
    substance.TablePackage,
    substance.TextInputPackage,
    substance.ToolbarPackage
  ].forEach((p) => {
    t.doesNotThrow(()=>{
      let configurator = new substance.Configurator();
      configurator.import(p);
    });
  });
  t.end();
});

const test$22 = substanceTest.module('paste');

test$22("Pasting plain text", function(t) {
  let { tx } = _fixture$3(simple);
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3
  });
  tx.paste('XXX');
  let p1 = tx.get('p1');
  t.equal(p1.content, '012XXX3456789');
  t.end();
});

test$22("Pasting a single paragraph", function(t) {
  let { tx } = _fixture$3(simple);
  let snippet = tx.createSnippet();
  let container = snippet.getContainer();
  let p = snippet.create({
    type: 'paragraph',
    id: substance.Document.TEXT_SNIPPET_ID,
    content: 'AABBCC'
  });
  container.show(p.id);
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  tx.paste(snippet);
  let p1 = tx.get('p1');
  t.equal(p1.content, '012AABBCC3456789', 'Plain text should be inserted.');
  t.end();
});

test$22("Pasting annotated text", function(t) {
  let { tx } = _fixture$3(simple);
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  let snippet = tx.createSnippet();
  let container = snippet.getContainer();
  let p = snippet.create({
    type: 'paragraph',
    id: substance.Document.TEXT_SNIPPET_ID,
    content: 'AABBCC'
  });
  container.show(p.id);
  snippet.create({
    type: 'strong',
    id: 's1',
    start: {
      path: [p.id, 'content'],
      offset: 2,
    },
    end: {
      offset: 4
    }
  });
  tx.paste(snippet);
  let p1 = tx.get('p1');
  t.equal(p1.content, '012AABBCC3456789', 'Plain text should be inserted.');
  let s1 = tx.get('s1');
  t.deepEqual(s1.path, [p1.id, 'content'], 'Annotation is bound to the correct path.');
  t.deepEqual([s1.start.offset, s1.end.offset], [5, 7], 'Annotation has correct range.');
  t.end();
});

test$22("Pasting two paragraphs", function(t) {
  let { tx } = _fixture$3(simple);
  let snippet = tx.createSnippet();
  let container = snippet.getContainer();
  let test1 = snippet.create({
    type: 'paragraph',
    id: 'test1',
    content: 'AA'
  });
  container.show(test1.id);
  let test2 = snippet.create({
    type: 'paragraph',
    id: 'test2',
    content: 'BB'
  });
  container.show(test2.id);
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  tx.paste(snippet);
  let body = tx.get('body');
  let p1 = tx.get('p1');
  t.equal(p1.content, '012AA', 'First part should be inserted into first paragraph.');
  t.equal(body.nodes[1], test2.id, 'Second part should go into a single paragraph.');
  t.equal(tx.get(body.nodes[2]).content, '3456789', 'Remaining part of first paragraph should be in a new paragraph.');
  t.equal(body.nodes[3], 'p2', 'After that should follow p2.');
  t.end();
});

function _fixture$3(seed) {
  let doc = createTestArticle(seed);
  let tx = new substance.EditingInterface(doc);
  return { doc: doc, tx: tx }
}

const test$23 = substanceTest.module('PathEventProxy');

function docWithTestNodes(tx) {
  headersAndParagraphs(tx);
  tx.create({
    type: "test-node",
    id: "test",
    arrayVal: [1,2,3]
  });
}

test$23("Updating a property", function(t) {
  var doc = createTestArticle(docWithTestNodes);
  var callCount = 0;
  doc.getEventProxy('path').on(['test', 'arrayVal'], function() {
    callCount++;
  });
  doc.update(['test', 'arrayVal'], { type: 'insert', pos: 1, value: '1000' } );
  t.equal(callCount, 1, "Event proxy listener should have been called.");
  t.end();
});

test$23("Setting a property", function(t) {
  var doc = createTestArticle(docWithTestNodes);
  var callCount = 0;
  doc.getEventProxy('path').on(['test', 'arrayVal'], function() {
    callCount++;
  });
  doc.set(['test', 'arrayVal'], [1,1,1]);
  t.equal(callCount, 1, "Event proxy listener should have been called.");
  t.end();
});

test$23("Setting a property and deleting the node afterwards", function(t) {
  var doc = createTestArticle(docWithTestNodes);
  var editorSession = new substance.EditorSession(doc, { configurator: new substance.Configurator() });
  var callCount = 0;
  doc.getEventProxy('path').on(['test', 'arrayVal'], function() {
    callCount++;
  });
  editorSession.transaction(function(tx) {
    tx.set(['test', 'arrayVal'], [1,1,1]);
    tx.delete('test');
  });
  t.equal(callCount, 0, "Event proxy listener doesn't get called when node is deleted.");
  t.end();
});

/* eslint-disable no-invalid-this, indent */
const Simple$1 = TestComponent.Simple;
const internal = substance.RenderingEngine._internal;

// runs RenderingEngine in regular mode
RenderingEngineTests();

// runs RenderingEngine in debug mode
RenderingEngineTests('debug');


function RenderingEngineTests(debug) {

  const test = substanceTest.module('RenderingEngine' + (debug ? ' [debug-mode]' : ''), {
    before: function() {
      substance.substanceGlobals.DEBUG_RENDERING = Boolean(debug);
    }
  });

  // NOTE: this is a set of white-box tests for the internal implementation
  // of ui/RenderingEngine.

  // __isMapped__ is an internal state variable that is used
  // by RenderingEnging to know, that a component can be reused. I.e.,
  // the VirtualComponent has been mapped successfully to a Component

  test('Components without refs are not mapped', function(t) {
    var comp = TestComponent.create(function($$) {
      return $$('div').append($$(Simple$1))
    });
    var vc = _capture(comp);
    t.notOk(vc._isMapped(vc._content.children[0]), 'child element should not be mapped.');
    t.end();
  });

  test('A component with ref is mapped', function(t) {
    var comp = TestComponent.create(function($$) {
      return $$('div').append($$(Simple$1).ref('foo'))
    });
    var vc = _capture(comp);
    t.ok(vc._isMapped(vc._content), 'root element should be mapped');
    t.ok(vc._isMapped(vc._getRef('foo')), "'foo' should be mapped");
    t.end();
  });

  // TODO: there are more case related nesting of components
  // and injecting components via props

  // __isRelocated__ is an internal state variable that is used
  // when a reused component has been rendered with a different parent

  test('Detecting relocation when injecting a new parent element', function(t) {
    function _render($$) {
      var el = $$('div');
      var parent = el;
      if (this.props.extraLayer) {
        var middle = $$('div');
        el.append(middle);
        parent = middle;
      }
      parent.append($$(Simple$1).ref('foo'));
      return el
    }
    var comp = TestComponent.create(_render);
    var vc = _capture(comp);
    t.notOk(vc._isRelocated(vc._getRef('foo')), "'foo' is not relocated the first time.");

    comp = TestComponent.create(_render, { extraLayer: false });
    _setProps(comp, { extraLayer: true });
    vc = _capture(comp);
    t.ok(vc._isRelocated(vc._getRef('foo')), "'foo' is relocated the second time.");

    comp = TestComponent.create(_render, { extraLayer: true });
    _setProps(comp, { extraLayer: false });
    vc = _capture(comp);
    t.ok(vc._isRelocated(vc._getRef('foo')), "'foo' is relocated the third time.");

    t.end();
  });

  test('Detecting relocation when injecting components (TextProperty use-case)', function(t) {
    /*
      This simulates a situation found often when rendering a TextProperty.
      Say a text property contains an inline node.
      Regularly, the content looks roughly like this:
      ```
        AAAA<div class="inline-node">BBBB</div>CCCC
      ```
      When selected there is a wrapper around this
      ```
        AAAA<div class='selection'><div>BBBB</div></div>CCCC
      ```
      Adding and removing this selection element leads to a situation
      where the inline node needs to be attached to varying parent
      elements.
    */
    function _render($$) {
      var el = $$('div');
      var parent = el;
      el.append('AAAA');
      if (this.props.extraLayer) {
        var middle = $$(Simple$1).ref('selection');
        el.append(middle);
        parent = middle;
      }
      parent.append($$(Simple$1).ref('foo').append('BBBB'));
      el.append('CCCC');
      return el
    }
    var comp = TestComponent.create(_render);
    var vc = _capture(comp);
    t.notOk(vc._isRelocated(vc._getRef('foo')), "'foo' is not relocated the first time.");

    comp = TestComponent.create(_render, { extraLayer: false });
    _setProps(comp, { extraLayer: true });
    vc = _capture(comp);
    t.ok(vc._isRelocated(vc._getRef('foo')), "'foo' is relocated the second time.");

    comp = TestComponent.create(_render, { extraLayer: true });
    _setProps(comp, { extraLayer: false });
    vc = _capture(comp);
    t.ok(vc._isRelocated(vc._getRef('foo')), "'foo' is relocated the third time.");

    t.end();
  });

  // NOTE: Dunno if this is a good test. Obviously it depends too much on the performance of the device
  test('(Performance) Rendering a large number of components with ref', (t) => {
    const N = 100;
    const M = 52;
    const data = [];
    for (let i = 0; i < N; i++) {
      let row = [];
      for (let j = 0; j < M; j++) {
        row.push(`${i}-${j}`);
      }
      data.push(row);
    }
    class Matrix extends substance.Component {
      render($$) {
        let el = $$('table').css({
          overflow: 'auto',
          height: '100px'
        });
        const data = this.props.data;
        for (let i = 0; i < data.length; i++) {
          let row = data[i];
          let rowEl = $$('tr');
          for (let j = 0; j < row.length; j++) {
            rowEl.append(
              $$('td')
                .ref(row[j])
                .text(row[j]));
          }
          el.append(rowEl);
        }
        return el
      }
    }
    let t0 = Date.now();
    let comp = Matrix.mount({data}, getMountPoint(t));
    let t1 = Date.now() - t0;
    t0 = Date.now();
    comp.setProps({data});
    let t2 = Date.now() - t0;
    t.ok(t1 < 1000, 'First rendering should be finished in reasonable time');
    t.ok(t2 < 1000, 'Second rendering should be finished in reasonable time');
    t.end();
  });
}

function _capture(comp) {
  var vc = internal._wrap(comp);
  var state = new substance.RenderingEngine.State();
  internal._capture(state, vc, 'force');
  vc._state = state;
  vc._isMapped = function(o) { return state.isMapped(o); };
  vc._isRelocated = function(o) { return state.isRelocated(o); };
  vc._getRef = function(ref) { return _getRef(vc, ref); };
  return vc
}

function _getRef(vc, ref) {
  return vc._content._context.refs[ref] || {}
}

function _setProps(comp, props) {
  comp.props = props;
  Object.freeze(props);
}

/* eslint-disable no-invalid-this */
const test$24 = substanceTest.module('Surface');

// This test was added to cover issue #82
test$24.UI("Set the selection after creating annotation.", function(t) {
  window.getSelection().removeAllRanges();
  let { editorSession, surface } = setupEditor(t, _p1$1);
  let sel = editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    endOffset: 5,
    containerId: 'body',
    surfaceId: 'body'
  });
  editorSession.transaction(function(tx) {
    tx.annotate({ type: "strong" });
  });
  let wsel = window.getSelection();
  let newSel = surface.domSelection.getSelection();
  t.equal(wsel.rangeCount, 1, "There should be a DOM selection.");
  t.ok(newSel.equals(sel), "New selection should be equal to initial selection.");
  t.end();
});

test$24.UI("Render a reverse selection.", function(t) {
  window.getSelection().removeAllRanges();
  let { editorSession } = setupEditor(t, _p1$1, _p2$1);
  editorSession.setSelection({
    type: 'container',
    startPath:['p1', 'content'],
    startOffset: 3,
    endPath: ['p2', 'content'],
    endOffset: 2,
    reverse: true,
    containerId: 'body',
    surfaceId: 'body'
  });
  var wsel = substance.BrowserDOMElement.getWindowSelection();
  t.ok(substance.BrowserDOMElement.isReverse(wsel.anchorNode, wsel.anchorOffset, wsel.focusNode, wsel.focusOffset));
  t.end();
});

const P1_TEXT$1 = 'abcdef';

function _p1$1(doc, body) {
  doc.create({
    type: 'paragraph',
    id: 'p1',
    content: P1_TEXT$1
  });
  body.show('p1');
}

const P2_TEXT$1 = 'ghijk';

function _p2$1(doc, body) {
  doc.create({
    type: 'paragraph',
    id: 'p2',
    content: P2_TEXT$1
  });
  body.show('p2');
}

const test$25 = substanceTest.module('TextOperation');

function checkTextTransform(test, a, b, input, expected) {
  let ops = substance.TextOperation.transform(a, b);
  let output = ops[1].apply(a.apply(input));
  test.deepEqual(output, expected, `(b' o a)('${input}') == '${expected}' with a=${a.toString()}, b'=${ops[1].toString()}`);
  output = ops[0].apply(b.apply(input));
  test.deepEqual(output, expected, `(a' o b)('${input}') == '${expected}' with b=${b.toString()}, a'=${ops[0].toString()}`);
}

test$25("Insert string", (t) => {
  let input = "Lorem ipsum";
  let expected = "Lorem bla ipsum";
  let a = substance.TextOperation.Insert(6, "bla ");
  t.equal(expected, a.apply(input));
  t.end();
});

test$25("Insert at last position", (t) => {
  let input = "Lorem ipsum";
  let expected = "Lorem ipsum.";
  let a = substance.TextOperation.Insert(11, ".");
  t.equal(expected, a.apply(input));
  t.end();
});

test$25("Invalid arguments", (t) => {
  t.throws(function() {
    new substance.TextOperation({});
  }, "Should throw for incomplete data.");
  t.throws(function() {
    new substance.TextOperation({type: "foo", pos: 0, str: ""});
  }, "Should throw for invalid type.");
  t.throws(function() {
    substance.TextOperation.Insert(-1, "");
  }, "Should throw for invalid position.");
  t.throws(function() {
    substance.TextOperation.Insert(-1, null);
  }, "Should throw for invalid string.");
  t.end();
});

test$25("TextOperation has length", (t) => {
  let op = substance.TextOperation.Delete(1, 'bla');
  t.equal(op.getLength(), 3, 'Length of ' + op.toString() + ' should be 3.');
  op = substance.TextOperation.Insert(1, 'blupp');
  t.equal(op.getLength(), 5, 'Length of ' + op.toString() + ' should be 5.');
  t.end();
});

test$25("JSON serialisation", (t) => {
  let op = substance.TextOperation.Delete(1, 'bla');
  let expected = {
    type: substance.TextOperation.DELETE,
    pos: 1,
    str: 'bla'
  };
  t.deepEqual(op.toJSON(), expected, 'Serialized operation should ok.');
  t.end();
});

test$25("JSON deserialisation", (t) => {
  let data = {
    type: substance.TextOperation.INSERT,
    pos: 1,
    str: "bla"
  };
  let op = substance.TextOperation.fromJSON(data);
  t.ok(op.isInsert(), 'Deserialized operation should be an insert operation.');
  t.ok(op.pos === 1, 'Deserialized operation should have offset==1.');
  t.ok(op.str === "bla", 'Deserialized operation should have string=="bla".');
  t.end();
});

test$25("Empty TextOperations are NOPs", (t) => {
  let op = substance.TextOperation.Insert(0, "");
  t.ok(op.isNOP(), 'Empty operations should be NOPs.');
  t.end();
});

test$25("Can't apply on a too short string", (t) => {
  let op = substance.TextOperation.Insert(6, "bla");
  t.throws(function() {
    op.apply('bla');
  }, "Should throw if string is too short.");
  op = substance.TextOperation.Delete(2, "bla");
  t.throws(function() {
    op.apply('bla');
  }, "Should throw if string is too short.");
  t.end();
});

test$25("Can be applied on custom String implementation", (t) => {
  let CustomString = function(str) {
    this.arr = str.split('');
    this.splice = function(pos, remove, insert) {
      this.arr.splice(pos, remove);
      if (insert) {
        this.arr = this.arr.slice(0, pos).concat(insert.split('')).concat(this.arr.slice(pos));
      }
    };
    this.toString = function() {
      return this.arr.join('')
    };
  };
  Object.defineProperty(CustomString.prototype, 'length', {
    get: function() {
      return this.arr.length
    }
  });
  let str = new CustomString('Lorem ipsum.');
  let op = substance.TextOperation.Insert(6, "bla ");
  op.apply(str);
  t.equal(str.toString(), 'Lorem bla ipsum.', 'Insert operation should work on custom string.');
  str = new CustomString('Lorem bla ipsum.');
  op = substance.TextOperation.Delete(6, "bla ");
  op.apply(str);
  t.equal(str.toString(), 'Lorem ipsum.', 'Delete operation should work on custom string.');
  t.end();
});

test$25("Inversion of Insert = Delete", (t) => {
  let op = substance.TextOperation.Insert(6, "bla");
  let inverse = op.invert();
  t.ok(inverse.isDelete(), 'Inverted operation should be a delete op.');
  t.equal(inverse.pos, op.pos, 'Inverted operation should have the same offset.');
  t.equal(inverse.str, op.str, 'Inverted operation should have the same string data.');
  t.end();
});

test$25("Inversion of Delete = Insert", (t) => {
  let op = substance.TextOperation.Delete(6, "bla");
  let inverse = op.invert();
  t.ok(inverse.isInsert(), 'Inverted operation should be a insert op.');
  t.equal(inverse.pos, op.pos, 'Inverted operation should have the same offset.');
  t.equal(inverse.str, op.str, 'Inverted operation should have the same string data.');
  t.end();
});

test$25("Transformation: a=Insert, b=Insert, a before b", (t) => {
  let input = "Lorem ipsum";
  let expected = "Lorem bla ipsum blupp";
  let a = substance.TextOperation.Insert(6, "bla ");
  let b = substance.TextOperation.Insert(11, " blupp");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Insert, b=Insert, same position", (t) => {
  let input = "Lorem ipsum";
  let a = substance.TextOperation.Insert(6, "bla ");
  let b = substance.TextOperation.Insert(6, "blupp ");
  let expected = "Lorem bla blupp ipsum";
  // applying b first gives a different result
  let expected_2 = "Lorem blupp bla ipsum";
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected_2);
  t.end();
});

test$25("Transformation: a=Delete, b=Delete, a before b", (t) => {
  let input = "Lorem ipsum dolor sit amet";
  let expected = "Lorem dolor amet";
  let a = substance.TextOperation.Delete(6, "ipsum ");
  let b = substance.TextOperation.Delete(18, "sit ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Delete, b=Delete, overlapping", (t) => {
  let input = "Lorem ipsum dolor sit amet";
  let expected = "Lorem amet";
  let a = substance.TextOperation.Delete(6, "ipsum dolor sit ");
  let b = substance.TextOperation.Delete(12, "dolor ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Delete, b=Delete, same position", (t) => {
  let input = "Lorem ipsum dolor sit amet";
  let expected = "Lorem amet";
  let a = substance.TextOperation.Delete(6, "ipsum dolor ");
  let b = substance.TextOperation.Delete(6, "ipsum dolor sit ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Insert, b=Delete", (t) => {
  let input = "Lorem dolor sit amet";
  let expected = "Lorem ipsum dolor amet";
  let a = substance.TextOperation.Insert(6, "ipsum ");
  let b = substance.TextOperation.Delete(12, "sit ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Insert, b=Delete, a after b", (t) => {
  let input = "Lorem ipsum dolor amet";
  let expected = "Lorem dolor sit amet";
  let a = substance.TextOperation.Insert(18, "sit ");
  let b = substance.TextOperation.Delete(6, "ipsum ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Insert, b=Delete, overlap", (t) => {
  let input = "Lorem dolor sit amet";
  let expected = "Lorem amet";
  let a = substance.TextOperation.Insert(12, "ipsum ");
  let b = substance.TextOperation.Delete(6, "dolor sit ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformations can be done inplace (optimzation for internal use)", (t) => {
  let a = substance.TextOperation.Insert(6, "bla ");
  let b = substance.TextOperation.Insert(6, "blupp ");
  let ops = substance.TextOperation.transform(a, b, {inplace: true});
  t.ok(a.pos === ops[0].pos && b.pos === ops[1].pos, "Transformation should be done inplace.");
  t.end();
});

// Note: In the case of TextOperations conflicts are soft, i.e., there is a defined result
// in such cases. However in certain situations it makes sense to detect such cases, e.g. to notify
// the user to review the result.

test$25("Conflict: Insert at the same position", (t) => {
  let a = substance.TextOperation.Insert(6, "bla");
  let b = substance.TextOperation.Insert(6, "blupp");
  t.ok(a.hasConflict(b), 'Two inserts are considered a conflict if they are at the same position.');
  t.end();
});

test$25("Conflict: Delete with overlapping range", (t) => {
  let a = substance.TextOperation.Delete(4, "bla");
  let b = substance.TextOperation.Delete(6, "blupp");
  t.ok(a.hasConflict(b) && b.hasConflict(a), 'Two deletes are considered a conflict if they overlap.');
  t.end();
});

test$25("Conflict: Delete and Insert with overlapping range", (t) => {
  let a = substance.TextOperation.Insert(4, "bla");
  let b = substance.TextOperation.Delete(2, "blupp");
  t.ok(a.hasConflict(b) && b.hasConflict(a), 'Inserts and Deletes are considered a conflict if they overlap.');
  t.end();
});

test$25("With option 'no-conflict' conflicting operations can not be transformed.", (t) => {
  let a = substance.TextOperation.Insert(4, "bla");
  let b = substance.TextOperation.Delete(2, "blupp");
  t.throws(function() {
    substance.TextOperation.transform(a, b, { "no-conflict": true });
  }, 'Transforming conflicting ops should throw when option "no-conflict" is enabled.');
  t.end();
});

const test$26 = substanceTest.module('TextPropertyComponent');

test$26("Get coordinate of empty property", function(t) {
  var doc = createTestArticle(simple);
  doc.create({
    type: 'paragraph',
    id: 'empty',
    content: ''
  });
  var comp = substance.TextPropertyComponent.mount({
    doc: doc,
    path: ['empty', 'content']
  }, getMountPoint(t));

  var coor = comp.getDOMCoordinate(0);

  t.notNil(coor, 'Coordinate should be not null.');
  t.equal(coor.container, comp.el.getNativeElement(), 'element should be property element');
  t.equal(coor.offset, 0, 'offset should be 0');
  t.end();
});

const test$27 = substanceTest.module('TreeIndex');

test$27("Setting and getting values from a TreeIndex", function(t) {
  var adapter = new substance.TreeIndex();
  adapter.set(['a'], 1);
  adapter.set(['c', 'b'], 2);
  t.equal(adapter.get('a'), 1, 'Top level value should be retrieved correctly.');
  t.equal(adapter.get(['c', 'b']), 2, 'Second level value should be retrieved correctly.');
  t.end();
});

test$27("Getting with invalid arguments", function(t) {
  var adapter = new substance.TreeIndex();
  adapter.set(['a'], 1);
  adapter.set(['c', 'b'], 2);
  t.isNil(adapter.get('e'), 'Should return no value for unknown id');
  t.isNil(adapter.get(['c', 'd']), 'Should return no value for unknown path');
  t.isNil(adapter.get(null), 'Should return no value for null');
  t.isNil(adapter.get(), 'Should return no value for no path');
  t.isNil(adapter.get([]), 'Should return no value for empty path');
  t.isNil(adapter.get({}), 'Should return no value for object');
  t.isNil(adapter.get(1), 'Should return no value for a number');
  t.end();
});

// test("Wrapping an existing object", function(t) {
//   // Note: this way you can create an adapter for plain objects
//   var mine = {}
//   var adapter = TreeIndex.wrap(mine)
//   adapter.set(['a', 'b'], 2)
//   t.equal(adapter.get(['a', 'b']), 2, 'Value should be retrievable.')
//   t.equal(mine.a.b, 2, 'Value should have been written inplace.')
//   t.deepEqual(adapter.get('a'), mine.a, 'Get on first level should return plain content.')
// })

test$27("Getting values recursively", function(t) {
  var adapter = new substance.TreeIndex();
  adapter.set(['a', 'b'], 'foo');
  adapter.set(['a', 'c', 'd'], 'bar');
  t.deepEqual(adapter.getAll('a'), { b: 'foo', d: 'bar'}, 'Values should be collected together into an object');
  t.end();
});

test$27("Arrays: basic usage", function(t) {
  var adapter = new substance.TreeIndex.Arrays();
  t.throws(function() {
    adapter.set('a', []);
  }, 'TreeIndex.set is not allowed for array type');
  adapter.add('a', 1);
  t.deepEqual(adapter.get('a'), [1], 'Get should return an array.');
  adapter.add(['a', 'b'], 2);
  t.deepEqual(adapter.get('a'), [1], 'Content should still be the same after adding a nested value.');
  t.deepEqual(adapter.getAll('a'), [1, 2], 'Collect all values into an array.');
  adapter.add(['a', 'b'], 3);
  adapter.remove(['a', 'b'], 2);
  t.deepEqual(adapter.get(['a', 'b']), [3], 'Only one value should be left after removal.');
  adapter.delete(['a','b']);
  t.isNil(adapter.get(['a', 'b']), 'Value should now be deleted.');
  t.end();
});

const test$28 = substanceTest.module('utils');

test$28('flattenOften should flatten multiple rounds', (t) => {
  let arr = [1, [2, 3, [4, 5]]];
  let result = substance.flattenOften(arr, 2);
  t.deepEqual([1, 2, 3, 4, 5], result, 'array should be flattened.');
  t.end();
});

test$28('flattenOften should stop at max', (t) => {
  let arr = [1, [2, 3, [4, 5]]];
  let result = substance.flattenOften(arr, 1);
  t.deepEqual([1, 2, 3, [4, 5]], result, 'array should be flattened only once.');
  t.end();
});

test$28('isEqual on primitives', (t) => {
  t.ok(substance.isEqual(1,1), '1 === 1');
  t.ok(substance.isEqual("foo", "foo"), '"foo" === "foo"');
  t.notOk(substance.isEqual("foo", "bar"), '"foo" !== "bar"');
  t.notOk(substance.isEqual(null, undefined), 'null !== undefined');
  t.end();
});

test$28('isEqual on arrays', (t) => {
  t.ok(substance.isEqual([1, [2, 3, [4, 5]]], [1, [2, 3, [4, 5]]]), 'should do a deep array comparison');
  t.notOk(substance.isEqual([1, [2, 3, [4, 5]]], [1, [2, 3, [4, 5, 6]]]), 'should detect inequality on deeper level');
  t.end();
});

test$28('isEqual on objects', (t) => {
  t.ok(substance.isEqual({ foo: 'foo', bar: { bla: 'bla', blupp: {} }, baz: [1, 2, 3] }, { foo: 'foo', bar: { bla: 'bla', blupp: {} }, baz: [1, 2, 3] }), 'should do a deep object comparison');
  t.notOk(substance.isEqual({ foo: 'foo', bar: { bla: 'bla', blupp: {} }, baz: [1, 2, 3] }, { foo: 'foo', bar: { bla: 'bla', blupp: { a: 'a'} }, baz: [1, 2, 3] }), 'should detect inequality on deeper level');
  t.end();
});

const CONTENT$2 = '0123456789';

XMLExporterTests();

if (substance.platform.inBrowser) {
  XMLExporterTests('memory');
}

function XMLExporterTests(memory) {

  const test = substanceTest.module('XMLExporter' + ( memory ? ' [memory]' : ''), {
    before: function(t) {
      t.elementFactory = memory ? substance.MemoryDOMElement.createDocument('xml') : substance.DefaultDOMElement.createDocument('xml');
    }
  });

  test("Exporting paragraph", function(t) {
    let { doc, exporter } = setup(t);
    let p1 = doc.create({ type: 'paragraph', id: 'p1', content: CONTENT$2 });
    let el = exporter.convertNode(p1);
    let actual = el.serialize();
    let expected = '<p id="p1">' + CONTENT$2 + '</p>';
    t.equal(actual, expected);
    t.end();
  });

  test("Exporting paragraph with strong", function(t) {
    let { doc, exporter } = setup(t);
    let p1 = doc.create({ type: 'paragraph', id: 'p1', content: CONTENT$2 });
    doc.create({
      type: 'strong',
      id: 's1',
      start: {
        path: ['p1', 'content'],
        offset: 4
      },
      end: {
        offset: 7
      }
    });
    let el = exporter.convertNode(p1);
    let actual = el.serialize();
    let expected = '<p id="p1">0123<strong id="s1">456</strong>789</p>';
    t.equal(actual, expected);
    t.end();
  });

  test("Exporting h1", function(t) {
    let { doc, exporter } = setup(t);
    let h1 = doc.create({ type: 'heading', id: 'h1', level: 1, content: CONTENT$2 });
    let el = exporter.convertNode(h1);
    let actual = el.serialize();
    let expected = '<h1 id="h1">' + CONTENT$2 + '</h1>';
    t.equal(actual, expected);
    t.end();
  });

  test("Exporting h2", function(t) {
    let { doc, exporter } = setup(t);
    let h2 = doc.create({ type: 'heading', id: 'h2', level: 2, content: CONTENT$2 });
    let el = exporter.convertNode(h2);
    let actual = el.serialize();
    let expected = '<h2 id="h2">' + CONTENT$2 + '</h2>';
    t.equal(actual, expected);
    t.end();
  });

  test("Exporting simple document", function(t) {
    let { doc, exporter } = setup(t, simple);
    let rootEl = exporter.exportDocument(doc);
    let actual = rootEl.serialize();
    let expected = [
      '<article>',
      '<p id="p1">' + CONTENT$2 + '</p>',
      '<p id="p2">' + CONTENT$2 + '</p>',
      '<p id="p3">' + CONTENT$2 + '</p>',
      '<p id="p4">' + CONTENT$2 + '</p>',
      '</article>'
    ].join('');
    t.equal(expected, actual);
    t.end();
  });

  test("Exporting meta", function(t) {
    let { doc, exporter } = setup(t);
    let meta = doc.create({
      type: "meta",
      id: "meta",
      title: 'Untitled'
    });
    let el = exporter.convertNode(meta);
    let actual = el.serialize();
    let expected = '<meta id="meta"><title>Untitled</title></meta>';
    t.equal(actual, expected);
    t.ok(true);
    t.end();
  });

  // FIXME: broken since introduction of file nodes
  // test("Exporting image", function(t) {
  //   let data = { type: 'image', id: 'img1', 'src': 'img1.png', 'previewSrc': 'img1preview.png' }
  //   let img = doc.create(data)
  //   let el = exporter.convertNode(img)
  //   t.equal(el.tagName.toLowerCase(), 'image')
  //   t.equal(el.id, 'img1')
  //   t.equal(el.getAttribute('src'), 'img1.png')
  //   t.end()
  // })

  function setup(t, fixture) {
    let config = getTestConfig();
    let exporter = config.createExporter('xml', {}, { elementFactory: t.elementFactory });
    let doc = createTestArticle(fixture);
    return { exporter, doc }
  }
}

const test$29 = substanceTest.module('XMLImporter');

const CONTENT$3 = '0123456789';

test$29("Importing paragraph", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<p id="p1">' + CONTENT$3 + '</p>';
  let el = substance.DefaultDOMElement.parseSnippet(xml, 'xml');
  let node = importer.convertElement(el);
  checkValues(t, node, {
    id: "p1",
    type: "paragraph",
    content: CONTENT$3
  });
  t.end();
});

test$29("Importing paragraph with strong", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<p id="p1">0123<strong id="s1">456</strong>789</p>';
  let el = substance.DefaultDOMElement.parseSnippet(xml, 'xml');
  let p1 = importer.convertElement(el);
  let doc = p1.getDocument();
  let s1 = doc.get('s1');
  t.equal(CONTENT$3, p1.content);
  t.equal('456', s1.getText());
  t.end();
});

test$29("Importing h1", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<h1 id="h1">' + CONTENT$3 + '</h1>';
  let el = substance.DefaultDOMElement.parseSnippet(xml, 'xml');
  let node = importer.convertElement(el);
  checkValues(t, node, {
    id: "h1",
    type: "heading",
    level: 1,
    content: CONTENT$3
  });
  t.end();
});

test$29("Importing h2", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<h2 id="h2">' + CONTENT$3 + '</h2>';
  let el = substance.DefaultDOMElement.parseSnippet(xml, 'xml');
  let node = importer.convertElement(el);
  checkValues(t, node, {
    id: "h2",
    type: "heading",
    level: 2,
    content: CONTENT$3
  });
  t.end();
});

test$29("Importing meta", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<meta><title>' + CONTENT$3 + '</title></meta>';
  let el = substance.DefaultDOMElement.parseSnippet(xml, 'xml');
  let node = importer.convertElement(el);
  checkValues(t, node, {
    id: 'meta',
    type: 'meta',
    title: CONTENT$3
  });
  t.end();
});

// FIXME: broken since introduction of file nodes
test$29("Importing image", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<img id="img1" src="someimage.png"></img>';
  let el = substance.DefaultDOMElement.parseSnippet(xml, 'xml');
  let node = importer.convertElement(el);
  t.equal(node.type, 'image', 'Created node should be of type "image"');
  t.equal(node.id, 'img1', '.. with correct id');
  let file = node.getImageFile();
  t.notNil(file, 'A file node should have been created');
  t.equal(file.url, 'someimage.png', '.. containing the correct URL');
  t.end();
});

function _setupImporter$1(options = {}) {
  let config = getTestConfig();
  let importer = config.createImporter('xml', {}, options);
  return importer
}

})));

//# sourceMappingURL=./tests.js.map